@@ -1,4 +1,4 @@
-/* $Id: alglin1.c,v 1.57 2001/10/01 12:11:27 karim Exp $
+/* $Id: alglin1.c,v 1.61 2001/12/04 21:05:09 karim Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -149,6 +149,8 @@ vconcat(GEN A, GEN B)
   long la,ha,hb,hc,i,j;
   GEN M,a,b,c;
 
+  if (!A) return B;
+  if (!B) return A;
   la = lg(A); if (la==1) return A;
   ha = lg(A[1]); M = cgetg(la,t_MAT);
   hb = lg(B[1]); hc = ha+hb-1;
@@ -668,7 +670,8 @@ GEN
 zeromat(long m, long n)
 {
   GEN y = cgetg(n+1,t_MAT);
-  long i; for (i=1; i<=n; i++) y[i]=(long)zerocol(m);
+  GEN v = zerocol(m);
+  long i; for (i=1; i<=n; i++) y[i]=(long)v;
   return y;
 }
 
@@ -1719,8 +1722,8 @@ gauss_get_pivot_NZ(GEN x, GEN x0/* unused */, GEN c, long i0)
 
 }
 
-/* ~ 0 compared to reference y */
-static int
+/* x ~ 0 compared to reference y */
+int
 approx_0(GEN x, GEN y)
 {
   long tx = typ(x);
@@ -2130,7 +2133,7 @@ FpM_mul(GEN x, GEN y, GEN p)
   long i,j,l,lx=lg(x), ly=lg(y);
   GEN z;
   if (ly==1) return cgetg(ly,t_MAT);
-  if (lx != lg(y[1])) err(operi,"* [mod p]",t_MAT,t_MAT);
+  if (lx != lg(y[1])) err(operi,"* [mod p]",x,y);
   z=cgetg(ly,t_MAT);
   if (lx==1)
   {
@@ -2164,7 +2167,7 @@ FpM_FpV_mul(GEN x, GEN y, GEN p)
 {
   long i,l,lx=lg(x), ly=lg(y);
   GEN z;
-  if (lx != ly) err(operi,"* [mod p]",t_MAT,t_COL);
+  if (lx != ly) err(operi,"* [mod p]",x,y);
   z=cgetg(ly,t_COL);
   if (lx==1) return cgetg(1,t_COL);
   l=lg(x[1]);
@@ -2331,7 +2334,7 @@ FpM_gauss_pivot(GEN x, GEN p, GEN *dd, long *rr)
   GEN c,d,piv;
   long i,j,k,r,t,n,m,av,lim;
 
-  if (!p) return gauss_pivot(x,dd,rr);
+  if (!p) { gauss_pivot(x,dd,rr); return; }
   if (typ(x)!=t_MAT) err(typeer,"FpM_gauss_pivot");
   n=lg(x)-1; if (!n) { *dd=NULL; *rr=0; return; }
 
