@@ -1,4 +1,4 @@
-/* $Id: alglin2.c,v 1.33 2001/10/14 23:20:48 karim Exp $
+/* $Id: alglin2.c,v 1.44 2001/12/14 16:47:44 karim Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -940,63 +940,63 @@ matrixqz0(GEN x,GEN p)
   err(flagerr,"matrixqz"); return NULL; /* not reached */
 }
 
+static int
+ZV_isin(GEN x)
+{
+  long i,l = lg(x);
+  for (i=1; i<l; i++)
+    if (typ(x[i]) != t_INT) return 0;
+  return 1;
+}
+
 GEN
 matrixqz(GEN x, GEN p)
 {
   ulong av = avma, av1, lim;
-  long i,j,j1,m,n,t,nfact;
-  GEN p1,p2,p3,unmodp;
+  long i,j,j1,m,n,nfact;
+  GEN p1,p2,p3;
 
-  if (typ(x)!=t_MAT) err(typeer,"matrixqz");
-  n=lg(x)-1; if (!n) return gcopy(x);
-  m=lg(x[1])-1;
+  if (typ(x) != t_MAT) err(typeer,"matrixqz");
+  n = lg(x)-1; if (!n) return gcopy(x);
+  m = lg(x[1])-1;
   if (n > m) err(talker,"more rows than columns in matrixqz");
   if (n==m)
   {
-    p1=det(x);
+    p1 = det(x);
     if (gcmp0(p1)) err(talker,"matrix of non-maximal rank in matrixqz");
-    avma=av; return idmat(n);
+    avma = av; return idmat(n);
   }
-  p1=cgetg(n+1,t_MAT);
+  /* m > n */
+  p1 = x; x = cgetg(n+1,t_MAT);
   for (j=1; j<=n; j++)
   {
-    p2=gun; p3=(GEN)x[j];
-    for (i=1; i<=m; i++)
-    {
-      t=typ(p3[i]);
-      if (t != t_INT && !is_frac_t(t))
-        err(talker,"not a rational or integral matrix in matrixqz");
-      p2=ggcd(p2,(GEN)p3[i]);
-    }
-    p1[j]=ldiv(p3,p2);
+    x[j] = (long)primpart((GEN)p1[j]);
+    if (!ZV_isin((GEN)p1[j])) err(talker, "not a rational matrix in matrixqz");
   }
-  x=p1; unmodp=cgetg(3,t_INTMOD); unmodp[2]=un;
+  /* x integral */
 
   if (gcmp0(p))
   {
-    p1=cgetg(n+1,t_MAT); p2=gtrans(x);
-    for (j=1; j<=n; j++) p1[j]=p2[j];
-    p3=det(p1); p1[n]=p2[n+1]; p3=ggcd(p3,det(p1));
-    if (!signe(p3))
+    p1 = gtrans(x); setlg(p1,n+1);
+    p2 = det(p1); p1[n] = p1[n+1]; p2 = ggcd(p2,det(p1));
+    if (!signe(p2))
       err(impl,"matrixqz when the first 2 dets are zero");
-    if (gcmp1(p3)) return gerepilecopy(av,x);
+    if (gcmp1(p2)) return gerepilecopy(av,x);
 
-    p3=factor(p3); p1=(GEN)p3[1]; nfact=lg(p1)-1;
+    p1 = (GEN)factor(p2)[1];
   }
-  else
-  {
-    p1=cgetg(2,t_VEC); p1[1]=(long)p; nfact=1;
-  }
-  av1=avma; lim=stack_lim(av1,1);
+  else p1 = _vec(p);
+  nfact = lg(p1)-1;
+  av1 = avma; lim = stack_lim(av1,1);
   for (i=1; i<=nfact; i++)
   {
-    p=(GEN)p1[i]; unmodp[1]=(long)p;
+    p = (GEN)p1[i];
     for(;;)
     {
-      p2=ker(gmul(unmodp,x));
+      p2 = FpM_ker(x, p);
       if (lg(p2)==1) break;
 
-      p2=centerlift(p2); p3=gdiv(gmul(x,p2),p);
+      p2 = centermod(p2,p); p3 = gdiv(gmul(x,p2), p);
       for (j=1; j<lg(p2); j++)
       {
 	j1=n; while (gcmp0(gcoeff(p2,j1,j))) j1--;
@@ -1005,7 +1005,7 @@ matrixqz(GEN x, GEN p)
       if (low_stack(lim, stack_lim(av1,1)))
       {
 	if (DEBUGMEM>1) err(warnmem,"matrixqz");
-	x=gerepilecopy(av1,x);
+	x = gerepilecopy(av1,x);
       }
     }
   }
@@ -1013,84 +1013,133 @@ matrixqz(GEN x, GEN p)
 }
 
 static GEN
-matrixqz_aux(GEN x, long m, long n)
+Z_V_mul(GEN u, GEN A)
+{
+  if (gcmp1(u)) return A;
+  if (gcmp_1(u)) return gneg(A);
+  if (gcmp0(u)) return zerocol(lg(A)-1);
+  return gmul(u,A);
+}
+
+static GEN
+QV_lincomb(GEN u, GEN v, GEN A, GEN B)
+{
+  if (!signe(u)) return Z_V_mul(v,B);
+  if (!signe(v)) return Z_V_mul(u,A);
+  return gadd(Z_V_mul(u,A), Z_V_mul(v,B));
+}
+
+/* cf ZV_elem */
+/* zero aj = Aij (!= 0)  using  ak = Aik (maybe 0), via linear combination of
+ * A[j] and A[k] of determinant 1. */
+static void
+QV_elem(GEN aj, GEN ak, GEN A, long j, long k)
+{
+  GEN p1,u,v,d, D;
+
+  if (gcmp0(ak)) { lswap(A[j],A[k]); return; }
+  D = mpppcm(denom(aj), denom(ak));
+  if (!is_pm1(D)) { aj = gmul(aj,D); ak = gmul(ak,D); }
+  d = bezout(aj,ak,&u,&v);
+  /* frequent special case (u,v) = (1,0) or (0,1) */
+  if (!signe(u))
+  { /* ak | aj */
+    p1 = negi(divii(aj,ak));
+    A[j]   = (long)QV_lincomb(gun, p1, (GEN)A[j], (GEN)A[k]);
+    return;
+  }
+  if (!signe(v))
+  { /* aj | ak */
+    p1 = negi(divii(ak,aj));
+    A[k]   = (long)QV_lincomb(gun, p1, (GEN)A[k], (GEN)A[j]);
+    lswap(A[j], A[k]);
+    return;
+  }
+
+  if (!is_pm1(d)) { aj = divii(aj,d); ak = divii(ak,d); }
+  p1 = (GEN)A[k]; aj = negi(aj);
+  A[k] = (long)QV_lincomb(u,v, (GEN)A[j],p1);
+  A[j] = (long)QV_lincomb(aj,ak, p1,(GEN)A[j]);
+}
+
+extern GEN hnfall_i(GEN A, GEN *ptB, long remove);
+
+static GEN
+matrixqz_aux(GEN A)
 {
   ulong av = avma, lim = stack_lim(av,1);
-  long i,j,k,in[2];
-  GEN p1;
+  long i,j,k,n,m;
+  GEN a;
 
-  for (i=1; i<=m; i++)
+  n = lg(A); if (n == 1) return cgetg(1,t_MAT);
+  m = lg(A[1]);
+  for (i=1; i<m; i++)
   {
-    for(;;)
+    for (j=k=1; j<n; j++)
     {
-      long fl=0;
+      GEN a = gcoeff(A,i,j);
+      if (gcmp0(a)) continue;
 
-      for (j=1; j<=n; j++)
-	if (!gcmp0(gcoeff(x,i,j)))
-	  { in[fl]=j; fl++; if (fl==2) break; }
-      if (j>n) break;
-
-      j=(gcmp(gabs(gcoeff(x,i,in[0]),DEFAULTPREC),
-	      gabs(gcoeff(x,i,in[1]),DEFAULTPREC)) > 0)? in[1]: in[0];
-      p1=gcoeff(x,i,j);
-      for (k=1; k<=n; k++)
-	if (k!=j)
-	  x[k]=lsub((GEN)x[k],gmul(ground(gdiv(gcoeff(x,i,k),p1)),(GEN)x[j]));
+      k = j+1; if (k == n) k = 1;
+      /* zero a = Aij  using  b = Aik */
+      QV_elem(a, gcoeff(A,i,k), A, j,k);
     }
-
-    j=1; while (j<=n && gcmp0(gcoeff(x,i,j))) j++;
-    if (j<=n)
+    a = gcoeff(A,i,k);
+    if (!gcmp0(a))
     {
-      p1=denom(gcoeff(x,i,j));
-      if (!gcmp1(p1)) x[j]=lmul(p1,(GEN)x[j]);
+      a = denom(a);
+      if (!is_pm1(a)) A[k] = lmul(a, (GEN)A[k]);
     }
     if (low_stack(lim, stack_lim(av,1)))
     {
       if(DEBUGMEM>1) err(warnmem,"matrixqz_aux");
-      x=gerepilecopy(av,x);
+      A = gerepilecopy(av,A);
     }
   }
-  return hnf(x);
+  return m > 100? hnfall_i(A,NULL,1): hnf(A);
 }
 
 GEN
 matrixqz2(GEN x)
 {
-  long av = avma,m,n;
-
+  ulong av = avma;
   if (typ(x)!=t_MAT) err(typeer,"matrixqz2");
-  n=lg(x)-1; if (!n) return gcopy(x);
-  m=lg(x[1])-1; x=dummycopy(x);
-  return gerepileupto(av, matrixqz_aux(x,m,n));
+  x = dummycopy(x);
+  return gerepileupto(av, matrixqz_aux(x));
 }
 
 GEN
 matrixqz3(GEN x)
 {
-  long av=avma,av1,j,j1,k,m,n,lim;
+  ulong av = avma, av1, lim;
+  long j,j1,k,m,n;
   GEN c;
 
   if (typ(x)!=t_MAT) err(typeer,"matrixqz3");
-  n=lg(x)-1; if (!n) return gcopy(x);
-  m=lg(x[1])-1; x=dummycopy(x); c=new_chunk(n+1);
-  for (j=1; j<=n; j++) c[j]=0;
-  av1=avma; lim=stack_lim(av1,1);
-  for (k=1; k<=m; k++)
+  n = lg(x); if (n==1) return gcopy(x);
+  m = lg(x[1]); x = dummycopy(x);
+  c = cgetg(n,t_VECSMALL);
+  for (j=1; j<n; j++) c[j] = 0;
+  av1 = avma; lim = stack_lim(av1,1);
+  for (k=1; k<m; k++)
   {
-    j=1; while (j<=n && (c[j] || gcmp0(gcoeff(x,k,j)))) j++;
-    if (j<=n)
-    {
-      c[j]=k; x[j]=ldiv((GEN)x[j],gcoeff(x,k,j));
-      for (j1=1; j1<=n; j1++)
-	if (j1!=j) x[j1]=lsub((GEN)x[j1],gmul(gcoeff(x,k,j1),(GEN)x[j]));
-      if (low_stack(lim, stack_lim(av1,1)))
+    j=1; while (j<n && (c[j] || gcmp0(gcoeff(x,k,j)))) j++;
+    if (j==n) continue;
+   
+    c[j]=k; x[j]=ldiv((GEN)x[j],gcoeff(x,k,j));
+    for (j1=1; j1<n; j1++)
+      if (j1!=j)
       {
-	if(DEBUGMEM>1) err(warnmem,"matrixqz3");
-	x=gerepilecopy(av1,x);
+        GEN t = gcoeff(x,k,j1);
+        if (!gcmp0(t)) x[j1] = lsub((GEN)x[j1],gmul(t,(GEN)x[j]));
       }
+    if (low_stack(lim, stack_lim(av1,1)))
+    {
+      if(DEBUGMEM>1) err(warnmem,"matrixqz3");
+      x = gerepilecopy(av1,x);
     }
   }
-  return gerepileupto(av, matrixqz_aux(x,m,n));
+  return gerepileupto(av, matrixqz_aux(x));
 }
 
 GEN
@@ -2214,9 +2263,9 @@ allhnfmod(GEN x,GEN dm,int flag)
         if (lgefint(p1[k]) > ldm) p1[k] = lresii((GEN)p1[k], dm);
       if (low_stack(lim, stack_lim(av,1)))
       {
+        GEN *gptr[2]; gptr[0]=&w; gptr[1]=&dm;
         if (DEBUGMEM>1) err(warnmem,"allhnfmod[2]. i=%ld", i);
-        tetpil=avma; w=gerepile(av,tetpil,ZM_copy(w));
-        diag = gcoeff(w,i,i);
+        gerepilemany(av,gptr,2); diag = gcoeff(w,i,i);
       }
     }
   }
@@ -2314,29 +2363,26 @@ findi(GEN M)
   return 0;
 }
 
+static long
+findi_normalize(GEN Aj, GEN B, long j, GEN **lambda)
+{
+  long r = findi(Aj);
+  if (r && signe(Aj[r]) < 0)
+  {
+    neg_col(Aj); if (B) neg_col((GEN)B[j]);
+    Minus(j,lambda);
+  }
+  return r;
+}
+
 static void
 reduce2(GEN A, GEN B, long k, long j, long *row, GEN **lambda, GEN *D)
 {
   GEN q;
   long i, row0, row1;
 
-  row0 = findi((GEN)A[j]);
-  if (row0 && signe(gcoeff(A,row0,j)) < 0)
-  {
-    neg_col((GEN)A[j]);
-    if (B) neg_col((GEN)B[j]);
-    Minus(j,lambda);
-  }
-
-  row1 = findi((GEN)A[k]);
-  if (row1 && signe(gcoeff(A,row1,k)) < 0)
-  {
-    neg_col((GEN)A[k]);
-    if (B) neg_col((GEN)B[k]);
-    Minus(k,lambda);
-  }
-
-  row[0]=row0; row[1]=row1;
+  row[0] = row0 = findi_normalize((GEN)A[j], B,j,lambda);
+  row[1] = row1 = findi_normalize((GEN)A[k], B,k,lambda);
   if (row0)
     q = truedvmdii(gcoeff(A,row0,k), gcoeff(A,row0,j), NULL);
   else if (absi_cmp(shifti(lambda[k][j], 1), D[j]) > 0)
@@ -2398,40 +2444,6 @@ hnfswap(GEN A, GEN B, long k, GEN **lambda, GEN *D)
   D[k-1] = divii(addii(p1,p2), D[k-1]);
 }
 
-/* A[k] = 0,  A[nz] != 0,  k > nz,  A[j] = 0 for all j < nz.
- * "Deep insert" A[k] at nz */
-static void
-trivswap(GEN *A, GEN *B, long nz, long k, GEN **lambda, GEN *D)
-{
-  GEN p1;
-  long i,j,n = lg(A);
-
-  p1 = A[nz]; /* cycle A */
-  for (j = nz; j < k; j++) swap(A[j+1], p1); 
-  A[nz] = p1; /* = [0...0] */
-
-  if (B)
-  {
-    p1 = B[nz]; /* cycle B */
-    for (j = nz; j < k; j++) swap(B[j+1], p1); 
-    B[nz] = p1;
-  }
-
-  p1 = D[nz]; /* cycle D */
-  for (j = nz; j < k; j++) swap(D[j+1], p1);
-  D[nz] = gun;
-
-  for (j=k-1; j>=nz; j--) /* cycle lambda */
-    for (i=k-1; i>=nz; i--) lambda[i+1][j+1] = lambda[i][j];
-  for (j=n-1; j>k; j--)
-    for (i=k-1; i>=nz; i--) 
-    {
-      lambda[i+1][j] = lambda[i][j];
-      lambda[j][i+1] = lambda[j][i];
-    }
-  for (i=1; i<n; i++) lambda[nz][i] = lambda[i][nz] = gzero;
-}
-
 static GEN
 fix_rows(GEN A)
 {
@@ -2449,12 +2461,11 @@ fix_rows(GEN A)
 }
 
 GEN
-hnflll_i(GEN A, GEN *ptB, int remove)
+hnflll_i(GEN A, GEN *ptB)
 {
   ulong av = avma, lim = stack_lim(av,3);
   long m1 = 1, n1 = 1; /* alpha = m1/n1. Maybe 3/4 here ? */
   long row[2], do_swap,i,n,k;
-  long nzcol = 1; /* index of 1st (maybe) non-0 col [only updated when !B] */
   GEN z,B, **lambda, *D;
   GEN *gptr[4];
 
@@ -2462,59 +2473,43 @@ hnflll_i(GEN A, GEN *ptB, int remove)
   n = lg(A);
   A = ZM_copy(fix_rows(A));
   B = ptB? idmat(n-1): NULL;
-  D = (GEN*) cgetg(n+1, t_VEC); D++; /* hack: need a "sentinel" D[0] */
-  if (n == 2) /* handle trivial case: return negative diag coeff otherwise */
-  {
-    i = findi((GEN)A[1]);
-    if (i && signe(gcoeff(A,i,1)) < 0)
-    {
-      neg_col((GEN)A[1]);
-      if (B) neg_col((GEN)B[1]);
-    }
-  }
-  lambda = (GEN**) cgetg(n,t_MAT);
-  for (i=1; i<n; i++) { D[i] = gun; lambda[i] = (GEN*)zerocol(n-1); }
-  D[0] = gun; k = 2;
+  D = (GEN*)cgetg(n+1,t_VEC); lambda = (GEN**) cgetg(n,t_MAT);
+  D++;
+  for (i=0; i<n; i++) D[i] = gun;
+  for (i=1; i<n; i++) lambda[i] = (GEN*)zerocol(n-1);
+  k = 2;
   while (k < n)
   {
     reduce2(A,B,k,k-1,row,lambda,D);
-    if (remove)
-    { /* not interested in kernel. Eliminate 0 cols */
-      if (!row[0] || !row[1])
-      {
-        while (!findi((GEN)A[nzcol]) && nzcol < n) nzcol++;
-        /* A[nzcol] != 0,  A[i] = 0 for i < nzcol */
-        if (!row[0] && k-1 > nzcol)
-          {trivswap((GEN*)A,(GEN*)B,nzcol,k-1, lambda,D); nzcol++;}
-        if (!row[1] && k   > nzcol)
-          {trivswap((GEN*)A,(GEN*)B,nzcol,k  , lambda,D); nzcol++;}
-        if (k <= nzcol) k = nzcol+1;
-        continue;
-      }
-      do_swap = (row[0] && row[0] <= row[1]);
+    if (row[0])
+      do_swap = (!row[1] || row[0] <= row[1]);
+    else if (!row[1])
+    { /* row[0] == row[1] == 0 */
+      ulong av1 = avma;
+      z = addii(mulii(D[k-2],D[k]), sqri(lambda[k][k-1]));
+      do_swap = (cmpii(mulsi(n1,z), mulsi(m1,sqri(D[k-1]))) < 0);
+      avma = av1;
     }
     else
-    {
-      if (row[0])
-        do_swap = (!row[1] || row[0] <= row[1]);
-      else if (row[1])
-        do_swap = 0;
-      else
-      { /* row[0] == row[1] == 0 */
-        ulong av1 = avma;
-        z = addii(mulii(D[k-2],D[k]), sqri(lambda[k][k-1]));
-        do_swap = (cmpii(mulsi(n1,z), mulsi(m1,sqri(D[k-1]))) < 0);
-        avma = av1;
-      }
-    }
+      do_swap = 0;
     if (do_swap)
     {
       hnfswap(A,B,k,lambda,D);
-      if (k > nzcol+1) k--;
+      if (k > 2) k--;
     }
     else
     {
-      for (i=k-2; i>=nzcol; i--) reduce2(A,B,k,i,row,lambda,D);
+      for (i=k-2; i; i--)
+      {
+        reduce2(A,B,k,i,row,lambda,D);
+        if (low_stack(lim, stack_lim(av,3)))
+        {
+          GEN a = (GEN)lambda, b = (GEN)(D-1); /* gcc -Wall */
+          gptr[0]=&A; gptr[1]=&a; gptr[2]=&b; gptr[3]=&B; 
+          if (DEBUGMEM) err(warnmem,"hnflll (reducing), i = %ld",i);
+          gerepilemany(av,gptr,B? 4: 3); lambda = (GEN**)a; D = (GEN*)(b+1);
+        }
+      }
       k++;
     }
     if (low_stack(lim, stack_lim(av,3)))
@@ -2525,13 +2520,8 @@ hnflll_i(GEN A, GEN *ptB, int remove)
       gerepilemany(av,gptr,B? 4: 3); lambda = (GEN**)a; D = (GEN*)(b+1);
     }
   }
-  if (remove)
-  {
-    for (i=nzcol; i<n; i++)
-      if (findi((GEN)A[i])) break;
-    i--; A += i; A[0] = evaltyp(t_MAT)|evallg(n-i);
-    if (B) { B += i; B[0] = evaltyp(t_MAT)|evallg(n-i); }
-  }
+  /* handle trivial case: return negative diag coeff otherwise */
+  if (n == 2) (void)findi_normalize((GEN)A[1], B,1,lambda);
   A = fix_rows(A);
   gptr[0] = &A; gptr[1] = &B;
   gerepilemany(av, gptr, B? 2: 1);
@@ -2543,7 +2533,7 @@ GEN
 hnflll(GEN A)
 {
   GEN B, z = cgetg(3, t_VEC);
-  z[1] = (long)hnflll_i(A, &B, 0);
+  z[1] = (long)hnflll_i(A, &B);
   z[2] = (long)B; return z;
 }
 
@@ -2562,15 +2552,15 @@ reduce1(GEN A, GEN B, long k, long j, GEN **lambda, GEN *D)
   else
     return;
 
-  if (! gcmp0(q))
+  if (signe(q))
   {
+    GEN *Lk = lambda[k], *Lj = lambda[j];
     q = mynegi(q);
     A[k] = laddii((GEN)A[k], mulii(q,(GEN)A[j]));
     elt_col((GEN)B[k],(GEN)B[j],q);
-    lambda[k][j] = addii(lambda[k][j], mulii(q,D[j]));
+    Lk[j] = addii(Lk[j], mulii(q,D[j]));
     for (i=1; i<j; i++)
-      if (signe(lambda[j][i]))
-        lambda[k][i] = addii(lambda[k][i], mulii(q,lambda[j][i]));
+      if (signe(Lj[i])) Lk[i] = addii(Lk[i], mulii(q,Lj[i]));
   }
 }
 
@@ -2578,17 +2568,15 @@ GEN
 extendedgcd(GEN A)
 {
   long m1 = 1, n1 = 1; /* alpha = m1/n1. Maybe 3/4 here ? */
-  long av = avma, tetpil, do_swap,i,j,n,k;
-  GEN p1,p2,B, **lambda, *D;
+  long av = avma, do_swap,i,n,k;
+  GEN z,B, **lambda, *D;
 
-  n = lg(A); B = idmat(n-1); A = ZM_copy(A);
-  D = (GEN*) cgeti(n); lambda = (GEN**) cgetg(n,t_MAT);
+  n = lg(A);
+  A = dummycopy(A);
+  B = idmat(n-1);
+  D = (GEN*)new_chunk(n); lambda = (GEN**) cgetg(n,t_MAT);
   for (i=0; i<n; i++) D[i] = gun;
-  for (i=1; i<n; i++)
-  {
-    lambda[i] = (GEN*) cgetg(n,t_COL);
-    for(j=1;j<n;j++) lambda[i][j] = gzero;
-  }
+  for (i=1; i<n; i++) lambda[i] = (GEN*)zerocol(n-1);
   k = 2;
   while (k < n)
   {
@@ -2596,19 +2584,17 @@ extendedgcd(GEN A)
     if (signe(A[k-1])) do_swap = 1;
     else if (!signe(A[k]))
     {
-      long av1=avma;
-      p1 = addii(mulii(D[k-2],D[k]), sqri(lambda[k][k-1]));
-      p1 = mulsi(n1,p1);
-      p2 = mulsi(m1,sqri(D[k-1]));
-      do_swap = (cmpii(p1,p2) < 0);
-      avma=av1;
+      ulong av1 = avma;
+      z = addii(mulii(D[k-2],D[k]), sqri(lambda[k][k-1]));
+      do_swap = (cmpii(mulsi(n1,z), mulsi(m1,sqri(D[k-1]))) < 0);
+      avma = av1;
     }
     else do_swap = 0;
 
     if (do_swap)
     {
       hnfswap(A,B,k,lambda,D);
-      if (k>2) k--;
+      if (k > 2) k--;
     }
     else
     {
@@ -2616,16 +2602,15 @@ extendedgcd(GEN A)
       k++;
     }
   }
-  if (signe(A[n-1])<0)
+  if (signe(A[n-1]) < 0)
   {
     A[n-1] = (long)mynegi((GEN)A[n-1]);
     neg_col((GEN)B[n-1]);
   }
-  tetpil = avma;
-  p1 = cgetg(3,t_VEC);
-  p1[1]=lcopy((GEN)A[n-1]);
-  p1[2]=lcopy(B);
-  return gerepile(av,tetpil,p1);
+  z = cgetg(3,t_VEC);
+  z[1] = A[n-1];
+  z[2] = (long)B;
+  return gerepilecopy(av, z);
 }
 
 /* HNF with permutation. TODO: obsolete, replace with hnflll. */
@@ -2906,208 +2891,220 @@ trivsmith(long all)
   z[3]=lgetg(1,t_MAT); return z;
 }
 
-/* Return the smith normal form d of matrix x. If all != 0 return [d,u,v],
- * where d = u.x.v
- */
-static GEN
-smithall(GEN x, long all)
+static void
+snf_pile(ulong av, GEN *x, GEN *U, GEN *V)
 {
-  long av,tetpil,i,j,k,l,c,fl,n,s1,s2,lim;
-  GEN p1,p2,p3,p4,z,b,u,v,d,ml,mr,mun,mdet,ys;
+  GEN *gptr[3];
+  int c = 1; gptr[0]=x;
+  if (*U) gptr[c++] = U;
+  if (*V) gptr[c++] = V;
+  gerepilemany(av,gptr,c);
+}
+
+/* Return the SNF of matrix x. If all != 0 return [d,u,v],
+ * where d = u.x.v */
+GEN
+smithall(GEN x, GEN *ptU, GEN *ptV)
+{
+  ulong av = avma, lim = stack_lim(av,1);
+  long i,j,k,l,c,n,s1,s2;
+  GEN p1,p2,p3,p4,b,u,v,d,U,V,mun,mdet,ys;
 
   if (typ(x)!=t_MAT) err(typeer,"smithall");
-  if (DEBUGLEVEL>=9) outerr(x);
-  n=lg(x)-1;
-  if (!n) return trivsmith(all);
+  if (DEBUGLEVEL>8) outerr(x);
+  n = lg(x)-1;
+  if (!n) {
+    if (ptU) *ptU = cgetg(1,t_MAT);
+    if (ptV) *ptV = cgetg(1,t_MAT);
+    return cgetg(1,t_MAT);
+  }
   if (lg(x[1]) != n+1) err(mattype1,"smithall");
   for (i=1; i<=n; i++)
     for (j=1; j<=n; j++)
       if (typ(coeff(x,i,j)) != t_INT)
         err(talker,"non integral matrix in smithall");
 
-  av = avma; lim = stack_lim(av,1);
-  x = dummycopy(x); mdet = detint(x);
-  if (ishnfall(x)) { if (all) { ml=idmat(n); mr=idmat(n); } }
+  U = ptU? gun: NULL;
+  V = ptV? gun: NULL;
+  x = dummycopy(x);
+  if (ishnfall(x))
+  {
+    mdet = dethnf_i(x);
+    if (V) V = idmat(n);
+  }
   else
   {
+    mdet = detint(x);
     if (signe(mdet))
     {
-      p1=hnfmod(x,mdet);
-      if (all) { ml=idmat(n); mr=gauss(x,p1); }
+      p1 = hnfmod(x,mdet);
+      if (V) V = gauss(x,p1);
     }
     else
     {
-      if (all)
-      {
-        p1 = hnfall0(x,0);
-        ml = idmat(n); mr = (GEN)p1[2]; p1 = (GEN)p1[1];
-      }
-      else
-        p1 = hnf0(x,0);
+      p1 = hnfall_i(x, ptV, 0);
+      if (ptV) V = *ptV;
     }
     x = p1;
   }
+  if (U) U = idmat(n);
+ 
   p1=cgetg(n+1,t_VEC); for (i=1; i<=n; i++) p1[i]=lnegi(gcoeff(x,i,i));
-  p2=sindexsort(p1); ys=cgetg(n+1,t_MAT);
+  p2=sindexsort(p1);
+  ys = cgetg(n+1,t_MAT);
   for (j=1; j<=n; j++)
   {
     p1=cgetg(n+1,t_COL); ys[j]=(long)p1;
     for (i=1; i<=n; i++) p1[i]=coeff(x,p2[i],p2[j]);
   }
   x = ys;
-  if (all)
+  if (U)
   {
-    p3=cgetg(n+1,t_MAT); p4=cgetg(n+1,t_MAT);
-    for (j=1; j<=n; j++) { p3[j]=ml[p2[j]]; p4[j]=mr[p2[j]]; }
-    ml=p3; mr=p4;
+    p1 = cgetg(n+1,t_MAT);
+    for (j=1; j<=n; j++) p1[j]=U[p2[j]];
+    U = p1;
   }
+  if (V)
+  {
+    p1 = cgetg(n+1,t_MAT);
+    for (j=1; j<=n; j++) p1[j]=V[p2[j]];
+    V = p1;
+  }
+ 
   if (signe(mdet))
   {
     p1 = hnfmod(x,mdet);
-    if (all) mr=gmul(mr,gauss(x,p1));
+    if (V) V = gmul(V, gauss(x,p1));
   }
   else
   {
-    if (all)
-    {
-      p1 = hnfall0(x,0);
-      mr = gmul(mr,(GEN)p1[2]); p1 = (GEN)p1[1];
-    }
-    else
-      p1 = hnf0(x,0);
+    p1 = hnfall_i(x,ptV,0);
+    if (ptV) V = gmul(V, *ptV);
   }
-  x=p1; mun = negi(gun);
+  x = p1; mun = negi(gun);
 
-  if (DEBUGLEVEL>7) {fprintferr("starting SNF loop");flusherr();}
-  for (i=n; i>=2; i--)
+  if (DEBUGLEVEL>7) fprintferr("starting SNF loop");
+  for (i=n; i>1; i--)
   {
-    if (DEBUGLEVEL>7) {fprintferr("\ni = %ld: ",i);flusherr();}
+    if (DEBUGLEVEL>7) fprintferr("\ni = %ld: ",i);
     for(;;)
     {
       c = 0;
       for (j=i-1; j>=1; j--)
       {
 	p1=gcoeff(x,i,j); s1 = signe(p1);
-	if (s1)
-	{
-	  p2=gcoeff(x,i,i);
-          if (!absi_cmp(p1,p2))
+	if (!s1) continue;
+       
+        p2=gcoeff(x,i,i);
+        if (!absi_cmp(p1,p2))
+        {
+          s2=signe(p2);
+          if (s1 == s2) { d=p1; u=gun; p4=gun; }
+          else
           {
-            s2=signe(p2);
-            if (s1 == s2) { d=p1; u=gun; p4=gun; }
-            else
-	    {
-              if (s2>0) { u = gun; p4 = mun; }
-              else      { u = mun; p4 = gun; }
-	      d=(s1>0)? p1: absi(p1);
-	    }
-            v = gzero; p3 = u;
+            if (s2>0) { u = gun; p4 = mun; }
+            else      { u = mun; p4 = gun; }
+            d=(s1>0)? p1: absi(p1);
           }
-          else { d=bezout(p2,p1,&u,&v); p3=divii(p2,d); p4=divii(p1,d); }
-	  for (k=1; k<=i; k++)
-	  {
-	    b=addii(mulii(u,gcoeff(x,k,i)),mulii(v,gcoeff(x,k,j)));
-	    coeff(x,k,j)=lsubii(mulii(p3,gcoeff(x,k,j)),
-	                        mulii(p4,gcoeff(x,k,i)));
-	    coeff(x,k,i)=(long)b;
-	  }
-	  if (all) update(u,v,p3,p4,(GEN*)(mr+i),(GEN*)(mr+j));
-          if (low_stack(lim, stack_lim(av,1)))
-	  {
-	    if (DEBUGMEM>1) err(warnmem,"[1]: smithall");
-	    if (all)
-	    {
-	      GEN *gptr[3]; gptr[0]=&x; gptr[1]=&ml; gptr[2]=&mr;
-	      gerepilemany(av,gptr,3);
-	    }
-	    else x=gerepileupto(av, ZM_copy(x));
-	  }
-	}
+          v = gzero; p3 = u;
+        }
+        else { d=bezout(p2,p1,&u,&v); p3=divii(p2,d); p4=divii(p1,d); }
+        for (k=1; k<=i; k++)
+        {
+          b=addii(mulii(u,gcoeff(x,k,i)),mulii(v,gcoeff(x,k,j)));
+          coeff(x,k,j)=lsubii(mulii(p3,gcoeff(x,k,j)),
+                              mulii(p4,gcoeff(x,k,i)));
+          coeff(x,k,i)=(long)b;
+        }
+        if (V) update(u,v,p3,p4,(GEN*)(V+i),(GEN*)(V+j));
+        if (low_stack(lim, stack_lim(av,1)))
+        {
+          if (DEBUGMEM>1) err(warnmem,"[1]: smithall");
+          snf_pile(av, &x,&U,&V);
+        }
       }
-      if (DEBUGLEVEL>=8) {fprintferr("; ");flusherr();}
+      if (DEBUGLEVEL>7) fprintferr("; ");
       for (j=i-1; j>=1; j--)
       {
 	p1=gcoeff(x,j,i); s1 = signe(p1);
-	if (s1)
-	{
-	  p2=gcoeff(x,i,i);
-	  if (!absi_cmp(p1,p2))
+	if (!s1) continue;
+       
+        p2=gcoeff(x,i,i);
+        if (!absi_cmp(p1,p2))
+        {
+          s2 = signe(p2);
+          if (s1 == s2) { d=p1; u=gun; p4=gun; }
+          else
           {
-            s2 = signe(p2);
-            if (s1 == s2) { d=p1; u=gun; p4=gun; }
-            else
-	    {
-              if (s2>0) { u = gun; p4 = mun; }
-              else      { u = mun; p4 = gun; }
-	      d=(s1>0)? p1: absi(p1);
-	    }
-            v = gzero; p3 = u;
+            if (s2>0) { u = gun; p4 = mun; }
+            else      { u = mun; p4 = gun; }
+            d=(s1>0)? p1: absi(p1);
           }
-          else { d=bezout(p2,p1,&u,&v); p3=divii(p2,d); p4=divii(p1,d); }
-	  for (k=1; k<=i; k++)
-	  {
-	    b=addii(mulii(u,gcoeff(x,i,k)),mulii(v,gcoeff(x,j,k)));
-	    coeff(x,j,k)=lsubii(mulii(p3,gcoeff(x,j,k)),
-	                        mulii(p4,gcoeff(x,i,k)));
-	    coeff(x,i,k)=(long)b;
-	  }
-	  if (all) update(u,v,p3,p4,(GEN*)(ml+i),(GEN*)(ml+j));
-	  c = 1;
-	}
+          v = gzero; p3 = u;
+        }
+        else { d=bezout(p2,p1,&u,&v); p3=divii(p2,d); p4=divii(p1,d); }
+        for (k=1; k<=i; k++)
+        {
+          b=addii(mulii(u,gcoeff(x,i,k)),mulii(v,gcoeff(x,j,k)));
+          coeff(x,j,k)=lsubii(mulii(p3,gcoeff(x,j,k)),
+                              mulii(p4,gcoeff(x,i,k)));
+          coeff(x,i,k)=(long)b;
+        }
+        if (U) update(u,v,p3,p4,(GEN*)(U+i),(GEN*)(U+j));
+        c = 1;
       }
       if (!c)
       {
-	b=gcoeff(x,i,i); fl=1;
-	if (signe(b))
-	{
-	  for (k=1; k<i && fl; k++)
-	    for (l=1; l<i && fl; l++)
-	      fl = (int)!signe(resii(gcoeff(x,k,l),b));
-          /* cast to (int) necessary for gcc-2.95 on sparcv9-64 (IS) */
-	  if (!fl)
-	  {
-	    k--;
-	    for (l=1; l<=i; l++)
-	      coeff(x,i,l)=laddii(gcoeff(x,i,l),gcoeff(x,k,l));
-	    if (all) ml[i]=ladd((GEN)ml[i],(GEN)ml[k]);
-	  }
-	}
-        if (fl) break;
+	b = gcoeff(x,i,i);
+	if (!signe(b)) break;
+       
+        for (k=1; k<i; k++)
+        {
+          for (l=1; l<i; l++)
+            if (signe(resii(gcoeff(x,k,l),b))) break;
+          if (l != i) break;
+        }
+        if (k == i) break;
+       
+        /* x[k,l] != 0 mod b */
+        for (l=1; l<=i; l++)
+          coeff(x,i,l) = laddii(gcoeff(x,i,l),gcoeff(x,k,l));
+        if (U) U[i] = ladd((GEN)U[i],(GEN)U[k]);
       }
       if (low_stack(lim, stack_lim(av,1)))
       {
-	if (DEBUGMEM>1) err(warnmem,"[2]: smithall");
-	if (all)
-	{
-	  GEN *gptr[3]; gptr[0]=&x; gptr[1]=&ml; gptr[2]=&mr;
-	  gerepilemany(av,gptr,3);
-	}
-	else x=gerepileupto(av,ZM_copy(x));
+        if (DEBUGMEM>1) err(warnmem,"[2]: smithall");
+        snf_pile(av, &x,&U,&V);
       }
     }
   }
-  if (DEBUGLEVEL>7) {fprintferr("\n");flusherr();}
-  if (all)
-  {
-    for (k=1; k<=n; k++)
-      if (signe(gcoeff(x,k,k))<0)
-        { mr[k]=lneg((GEN)mr[k]); coeff(x,k,k)=lnegi(gcoeff(x,k,k)); }
-    tetpil=avma; z=cgetg(4,t_VEC);
-    z[1]=ltrans(ml); z[2]=lcopy(mr); z[3]=lcopy(x);
-    return gerepile(av,tetpil,z);
-  }
-  tetpil=avma; z=cgetg(n+1,t_VEC); j=n;
-  for (k=n; k; k--)
-    if (signe(gcoeff(x,k,k))) z[j--]=labsi(gcoeff(x,k,k));
-  for (   ; j; j--) z[j]=zero;
-  return gerepile(av,tetpil,z);
+  if (DEBUGLEVEL>7) fprintferr("\n");
+  for (k=1; k<=n; k++)
+    if (signe(gcoeff(x,k,k)) < 0)
+    {
+      if (V) V[k]=lneg((GEN)V[k]);
+      coeff(x,k,k) = lnegi(gcoeff(x,k,k));
+    }
+  if (!U && !V)
+    return gerepileupto(av, mattodiagonal(x));
+   
+  if (U) U = gtrans_i(U);
+  snf_pile(av, &x,&U,&V);
+  if (ptU) *ptU = U;
+  if (ptV) *ptV = V;
+  return x;
 }
 
 GEN
-smith(GEN x) { return smithall(x,0); }
+smith(GEN x) { return smithall(x, NULL,NULL); }
 
 GEN
-smith2(GEN x) { return smithall(x,1); }
+smith2(GEN x)
+{
+  GEN z = cgetg(4, t_VEC);
+  z[3] = (long)smithall(x, (GEN*)(z+1),(GEN*)(z+2));
+  return z;
+}
 
 /* Assume z was computed by [g]smithall(). Remove the 1s on the diagonal */
 GEN
@@ -3267,10 +3264,10 @@ matsnf0(GEN x,long flag)
   long av = avma;
   if (flag > 7) err(flagerr,"matsnf");
   if (typ(x) == t_VEC && flag & 4) return smithclean(x);
-  if (flag & 2) x = gsmithall(x,flag & 1);
-  else          x = smithall(x, flag & 1);
-  if (flag & 4) x = smithclean(x);
-  return gerepileupto(av, x);
+  if (flag & 2) x = flag&1 ? gsmith2(x): gsmith(x);
+  else          x = flag&1 ?  smith2(x):  smith(x);
+  if (flag & 4) x = gerepileupto(av, smithclean(x));
+  return x;
 }
 
 GEN
