@@ -1,4 +1,4 @@
-/* $Id: arith1.c,v 1.57 2001/10/15 19:01:32 karim Exp $
+/* $Id: arith1.c,v 1.69 2002/01/09 19:13:44 karim Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -530,7 +530,7 @@ gcarreparfait(GEN x)
 
     case t_INTMOD:
     {
-      GEN b, q, e;
+      GEN b, q;
       long w;
       a = (GEN)x[2]; if (!signe(a)) return gun;
       av = avma;
@@ -557,23 +557,18 @@ gcarreparfait(GEN x)
       if (i==0)
       {
         GEN d = mppgcd(a,q);
-        p1 = factor(d);
-        p = (GEN)p1[1];
-        e = (GEN)p1[2]; l = lg(e);
+        p = (GEN)factor(d)[1]; l = lg(p);
         for (i=1; i<l; i++)
         {
           v = pvaluation(a,(GEN)p[i],&p1);
-          w = itos((GEN)e[i]);
+          w = pvaluation(q,(GEN)p[i], &q);
           if (v < w && (v&1 || kronecker(p1,(GEN)p[i]) == -1)) 
             { avma = av; return gzero; }
-          q = diviiexact(q, gpowgs((GEN)p[i], w));
         }
         if (kronecker(a,q) == -1) { avma = av; return gzero; }
       }
       /* kro(a,q) = 1, q odd: need to factor q */
-      p1 = factor(q);
-      p = (GEN)p1[1]; 
-      e = (GEN)p1[2]; l = lg(e) - 1;
+      p = (GEN)factor(q)[1]; l = lg(p) - 1;
       /* kro(a,q) = 1, check all p|q but the last (product formula) */
       for (i=1; i<l; i++)
         if (kronecker(a,(GEN)p[i]) == -1) { avma = av; return gzero; }
@@ -1388,12 +1383,27 @@ mpinvmod(GEN a, GEN m)
 
 /*********************************************************************/
 /**                                                                 **/
-/**                   POWERING MODULO (A^N mod M)                   **/
+/**                    MODULAR EXPONENTIATION                       **/
 /**                                                                 **/
 /*********************************************************************/
 GEN resiimul(GEN x, GEN y);
-GEN resmod2n(GEN x, long y);
-GEN _resii(GEN x, GEN y) { return resii(x,y); }
+static GEN _resii(GEN x, GEN y) { return resii(x,y); }
+
+/* Montgomery reduction */
+
+typedef struct {
+  GEN N;
+  ulong inv; /* inv = -N^(-1) mod B, */
+} montdata;
+
+extern ulong invrev(ulong b);
+
+void
+init_montdata(GEN N, montdata *s)
+{
+  s->N = N;
+  s->inv = (ulong) -invrev(modBIL(N));
+}
 
 GEN
 init_remainder(GEN M)
@@ -1407,72 +1417,134 @@ init_remainder(GEN M)
 }
 
 /* optimal on UltraSPARC */
-static long RESIIMUL_LIMIT = 150;
+static long RESIIMUL_LIMIT   = 150;
+static long MONTGOMERY_LIMIT = 32;
 
 void
 setresiilimit(long n) { RESIIMUL_LIMIT = n; }
+void
+setmontgomerylimit(long n) { MONTGOMERY_LIMIT = n; }
 
-GEN
-powmodulo(GEN A, GEN N, GEN M)
+typedef struct {
+  GEN N;
+  GEN (*res)(GEN,GEN);
+  GEN (*mulred)(GEN,GEN,GEN);
+} muldata;
+
+/* reduction for multiplication by 2 */
+static GEN
+_redsub(GEN x, GEN N)
 {
-  GEN y;
-  long av = avma,av1,lim,man,k,nb,s, *p;
-  GEN (*mul)(GEN,GEN) = mulii;
-  GEN (*res)(GEN,GEN) = _resii;
+  return (cmpii(x,N) >= 0)? subii(x,N): x;
+}
+/* Montgomery reduction */
+extern GEN red_montgomery(GEN T, GEN N, ulong inv);
+static GEN
+montred(GEN x, GEN N)
+{
+  return red_montgomery(x, ((montdata*)N)->N, ((montdata*)N)->inv);
+}
+/* 2x mod N */
+static GEN
+_muli2red(GEN x, GEN y/* ignored */, GEN N) {
+  (void)y; return _redsub(shifti(x,1), N);
+}
+static GEN
+_muli2montred(GEN x, GEN y/* ignored */, GEN N) {
+  GEN n = ((montdata*)N)->N;
+  GEN z = _muli2red(x,y, n);
+  long l = lgefint(n);
+  while (lgefint(z) > l) z = subii(z,n);
+  return z;
+}
+static GEN
+_muli2invred(GEN x, GEN y/* ignored */, GEN N) {
+  return _muli2red(x,y, (GEN)N[1]);
+}
+/* xy mod N */
+static GEN
+_muliired(GEN x, GEN y, GEN N) { return resii(mulii(x,y), N); }
+static GEN
+_muliimontred(GEN x, GEN y, GEN N) { return montred(mulii(x,y), N); }
+static GEN
+_muliiinvred(GEN x, GEN y, GEN N) { return resiimul(mulii(x,y), N); }
 
-  if (typ(A) != t_INT || typ(N) != t_INT || typ(M) != t_INT) err(arither1);
-  s = signe(N);
+static GEN
+_mul(void *data, GEN x, GEN y)
+{
+  muldata *D = (muldata *)data;
+  return D->mulred(x,y,D->N);
+}
+static GEN
+_sqr(void *data, GEN x)
+{
+  muldata *D = (muldata *)data;
+  return D->res(sqri(x), D->N);
+}
+
+/* A^k mod N */
+GEN
+powmodulo(GEN A, GEN k, GEN N)
+{
+  ulong av = avma;
+  long t,s, lN;
+  int base_is_2, use_montgomery;
+  GEN y;
+  muldata  D;
+  montdata S;
+
+  if (typ(A) != t_INT || typ(k) != t_INT || typ(N) != t_INT) err(arither1);
+  s = signe(k);
   if (!s)
   {
-    k = signe(resii(A,M)); avma=av;
-    return k? gun: gzero;
+    t = signe(resii(A,N)); avma = av;
+    return t? gun: gzero;
   }
-  if (s < 0) { A = mpinvmod(A,M); N = absi(N); }
+  if (s < 0) y = mpinvmod(A,N);
   else
   {
-    A = modii(A,M);
-    if (!signe(A)) { avma=av; return gzero; }
+    y = modii(A,N);
+    if (!signe(y)) { avma = av; return gzero; }
   }
-  y = A;
-  if (lgefint(y)==3) switch(y[2])
+
+  base_is_2 = 0;
+  if (lgefint(y) == 3) switch(y[2])
   {
-    case 1: /* y = 1 */
-      avma=av; return gun;
-    case 2: /* y = 2, use shifti not mulii */
-    mul = (GEN (*)(GEN,GEN))shifti; A = (GEN)1L;
+    case 1: avma = av; return gun;
+    case 2: base_is_2 = 1; break;
   }
 
   /* TODO: Move this out of here and use for general modular computations */
-  if ((k = vali(M)) && k == expi(M))
-  { /* M is a power of 2 */
-    M = (GEN)k;
-    res = (GEN(*)(GEN,GEN))resmod2n;
-  }
-  else if (lgefint(M) > RESIIMUL_LIMIT && (lgefint(N) > 3 || N[2] > 4))
-  { /* compute x % M using multiplication by 1./M */
-    M = init_remainder(M);
-    res = (GEN(*)(GEN,GEN))resiimul;
-  }
-
-  p = N+2; man = *p; /* see puissii */
-  k = 1+bfffo(man); man<<=k; k = BITS_IN_LONG-k;
-  av1=avma; lim=stack_lim(av1,1);
-  for (nb=lgefint(N)-2;;)
+  lN = lgefint(N);
+  use_montgomery = mod2(N) && lN < MONTGOMERY_LIMIT;
+  if (use_montgomery)
   {
-    for (; k; man<<=1,k--)
-    {
-      y = res(sqri(y), M);
-      if (man < 0) y = res(mul(y,A), M);
-      if (low_stack(lim, stack_lim(av1,1)))
-      {
-        if (DEBUGMEM>1) err(warnmem,"powmodulo");
-        y = gerepileuptoint(av1,y);
-      }
-    }
-    if (--nb == 0) break;
-    man = *++p, k = BITS_IN_LONG;
+    init_montdata(N, &S);
+    y = resii(shifti(y, bit_accuracy(lN)), N);
+    D.mulred = base_is_2? &_muli2montred: &_muliimontred;
+    D.res = &montred;
+    D.N = (GEN)&S;
   }
-  return gerepileupto(av,y);
+  else if (lN > RESIIMUL_LIMIT
+       && (lgefint(k) > 3 || (((double)k[2])*expi(A)) > 2 + expi(N)))
+  {
+    D.mulred = base_is_2? &_muli2invred: &_muliiinvred;
+    D.res = &resiimul;
+    D.N = init_remainder(N);
+  }
+  else
+  {
+    D.mulred = base_is_2? &_muli2red: &_muliired;
+    D.res = &_resii;
+    D.N = N;
+  }
+  y = leftright_pow(y, k, (void*)&D, &_sqr, &_mul);
+  if (use_montgomery)
+  {
+    y = montred(y, (GEN)&S);
+    if (cmpii(y,N) >= 0) y = subii(y,N);
+  }
+  return gerepileuptoint(av,y);
 }
 
 /*********************************************************************/
@@ -1978,8 +2050,8 @@ bestappr(GEN x, GEN k)
       }
 
     case t_REAL:
-      p1=gun; p0=gfloor(x); q1=gzero; q0=gun; a=p0;
-      while (gcmp(q0,k)<=0)
+      p1=gun; a=p0=gfloor(x); q1=gzero; q0=gun;
+      while (cmpii(q0,k)<=0)
       {
 	x = gsub(x,a);
 	if (gcmp0(x)) { p1=p0; q1=q0; break; }
@@ -2290,7 +2362,11 @@ two_rank(GEN x)
 }
 
 #define MAXFORM 11
-#define _low(x) ({GEN __x=(GEN)x; modBIL(__x);})
+#if 0 /* gcc-ism */
+#  define _low(x) ({GEN __x=(GEN)x; modBIL(__x);})
+#else
+#  define _low(x) (__x = (GEN)(x), modBIL(__x))
+#endif
 
 /* h(x) for x<0 using Baby Step/Giant Step.
  * Assumes G is not too far from being cyclic.
@@ -2304,6 +2380,7 @@ classno(GEN x)
   GEN a,b,count,index,tabla,tablb,hash,p1,p2,hin,h,f,fh,fg,ftest;
   GEN Hf,D,fa;
   byteptr p = diffptr;
+  GEN __x;
 
   if (typ(x) != t_INT) err(arither1);
   s=signe(x); if (s>=0) return classno2(x);
