@@ -1,4 +1,4 @@
-/* $Id: arith2.c,v 1.30 2001/10/16 16:33:42 karim Exp $
+/* $Id: arith2.c,v 1.34 2001/12/05 20:48:01 karim Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -236,7 +236,7 @@ init_tinyprimes_tridiv(byteptr p);	/* in ifactor2.c */
 static ulong _maxprime = 0;
 
 ulong
-maxprime() { return _maxprime; }
+maxprime(void) { return _maxprime; }
 
 byteptr
 initprimes(long maxnum)
@@ -1555,14 +1555,13 @@ rhoreal_aux(GEN x, GEN D, GEN sqrtD, GEN isqrtD)
   GEN p1,p2, y = cgetg(6,t_VEC);
   GEN b = (GEN)x[2];
   GEN c = (GEN)x[3];
-  long s = signe(c);
 
   y[1] = (long)c;
-  p2 = (absi_cmp(isqrtD,c) >= 0)? isqrtD: c;
-  setsigne(c,1);
+  p2 = (absi_cmp(isqrtD,c) >= 0)? isqrtD: absi(c);
   p1 = shifti(c,1);
+  if (p1 == gzero) err(talker, "reducible form in rhoreal");
+  setsigne(p1,1); /* |2c| */
   p2 = divii(addii(p2,b), p1);
-  setsigne(c,s);
   y[2] = lsubii(mulii(p2,p1), b);
 
   p1 = shifti(subii(sqri((GEN)y[2]),D),-2);
@@ -2009,8 +2008,8 @@ bittest_many(GEN x, GEN gn, long c)
   }
   skip = (b1 < b2);			/* Skip the first word of the shift */
   l_res = l2 - l1 + 1 + 2 - skip + extra_words;	/* A coarse estimate */
-  p = new_chunk(l_res) + 2 + extra_words;
-  shift_r(p - skip, x + l1, x + l2 + 1, 0, b2);
+  p = (ulong*) (new_chunk(l_res) + 2 + extra_words);
+  shift_r(p - skip, (ulong*)x + l1, (ulong*)x + l2 + 1, 0, b2);
   if (two_adic) {			/* Check the low bits of x */
 	int i = lx, nonzero = 0;
 
@@ -2047,7 +2046,7 @@ bittest_many(GEN x, GEN gn, long c)
   if (pnew >= p - 2 + l_res)
       return gzero;
   l_res -= (pnew - p);
-  res = pnew - 2;
+  res = (GEN)(pnew - 2);
   res[1] = evalsigne(1)|evallgefint(l_res);
   res[0] = evaltyp(t_INT)|evallg(l_res);
   return res;
