@@ -1,4 +1,4 @@
-/* $Id: base3.c,v 1.42 2001/10/13 13:27:19 karim Exp $
+/* $Id: base3.c,v 1.47 2001/12/06 00:31:50 karim Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -39,6 +39,7 @@ isnfscalar(GEN x)
 {
   long lx=lg(x),i;
 
+  if (typ(x) != t_COL) return 0;
   for (i=2; i<lx; i++)
     if (!gcmp0((GEN)x[i])) return 0;
   return 1;
@@ -81,6 +82,7 @@ element_mul(GEN nf, GEN x, GEN y)
   if (ty==t_POLMOD) y=checknfelt_mod(nf,y,"element_mul");
   if (is_extscalar_t(tx)) return scal_mul(nf,x,y,ty);
   if (is_extscalar_t(ty)) return scal_mul(nf,y,x,tx);
+  if (tx != t_COL || ty != t_COL) err(typeer,"element_mul");
   if (isnfscalar(x)) return gmul((GEN)x[1],y);
   if (isnfscalar(y)) return gmul((GEN)y[1],x);
 
@@ -138,6 +140,7 @@ element_inv(GEN nf, GEN x)
     for (i=2; i<=N; i++) p1[i]=lcopy((GEN)x[i]);
     return p1;
   }
+  if (tx != t_COL) err(typeer,"element_inv");
   p = NULL;
   for (i=1; i<=N; i++)
     if (typ(x[i])==t_INTMOD)
@@ -182,6 +185,7 @@ element_div(GEN nf, GEN x, GEN y)
     p1=gdiv(gmodulcp(gmul((GEN)nf[7],x),(GEN)nf[1]),y);
     return gerepileupto(av, algtobasis(nf,p1));
   }
+  if (tx != t_COL || ty != t_COL) err(typeer,"element_div");
 
   if (isnfscalar(y)) return gdiv(x,(GEN)y[1]);
   if (isnfscalar(x))
@@ -307,6 +311,7 @@ element_sqr(GEN nf, GEN x)
   if (tx==t_POLMOD) x=checknfelt_mod(nf,x,"element_sqr");
   if (is_extscalar_t(tx))
     return gerepileupto(av, algtobasis(nf, gsqr(x)));
+  if (tx != t_COL) err(typeer,"element_sqr");
   if (isnfscalar(x))
   {
     s=cgetg(N+1,t_COL); s[1]=lsqr((GEN)x[1]);
@@ -345,50 +350,73 @@ element_sqr(GEN nf, GEN x)
   return v;
 }
 
+static GEN
+_mul(void *data, GEN x, GEN y) { return element_mul((GEN)data,x,y); }
+static GEN
+_sqr(void *data, GEN x) { return element_sqr((GEN)data,x); }
+
 /* Compute x^n in nf, left-shift binary powering */
 GEN
 element_pow(GEN nf, GEN x, GEN n)
 {
   ulong av = avma;
-  long s,N,i,j,m;
-  GEN y,p1,cx;
+  long s,N;
+  GEN y, cx;
 
   if (typ(n)!=t_INT) err(talker,"not an integer exponent in nfpow");
   nf=checknf(nf); N=degpol(nf[1]);
   s=signe(n); if (!s) return gscalcol_i(gun,N);
-  if (typ(x)!=t_COL) x=algtobasis(nf,x);
+  if (typ(x) != t_COL)
+  {
+    x = algtobasis(nf,x);
+    if (typ(x) != t_COL) err(typeer,"element_pow");
+  }
 
   if (isnfscalar(x))
   {
     y = gscalcol_i(gun,N);
     y[1] = (long)powgi((GEN)x[1],n); return y;
   }
-  cx = content(x);
-  if (gcmp1(cx)) cx = NULL; else x = gdiv(x,cx);
-  p1 = n+2; m = *p1;
-  y=x; j=1+bfffo(m); m<<=j; j = BITS_IN_LONG-j;
-  for (i=lgefint(n)-2;;)
-  {
-    for (; j; m<<=1,j--)
-    {
-      y = element_sqr(nf, y);
-      if (m<0) y=element_mul(nf, y, x);
-    }
-    if (--i == 0) break;
-    m = *++p1, j = BITS_IN_LONG;
-  }
+  x = primitive_part(x, &cx);
+  y = leftright_pow(x, n, (void*)nf, _sqr, _mul);
   if (s<0) y = element_inv(nf, y);
   if (cx) y = gmul(y, powgi(cx, n));
   return av==avma? gcopy(y): gerepileupto(av,y);
 }
 
+struct muldata {
+  GEN nf, p;
+  long I;
+};
+
+static GEN
+_mulidmod(void *data, GEN x, GEN y)
+{
+  struct muldata *D = (struct muldata*)data;
+  (void)y; /* ignored */
+  return FpV_red(element_mulid(D->nf, x, D->I), D->p);
+}
+static GEN
+_mulmod(void *data, GEN x, GEN y)
+{
+  struct muldata *D = (struct muldata*)data;
+  return FpV_red(element_muli(D->nf, x, y), D->p);
+}
+static GEN
+_sqrmod(void *data, GEN x)
+{
+  struct muldata *D = (struct muldata*)data;
+  return FpV_red(element_sqri(D->nf, x), D->p);
+}
+
 /* x in Z^n, compute lift(x^n mod p) */
 GEN
 element_pow_mod_p(GEN nf, GEN x, GEN n, GEN p)
 {
   ulong av = avma;
-  long s,N,i,j,m;
-  GEN y,p1;
+  struct muldata D;
+  long s,N;
+  GEN y;
 
   if (typ(n)!=t_INT) err(talker,"not an integer exponent in nfpow");
   nf=checknf(nf); N=degpol(nf[1]);
@@ -400,20 +428,10 @@ element_pow_mod_p(GEN nf, GEN x, GEN n, GEN p)
     y = gscalcol_i(gun,N);
     y[1] = (long)powmodulo((GEN)x[1],n,p); return y;
   }
-  p1 = n+2; m = *p1;
-  y=x; j=1+bfffo(m); m<<=j; j = BITS_IN_LONG-j;
-  for (i=lgefint(n)-2;;)
-  {
-    for (; j; m<<=1,j--)
-    {
-      y = element_sqri(nf, y);
-      if (m<0) y=element_muli(nf, y, x);
-      y = FpV_red(y, p);
-    }
-    if (--i == 0) break;
-    m = *++p1, j = BITS_IN_LONG;
-  }
-  if (s<0)  y = FpV_red(element_inv(nf,y), p);
+  D.nf = nf;
+  D.p = p;
+  y = leftright_pow(x,n, (void*)&D, &_sqrmod, &_mulmod);
+  if (s < 0)  y = FpV_red(element_inv(nf,y), p);
   return av==avma? gcopy(y): gerepileupto(av,y);
 }
 
@@ -422,28 +440,20 @@ GEN
 element_powid_mod_p(GEN nf, long I, GEN n, GEN p)
 {
   ulong av = avma;
-  long s,N,i,j,m;
-  GEN y,p1;
+  struct muldata D;
+  long s,N;
+  GEN y;
 
   if (typ(n)!=t_INT) err(talker,"not an integer exponent in nfpow");
   nf=checknf(nf); N=degpol(nf[1]);
   s=signe(n);
   if (!s || I == 1) return gscalcol_i(gun,N);
-  p1 = n+2; m = *p1;
   y = zerocol(N); y[I] = un;
-  j=1+bfffo(m); m<<=j; j = BITS_IN_LONG-j;
-  for (i=lgefint(n)-2;;)
-  {
-    for (; j; m<<=1,j--)
-    {
-      y = element_sqri(nf, y);
-      if (m<0) y=element_mulid(nf, y, I);
-      y = FpV_red(y, p);
-    }
-    if (--i == 0) break;
-    m = *++p1, j = BITS_IN_LONG;
-  }
-  if (s<0)  y = FpV_red(element_inv(nf,y), p);
+  D.nf = nf;
+  D.p = p;
+  D.I = I;
+  y = leftright_pow(y,n, (void*)&D, &_sqrmod, &_mulidmod);
+  if (s < 0) y = FpV_red(element_inv(nf,y), p);
   return av==avma? gcopy(y): gerepileupto(av,y);
 }
 
@@ -759,6 +769,8 @@ zsigne(GEN nf,GEN x,GEN arch)
     if (signe(arch[i]))
     {
       GEN y = poleval(x,(GEN)rac[i]);
+      if (lg(y) == 3 && typ(y) == t_REAL)
+        err(warner,"dubious value in zsigne. Increase nf precision?");
       if (e + gexpo((GEN)rac[i]) - gexpo(y) > B)
         err(precer, "zsigne");
       vecsign[j++] = (gsigne(y) > 0)? (long)_0: (long)_1;
@@ -902,12 +914,11 @@ element_powmodidele(GEN nf,GEN x,GEN k,GEN idele,GEN sarch)
 GEN
 smithrel(GEN H, GEN *newU, GEN *newUi)
 { 
-  GEN z,U,Ui,D,p1;
+  GEN U,Ui,D,p1;
   long l,c;
 
-  z = smith2(H);
-  U = (GEN)z[1];
-  D = (GEN)z[3]; l = lg(D);
+  D = smithall(H, &U, NULL);
+  l = lg(D);
   for (c=1; c<l; c++)
   {
     p1 = gcoeff(D,c,c);
