@@ -1,4 +1,4 @@
-/* $Id: base4.c,v 1.71 2001/10/13 13:27:19 karim Exp $
+/* $Id: base4.c,v 1.85 2002/01/10 22:09:27 karim Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -113,6 +113,7 @@ prime_to_ideal_aux(GEN nf, GEN vp)
   return hnfmodid(m,(GEN)vp[1]);
 }
 
+/* vp = [a,x,...], a in Z. Return (a,x)  [HACK: vp need not be prime] */
 GEN
 prime_to_ideal(GEN nf, GEN vp)
 {
@@ -611,8 +612,8 @@ idealfactor(GEN nf, GEN x)
   if (tx == id_PRIME)
   {
     y=cgetg(3,t_MAT);
-    y[1]=lgetg(2,t_COL); mael(y,1,1)=lcopy(x);
-    y[2]=lgetg(2,t_COL); mael(y,2,1)=un; return y;
+    y[1]=(long)_col(gcopy(x));
+    y[2]=(long)_col(gun); return y;
   }
   nf=checknf(nf); av=avma;
   if (tx == id_PRINCIPAL) x = principalideal_aux(nf,x);
@@ -1053,7 +1054,7 @@ idealaddmultoone(GEN nf, GEN list)
 /* multiplication */
 
 /* x integral ideal (without archimedean component) in HNF form
- * [a,alpha,n] corresponds to the ideal aZ_K+alpha Z_K (a is a
+ * [a,alpha] corresponds to the ideal aZ_K+alpha Z_K (a is a
  * rational integer). Multiply them
  */
 static GEN
@@ -1074,8 +1075,7 @@ idealmulspec(GEN nf, GEN x, GEN a, GEN alpha)
 /* x ideal (matrix form,maximal rank), vp prime ideal (primedec). Output the
  * product. Can be used for arbitrary vp of the form [p,a,e,f,b], IF vp
  * =pZ_K+aZ_K, p is an integer, and norm(vp) = p^f; e and b are not used.
- * For internal use.
- */
+ * For internal use. */
 GEN
 idealmulprime(GEN nf, GEN x, GEN vp)
 {
@@ -1086,6 +1086,10 @@ idealmulprime(GEN nf, GEN x, GEN vp)
   return denx? gdiv(x,denx): x;
 }
 
+GEN arch_mul(GEN x, GEN y);
+GEN vecpow(GEN x, GEN n);
+GEN vecmul(GEN x, GEN y);
+
 /* Assume ix and iy are integral in HNF form (or ideles of the same form).
  * HACK: ideal in iy can be of the form [a,b], a in Z, b in Z_K
  * For internal use. */
@@ -1104,7 +1108,7 @@ idealmulh(GEN nf, GEN ix, GEN iy)
   if (!f) return y;
 
   res[1]=(long)y;
-  if (f==3) y = gadd((GEN)ix[2],(GEN)iy[2]);
+  if (f==3) y = arch_mul((GEN)ix[2], (GEN)iy[2]);
   else
   {
     y = (f==2)? (GEN)iy[2]: (GEN)ix[2];
@@ -1327,12 +1331,11 @@ static GEN
 famat_makecoprime(GEN nf, GEN g, GEN e, GEN pr, GEN prn, GEN EX)
 {
   long i,k, l = lg(g), N = degpol(nf[1]);
-  GEN prnZ,cx,x,u,z, zpow = gzero, p = (GEN)pr[1], b = (GEN)pr[5];
+  GEN prnZ,cx,x,u, zpow = gzero, p = (GEN)pr[1], b = (GEN)pr[5];
   GEN mul = cgetg(N+1,t_MAT);
   GEN newg = cgetg(l+1, t_VEC); /* room for z */
 
   prnZ = gcoeff(prn, 1,1);
-  z = gmod(special_anti_uniformizer(nf, pr), prnZ);
   for (i=1; i<=N; i++) mul[i] = (long)element_mulid(nf,b,i);
   for (i=1; i < l; i++)
   {
@@ -1350,7 +1353,7 @@ famat_makecoprime(GEN nf, GEN g, GEN e, GEN pr, GEN prn, GEN EX)
   if (zpow == gzero) setlg(newg, l);
   else
   {
-    newg[i] = (long)z;
+    newg[i] = (long)FpV_red(special_anti_uniformizer(nf, pr), prnZ);
     e = concatsp(e, negi(zpow));
   }
   e = gmod(e, EX);
@@ -1406,6 +1409,28 @@ famat_to_nf_modidele(GEN nf, GEN g, GEN e, GEN bid)
   return set_sign_mod_idele(nf, to_famat(g,e), t, module, sarch);
 }
 
+GEN
+famat_to_arch(GEN nf, GEN fa, long prec)
+{
+  GEN g,e, y = NULL;
+  long i,l;
+
+  if (lg(fa) == 1) return zerovec(lg(nf[6])-1);
+  g = (GEN)fa[1]; 
+  e = (GEN)fa[2]; l = lg(e);
+  for (i=1; i<l; i++)
+  {
+    GEN t, x = (GEN)g[i];
+    if (typ(x) != t_COL) x = algtobasis(nf,x);
+    x = primpart(x);
+    /* multiplicative arch would be better (save logs), but exponents overflow
+     * [ could keep track of expo separately, but not worth it ] */
+    t = gmul(get_arch(nf,x,prec), (GEN)e[i]);
+    y = y? gadd(y,t): t;
+  }
+  return y;
+}
+
 GEN
 vecmul(GEN x, GEN y)
 {
@@ -1428,6 +1453,17 @@ vecinv(GEN x)
   return z;
 }
 
+GEN
+vecpow(GEN x, GEN n)
+{
+  long i,lx, tx = typ(x);
+  GEN z;
+  if (is_scalar_t(tx)) return powgi(x,n);
+  lx = lg(x); z = cgetg(lx, tx);
+  for (i=1; i<lx; i++) z[i] = (long)vecpow((GEN)x[i], n);
+  return z;
+}
+
 GEN
 vecdiv(GEN x, GEN y) { return vecmul(x, vecinv(y)); }
 
@@ -1450,8 +1486,9 @@ GEN
 arch_inv(GEN x) {
   switch (typ(x)) {
     case t_POLMOD: return ginv(x);
+    case t_COL:    return vecinv(x);
     case t_MAT:    return famat_inv(x);
-    default:       return gneg(x); /* t_COL, t_VEC */
+    default:       return gneg(x); /* t_VEC */
   }
 }
 
@@ -1459,6 +1496,7 @@ GEN
 arch_pow(GEN x, GEN n) {
   switch (typ(x)) {
     case t_POLMOD: return powgi(x,n);
+    case t_COL:    return vecpow(x,n);
     case t_MAT:    return famat_pow(x,n);
     default:       return gmul(n,x);
   }
@@ -1617,18 +1655,31 @@ idealpowprime_spec(GEN nf, GEN vp, GEN n, GEN *d)
   if (s < 0) n = negi(n);
   /* now n > 0 */
   x = dummycopy(vp);
-  n1 = dvmdii(n, (GEN)x[3], &r);
-  if (signe(r)) n1 = addis(n1,1); /* n1 = ceil(n/e) */
-  x[1] = (long)powgi((GEN)x[1],n1);
-  if (s < 0)
+  if (is_pm1(n)) /* n = 1 special cased for efficiency */
   {
-    x[2] = ldiv(element_pow(nf,(GEN)x[5],n), powgi((GEN)vp[1],subii(n,n1)));
-    *d = (GEN)x[1];
+    if (s < 0)
+    {
+      x[2] = x[5];
+      *d = (GEN)x[1];
+    }
+    else
+      *d = NULL;
   }
   else
   {
-    x[2] = (long)element_pow(nf,(GEN)x[2],n);
-    *d = NULL;
+    n1 = dvmdii(n, (GEN)x[3], &r);
+    if (signe(r)) n1 = addis(n1,1); /* n1 = ceil(n/e) */
+    x[1] = (long)powgi((GEN)x[1],n1);
+    if (s < 0)
+    {
+      x[2] = ldiv(element_pow(nf,(GEN)x[5],n), powgi((GEN)vp[1],subii(n,n1)));
+      *d = (GEN)x[1];
+    }
+    else
+    {
+      x[2] = (long)element_pow(nf,(GEN)x[2],n);
+      *d = NULL;
+    }
   }
   return x;
 }
@@ -1647,7 +1698,7 @@ idealpowprime(GEN nf, GEN vp, GEN n)
   return x;
 }
 
-/* x * vp^n */
+/* x * vp^n. Assume x in HNF (possibly non-integral) */
 GEN
 idealmulpowprime(GEN nf, GEN x, GEN vp, GEN n)
 {
@@ -1655,6 +1706,11 @@ idealmulpowprime(GEN nf, GEN x, GEN vp, GEN n)
 
   if (!signe(n)) return x;
   nf = checknf(nf);
+  
+  /* inert, special cased for efficiency */
+  if (itos((GEN)vp[4]) == degpol(nf[1]))
+    return gmul(x, powgi((GEN)vp[1], n));
+
   y = idealpowprime_spec(nf, vp, n, &d);
   denx = denom(x);
   if (gcmp1(denx)) denx = d; else
@@ -1665,6 +1721,11 @@ idealmulpowprime(GEN nf, GEN x, GEN vp, GEN n)
   x = idealmulspec(nf,x, (GEN)y[1], (GEN)y[2]);
   return denx? gdiv(x,denx): x;
 }
+GEN
+idealdivpowprime(GEN nf, GEN x, GEN vp, GEN n)
+{
+  return idealmulpowprime(nf,x,vp, negi(n));
+}
 
 /* raise the ideal x to the power n (in Z) */
 GEN
@@ -1730,39 +1791,53 @@ init_idele(GEN nf)
   x[2] = (long)zerovec(RU); return x;
 }
 
+static GEN
+_idealmulred(GEN nf, GEN x, GEN y, long prec)
+{
+  return ideallllred(nf,idealmul(nf,x,y), NULL, prec);
+}
+
+struct muldata {
+  GEN nf;
+  long prec;
+};
+
+static GEN
+_mul(void *data, GEN x, GEN y)
+{
+  struct muldata *D = (struct muldata *)data;
+  return _idealmulred(D->nf,x,y,D->prec);
+}
+static GEN
+_sqr(void *data, GEN x)
+{
+  return _mul(data,x,x);
+}
+
 /* compute x^n (x ideal, n integer), reducing along the way */
 GEN
 idealpowred(GEN nf, GEN x, GEN n, long prec)
 {
-  long i,j,m,av=avma, s = signe(n);
-  GEN y, p1;
+  long av=avma, s = signe(n);
+  struct muldata D;
+  GEN y;
 
   if (typ(n) != t_INT) err(talker,"non-integral exponent in idealpowred");
-  if (signe(n) == 0) return idealpow(nf,x,n);
-  p1 = n+2; m = *p1;
-  y = x; j=1+bfffo(m); m<<=j; j = BITS_IN_LONG-j;
-  for (i=lgefint(n)-2;;)
-  {
-    for (; j; m<<=1,j--)
-    {
-      y = idealmul(nf,y,y);
-      if (m < 0) y = idealmul(nf,y,x);
-      y = ideallllred(nf,y,NULL,prec);
-    }
-    if (--i == 0) break;
-    m = *++p1; j = BITS_IN_LONG;
-  }
+  if (s == 0) return idealpow(nf,x,n);
+  D.nf  = nf;
+  D.prec= prec;
+  y = leftright_pow(x, n, (void*)&D, &_sqr, &_mul);
+ 
   if (s < 0) y = idealinv(nf,y);
-  if (y == x) y = ideallllred(nf,x,NULL,prec);
+  else if (is_pm1(n)) y = ideallllred(nf,x,NULL,prec);
   return gerepileupto(av,y);
 }
 
 GEN
 idealmulred(GEN nf, GEN x, GEN y, long prec)
 {
-  long av=avma;
-  x = idealmul(nf,x,y);
-  return gerepileupto(av, ideallllred(nf,x,NULL,prec));
+  ulong av = avma;
+  return gerepileupto(av, _idealmulred(nf,x,y,prec));
 }
 
 long
@@ -2036,10 +2111,10 @@ minideal(GEN nf, GEN x, GEN vdir, long prec)
   return gerepileupto(av, principalidele(nf,y,prec));
 }
 static GEN
-appr_reduce(GEN s, GEN y, long N)
+appr_reduce(GEN s, GEN y)
 {
+  long i, N = lg(y)-1;
   GEN p1,u,z = cgetg(N+2,t_MAT);
-  long i;
 
   s=gmod(s,gcoeff(y,1,1)); y=gmul(y,lllint(y));
   for (i=1; i<=N; i++) z[i]=y[i]; z[N+1]=(long)s;
@@ -2050,170 +2125,150 @@ appr_reduce(GEN s, GEN y, long N)
   return gadd(s, gmul(y,u));
 }
 
-/* Given a fractional ideal x (if fl=0) or a prime ideal factorization
- * with possibly zero or negative exponents (if fl=1), gives a b such that
- * v_p(b)=v_p(x) for all prime ideals p dividing x (or in the factorization)
- * and v_p(b)>=0 for all other p, using the (standard) proof given in GTM 138.
- * Certainly not the most efficient, but sure.
- */
+/* all ep[i] >= 0, y  is  prod P[i]^(ep[i]+1)  [or NULL in which case it is
+ * computed]. Return s integral such that v_P[i](s) = ep[i] for all i */
+static GEN
+appr_pos(GEN nf, GEN y, GEN list, GEN ep)
+{
+  long i, r = lg(list);
+  GEN pr,v,s,t, z = cgetg(r,t_VEC);
+
+  if (!y)
+  {
+    y = idmat(degpol(nf[1]));
+    for (i=1; i<r; i++)
+      y = idealmulpowprime(nf,y,(GEN)list[i], addsi(1,(GEN)ep[i]));
+  }
+
+  for (i=1; i<r; i++)
+    z[i] = (long)idealdivpowprime(nf,y,(GEN)list[i], addsi(1,(GEN)ep[i]));
+  v = idealaddmultoone(nf,z); s = NULL;
+  for (i=1; i<r; i++)
+  {
+    pr = (GEN)list[i];
+    t = element_mul(nf,(GEN)v[i], element_pow(nf,(GEN)pr[2],(GEN)ep[i]));
+    s = s? gadd(s,t): t;
+  }
+  s = appr_reduce(s,y);
+  if (DEBUGLEVEL>2) fprintferr(" leaving idealappr_pos = %Z\n",s);
+  return s;
+}
+
+/* Given a fractional ideal x (if fl=0) or a prime ideal factorization with
+ * possibly zero or negative exponents (if fl=1), gives b such that
+ * v_p(b)=v_p(x) for all prime ideals p | x and v_p(b)>=0 for all other p,
+ * using the standard proof given in GTM 138. Certainly not the most
+ * efficient, but sure. */
 GEN
 idealappr0(GEN nf, GEN x, long fl)
 {
-  long av=avma,tetpil,i,j,k,l,N,r,r2;
-  GEN fact,fact2,list,ep,ep1,ep2,y,z,v,p1,p2,p3,p4,s,pr,alpha,beta,den;
+  long av=avma,i,j,k,l,N,r,r2;
+  GEN fact,fact2,list,ep,ep1,ep2,y,z,p1,p3,alpha,beta,den;
 
   if (DEBUGLEVEL>4)
   {
-    fprintferr(" entree dans idealappr0() :\n");
-    fprintferr(" x = "); outerr(x);
+    fprintferr(" entering idealappr0() :\n");
+    fprintferr(" x = %Z\n", x);
   }
-  nf=checknf(nf); N=degpol(nf[1]);
+  nf = checknf(nf); N = degpol(nf[1]);
   if (fl)
   {
     if (typ(x)!=t_MAT || lg(x)!=3)
       err(talker,"not a prime ideal factorization in idealappr0");
-    fact=x; list=(GEN)fact[1]; ep=(GEN)fact[2]; r=lg(list);
+    fact=x;
+    list= (GEN)fact[1];
+    ep  = (GEN)fact[2]; r = lg(ep);
     if (r==1) return gscalcol_i(gun,N);
     for (i=1; i<r; i++)
       if (signe(ep[i]) < 0) break;
     if (i < r)
     {
       ep1=cgetg(r,t_COL);
-      for (i=1; i<r; i++)
-        ep1[i] = (signe(ep[i])>=0)? zero: lnegi((GEN)ep[i]);
-      fact[2]=(long)ep1; beta=idealappr0(nf,fact,1);
-      fact2=idealfactor(nf,beta);
+      for (i=1; i<r; i++) ep1[i] = (signe(ep[i])>=0)? zero: lnegi((GEN)ep[i]);
+      beta = appr_pos(nf,NULL,list,ep1);
+      fact2= idealfactor(nf,beta);
       p1=(GEN)fact2[1]; r2=lg(p1);
       ep2=(GEN)fact2[2]; l=r+r2-1;
       z=cgetg(l,t_VEC); for (i=1; i<r; i++) z[i]=list[i];
       ep1=cgetg(l,t_VEC);
-      for (i=1; i<r; i++)
-        ep1[i] = (signe(ep[i])<=0)? zero: licopy((GEN)ep[i]);
-      j=r-1;
+      for (i=1; i<r; i++) ep1[i] = (signe(ep[i])<=0)? zero: ep[i];
+      j = r;
       for (i=1; i<r2; i++)
       {
         p3=(GEN)p1[i]; k=1;
         while (k<r &&
           (    !gegal((GEN)p3[1],gmael(list,k,1))
             || !element_val(nf,(GEN)p3[2],(GEN)list[k]) )) k++;
-        if (k==r) { j++; z[j]=(long)p3; ep1[j]=ep2[i]; }
+        if (k==r) { z[j]=(long)p3; ep1[j]=ep2[i]; j++; }
       }
-      fact=cgetg(3,t_MAT);
-      fact[1]=(long)z; setlg(z,j+1);
-      fact[2]=(long)ep1; setlg(ep1,j+1);
-      alpha=idealappr0(nf,fact,1); tetpil=avma;
-      if (DEBUGLEVEL>2)
-      {
-        fprintferr(" alpha = "); outerr(alpha);
-        fprintferr(" beta = "); outerr(beta);
-      }
-      return gerepile(av,tetpil,element_div(nf,alpha,beta));
+
+      setlg(z,j); setlg(ep1,j);
+      alpha = appr_pos(nf,NULL,z,ep1);
+      if (DEBUGLEVEL>2) fprintferr(" alpha = %Z\nbeta = %Z\n", alpha,beta);
+      return gerepileupto(av, element_div(nf,alpha,beta));
     }	
-    y=idmat(N);
-    for (i=1; i<r; i++)
-    {
-      pr=(GEN)list[i];
-      if (signe(ep[i]))
-      {
-        p4=addsi(1,(GEN)ep[i]); p1=powgi((GEN)pr[1],p4);
-	if (cmpis((GEN)pr[4],N))
-	{
-	  p2=cgetg(3,t_MAT);
-          p2[1]=(long)gscalcol_i(p1, N);
-	  p2[2]=(long)element_pow(nf,(GEN)pr[2],p4);
-          y=idealmat_mul(nf,y,p2);
-	}
-	else y=gmul(p1,y);
-      }
-      else y=idealmulprime(nf,y,pr);
-    }
+    y = NULL;
   }
   else
   {
     den=denom(x); if (gcmp1(den)) den=NULL; else x=gmul(den,x);
     x=idealhermite_aux(nf,x);
     fact=idealfactor(nf,x);
-    list=(GEN)fact[1]; ep=(GEN)fact[2]; r=lg(list);
+    list= (GEN)fact[1];
+    ep  = (GEN)fact[2]; r = lg(ep);
     if (r==1) { avma=av; return gscalcol_i(gun,N); }
     if (den)
     {
       fact2=idealfactor(nf,den);
       p1=(GEN)fact2[1]; r2=lg(p1);
-      l=r+r2-1;
-      z=cgetg(l,t_COL);   for (i=1; i<r; i++) z[i]=list[i];
-      ep1=cgetg(l,t_COL); for (i=1; i<r; i++) ep1[i]=ep[i];
-      j=r-1;
+      l = r+r2-1;
+      z   = cgetg(l,t_COL); for (i=1; i<r; i++) z[i]=list[i];
+      ep1 = cgetg(l,t_COL); for (i=1; i<r; i++) ep1[i]=ep[i];
+      j = r;
       for (i=1; i<r2; i++)
       {
 	p3=(GEN)p1[i]; k=1;
 	while (k<r && !gegal((GEN)list[k],p3)) k++;
-	if (k==r){ j++; z[j]=(long)p3; ep1[j]=zero; }
+	if (k==r) { z[j]=(long)p3; ep1[j]=zero; j++; }
       }
-      fact=cgetg(3,t_MAT);
-      fact[1]=(long)z; setlg(z,j+1);
-      fact[2]=(long)ep1; setlg(ep1,j+1);
-      alpha=idealappr0(nf,fact,1);
-      if (DEBUGLEVEL>2) { fprintferr(" alpha = "); outerr(alpha); }
-      tetpil=avma; return gerepile(av,tetpil,gdiv(alpha,den));
+      y = x; for (k=r; k<j; k++) y = idealmulprime(nf,y,(GEN)z[i]);
+      setlg(z, j); setlg(ep1, j);
+      alpha = appr_pos(nf,y,z,ep1);
+      if (DEBUGLEVEL>2) fprintferr(" alpha = %Z\n",alpha);
+      return gerepileupto(av, gdiv(alpha,den));
     }
-    y=x; for (i=1; i<r; i++) y=idealmulprime(nf,y,(GEN)list[i]);
+    y = x; for (i=1; i<r; i++) y = idealmulprime(nf,y,(GEN)list[i]);
   }
-
-  z=cgetg(r,t_VEC);
-  for (i=1; i<r; i++)
-  {
-    pr=(GEN)list[i]; p4=addsi(1, (GEN)ep[i]); p1=powgi((GEN)pr[1],p4);
-    if (cmpis((GEN)pr[4],N))
-    {
-      p2=cgetg(3,t_MAT);
-      p2[1]=(long)gscalcol_i(p1,N);
-      p2[2]=(long)element_pow(nf,(GEN)pr[5],p4);
-      z[i]=ldiv(idealmat_mul(nf,y,p2),p1);
-    }
-    else z[i]=ldiv(y,p1);
-  }
-  v=idealaddmultoone(nf,z);
-  s=cgetg(N+1,t_COL); for (i=1; i<=N; i++) s[i]=zero;
-  for (i=1; i<r; i++)
-  {
-    pr=(GEN)list[i];
-    if (signe(ep[i]))
-      s=gadd(s,element_mul(nf,(GEN)v[i],element_pow(nf,(GEN)pr[2],(GEN)ep[i])));
-    else s=gadd(s,(GEN)v[i]);
-  }
-  p3 = appr_reduce(s,y,N);
-  if (DEBUGLEVEL>2)
-    { fprintferr(" sortie de idealappr0 p3 = "); outerr(p3); }
-  return gerepileupto(av,p3);
+  return gerepileupto(av, appr_pos(nf,y,list,ep));
 }
 
 /* Given a prime ideal factorization x with possibly zero or negative exponents,
- * and a vector y of elements of nf, gives a b such that
- * v_p(b-y_p)>=v_p(x) for all prime ideals p in the ideal factorization
+ * and a vector w of elements of nf, gives a b such that
+ * v_p(b-w_p)>=v_p(x) for all prime ideals p in the ideal factorization
  * and v_p(b)>=0 for all other p, using the (standard) proof given in GTM 138.
  * Certainly not the most efficient, but sure.
  */
 GEN
-idealchinese(GEN nf, GEN x, GEN y)
+idealchinese(GEN nf, GEN x, GEN w)
 {
-  long ty=typ(y),av=avma,i,j,k,l,N,r,r2;
-  GEN fact,fact2,list,ep,ep1,ep2,z,t,v,p1,p2,p3,p4,s,pr,den;
+  long ty=typ(w),av=avma,i,j,k,l,N,r,r2;
+  GEN fact,fact2,list,ep,ep1,ep2,z,t,y,v,p1,p3,s,den;
 
   if (DEBUGLEVEL>4)
   {
     fprintferr(" entree dans idealchinese() :\n");
-    fprintferr(" x = "); outerr(x);
-    fprintferr(" y = "); outerr(y);
+    fprintferr(" x = %Z\n", x);
+    fprintferr(" w = %Z\n", w);
   }
   nf=checknf(nf); N=degpol(nf[1]);
   if (typ(x)!=t_MAT ||(lg(x)!=3))
     err(talker,"not a prime ideal factorization in idealchinese");
   fact=x; list=(GEN)fact[1]; ep=(GEN)fact[2]; r=lg(list);
-  if (!is_vec_t(ty) || lg(y)!=r)
+  if (!is_vec_t(ty) || lg(w)!=r)
     err(talker,"not a suitable vector of elements in idealchinese");
   if (r==1) return gscalcol_i(gun,N);
 
-  den=denom(y);
+  den=denom(w);
   if (!gcmp1(den))
   {
     fact2=idealfactor(nf,den);
@@ -2221,56 +2276,36 @@ idealchinese(GEN nf, GEN x, GEN y)
     ep2=(GEN)fact2[2]; l=r+r2-1;
     z=cgetg(l,t_VEC); for (i=1; i<r; i++) z[i]=list[i];
     ep1=cgetg(l,t_VEC); for (i=1; i<r; i++) ep1[i]=ep[i];
-    j=r-1;
+    j=r;
     for (i=1; i<r2; i++)
     {
       p3=(GEN)p1[i]; k=1;
       while (k<r && !gegal((GEN)list[k],p3)) k++;
-      if (k==r) { j++; z[j]=(long)p3; ep1[j]=ep2[i]; }
+      if (k==r) { z[j]=(long)p3; ep1[j]=ep2[i]; j++; }
       else ep1[k]=ladd((GEN)ep1[k],(GEN)ep2[i]);
     }
-    r=j+1; setlg(z,r); setlg(ep1,r); list=z; ep=ep1;
+    r=j; setlg(z,r); setlg(ep1,r); list=z; ep=ep1;
   }
   for (i=1; i<r; i++)
-    if (signe(ep[i])<0) ep[i] = zero;
-  t=idmat(N);
-  for (i=1; i<r; i++)
-  {
-    pr=(GEN)list[i]; p4=(GEN)ep[i];
-    if (signe(p4))
-    {
-      if (cmpis((GEN)pr[4],N))
-      {
-	p2=cgetg(3,t_MAT);
-        p2[1]=(long)gscalcol_i(powgi((GEN)pr[1],p4), N);
-	p2[2]=(long)element_pow(nf,(GEN)pr[2],p4);
-        t=idealmat_mul(nf,t,p2);
-      }
-      else t=gmul(powgi((GEN)pr[1],p4),t);
-    }
-  }
-  z=cgetg(r,t_VEC);
-  for (i=1; i<r; i++)
-  {
-    pr=(GEN)list[i]; p4=(GEN)ep[i];
-    if (cmpis((GEN)pr[4],N))
-    {
-      p2=cgetg(3,t_MAT); p1=powgi((GEN)pr[1],p4);
-      p2[1]=(long)gscalcol_i(p1,N);
-      p2[2]=(long)element_pow(nf,(GEN)pr[5],p4);
-      z[i]=ldiv(idealmat_mul(nf,t,p2),p1);
-    }
-    else z[i]=ldiv(t,powgi((GEN)pr[1],p4));
-  }
-  v=idealaddmultoone(nf,z);
-  s=cgetg(N+1,t_COL); for (i=1; i<=N; i++) s[i]=zero;
-  for (i=1; i<r; i++)
-    s = gadd(s,element_mul(nf,(GEN)v[i],(GEN)y[i]));
+    if (signe(ep[i]) < 0) ep[i] = zero;
 
-  p3 = appr_reduce(s,t,N);
-  if (DEBUGLEVEL>2)
-    { fprintferr(" sortie de idealchinese() : p3 = "); outerr(p3); }
-  return gerepileupto(av,p3);
+  y = idmat(N);
+  for (i=1; i<r; i++)
+    y = idealmulpowprime(nf,y, (GEN)list[i], (GEN)ep[i]);
+
+  z = cgetg(r,t_VEC);
+  for (i=1; i<r; i++)
+    z[i] = (long)idealdivpowprime(nf,y, (GEN)list[i], (GEN)ep[i]);
+  v = idealaddmultoone(nf,z); s = NULL;
+  for (i=1; i<r; i++)
+  {
+    t = element_mul(nf,(GEN)v[i], (GEN)w[i]);
+    s = s? gadd(s,t): t;
+  }
+
+  s = appr_reduce(s,y);
+  if (DEBUGLEVEL>2) fprintferr(" leaving idealchinese = %Z\n",s);
+  return gerepileupto(av,s);
 }
 
 GEN
@@ -2326,10 +2361,11 @@ idealcoprime(GEN nf, GEN x, GEN y)
     fprintferr(" x = "); outerr(x);
     fprintferr(" y = "); outerr(y);
   }
-  fact=idealfactor(nf,y); list=(GEN)fact[1];
-  r=lg(list); ep = (GEN)fact[2];
-  for (i=1; i<r; i++) ep[i]=lstoi(-idealval(nf,x,(GEN)list[i]));
-  tetpil=avma; p2=idealappr0(nf,fact,1);
+  fact=idealfactor(nf,y);
+  list=(GEN)fact[1];
+  ep = (GEN)fact[2]; r=lg(ep);
+  for (i=1; i<r; i++) ep[i] = lstoi(-idealval(nf,x,(GEN)list[i]));
+  tetpil=avma; p2 = idealappr0(nf,fact,1);
   if (DEBUGLEVEL>4)
     { fprintferr(" sortie de idealcoprime() : p2 = "); outerr(p2); }
   return gerepile(av,tetpil,p2);
