@@ -1,4 +1,4 @@
-/* $Id: base5.c,v 1.17 2001/10/01 12:11:29 karim Exp $
+/* $Id: base5.c,v 1.18 2001/12/06 17:30:28 karim Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -586,9 +586,9 @@ rnfidealreltoabs(GEN rnf,GEN x)
       p3[(i-1)*m+j]=(long)p5;
     }
   }
-  p1 = gmul(basinv,p3); c = content(p1);
+  p1 = gmul(basinv,p3);
+  p1 = primitive_part(p1, &c);
   p2 = gmael(p1,1,1); /* x \cap Z */
-  if (is_pm1(c)) c = NULL; else { p1 = gdiv(p1, c); p2 = gdiv(p2, c); }
   p1 = hnfmodid(p1, p2);
   if (c) p1 = gmul(p1, c);
   return gerepileupto(av, p1);
