@@ -1,4 +1,4 @@
-/* $Id: buch1.c,v 1.35 2001/10/12 11:22:43 karim Exp $
+/* $Id: buch1.c,v 1.37 2001/11/22 22:38:08 karim Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -1096,9 +1096,9 @@ desalloc(long **mat)
   }
 }
 
-/* L-function */
+/* ~ L(kro_D, 1) */
 static GEN
-lfunc(GEN Disc)
+Lval1(GEN Disc)
 {
   long av=avma, p;
   GEN y=realun(DEFAULTPREC);
@@ -1116,24 +1116,19 @@ lfunc(GEN Disc)
 static GEN
 get_clgp(GEN Disc, GEN W, GEN *ptmet, long prec)
 {
-  GEN p1,p2,res,*init, u1u2=smith2(W), u1=(GEN)u1u2[1], met=(GEN)u1u2[3];
-  long c,i,j, l = lg(met);
+  GEN res,*init, u1, met = smithrel(W,NULL,&u1);
+  long c,i,j, l = lg(W);
 
-  u1 = reducemodmatrix(ginv(u1), W);
-  for (c=1; c<l; c++)
-    if (gcmp1(gcoeff(met,c,c))) break;
+  c = lg(met);
   if (DEBUGLEVEL) msgtimer("smith/class group");
   res=cgetg(c,t_VEC); init = (GEN*)cgetg(l,t_VEC);
   for (i=1; i<l; i++)
     init[i] = primeform(Disc,stoi(labs(vectbase[vperm[i]])),prec);
   for (j=1; j<c; j++)
   {
-    p1 = NULL;
+    GEN p1 = NULL;
     for (i=1; i<l; i++)
-    {
-      p2 = gpui(init[i], gcoeff(u1,i,j), prec);
-      p1 = comp(p1,p2);
-    }
+      p1 = comp(p1, powgi(init[i], gcoeff(u1,i,j)));
     res[j] = (long)p1;
   }
   if (DEBUGLEVEL) msgtimer("generators");
@@ -1601,8 +1596,8 @@ gcdreal(GEN a,GEN b,long *pte)
   long e;
   GEN k1,r;
 
-  if (!signe(a)) return mpcopy(b);
-  if (!signe(b)) return mpcopy(a);
+  if (!signe(a)) return mpabs(b);
+  if (!signe(b)) return mpabs(a);
 
   if (typ(a)==t_INT)
   {
@@ -1649,11 +1644,12 @@ get_reg(GEN matc, long sreg)
 GEN
 buchquad(GEN D, double cbach, double cbach2, long RELSUP0, long flag, long prec)
 {
-  long av0 = avma,av,tetpil,KCCO,KCCOPRO,i,j,s, *ex,**mat;
+  ulong av0 = avma, av;
+  long KCCO,KCCOPRO,i,j,s, *ex,**mat;
   long extrarel,nrelsup,nreldep,LIMC,LIMC2,cp,nbram,nlze;
-  GEN p1,h,W,met,res,basecl,dr,c_1,pdep,C,B,extramat,extraC;
-  GEN reg,vecexpo,glog2,cst;
-  double drc,lim,LOGD;
+  GEN p1,h,W,met,res,basecl,dr,pdep,C,B,extramat,extraC;
+  GEN R,vecexpo,glog2,resc,Res,z;
+  double c_1,drc,lim,LOGD;
 
   Disc = D; if (typ(Disc)!=t_INT) err(typeer,"buchquad");
   s=mod4(Disc);
@@ -1682,8 +1678,18 @@ buchquad(GEN D, double cbach, double cbach2, long RELSUP0, long flag, long prec)
     err(warner,"not a fundamental discriminant in quadclassunit");
   buch_init(); RELSUP = RELSUP0;
   dr=cgetr(3); affir(Disc,dr); drc=fabs(rtodbl(dr)); LOGD=log(drc);
-  lim=sqrt(drc); cst = mulrr(lfunc(Disc), dbltor(lim));
+
+  lim = sqrt(drc);
+  resc = dbltor(lim); Res = Lval1(Disc);
+  /* resc = sqrt(D) w / 2^r1 (2pi)^r2 ~ hR / L(chi,1) */
+  if (PRECREG)
+    resc = gmul2n(resc,-1);
+  else
+    resc = divrr(resc, mppi(DEFAULTPREC));
+  z = mulrr(Res, resc); /* ~ hR */
+  R = gun;
   if (!PRECREG) lim /= sqrt(3.);
+
   cp = (long)exp(sqrt(LOGD*log(LOGD)/8.0));
   if (cp < 13) cp = 13;
   av = avma; cbach /= 2;
@@ -1781,56 +1787,48 @@ EXTRAREL:
     }
   }
   /* tentative class number */
-  h=gun; for (i=1; i<lg(W); i++) h=mulii(h,gcoeff(W,i,i));
+  h = dethnf_i(W);
   if (PRECREG)
   {
     /* tentative regulator */
-    reg = get_reg(C, KCCOPRO - (lg(B)-1) - (lg(W)-1));
-    if (!reg)
+    R = get_reg(C, KCCOPRO - (lg(B)-1) - (lg(W)-1));
+    if (!R)
     {
-      desalloc(mat);
-      prec = (PRECREG<<1)-2; goto INCREASE;
+      prec = (PRECREG<<1)-2;
+      desalloc(mat); goto INCREASE;
     }
-    if (gexpo(reg)<=-3)
+    if (gexpo(R) <= -3)
     {
       if (++nrelsup <= 7)
       {
-        if (DEBUGLEVEL) { fprintferr("regulateur nul\n"); flusherr(); }
+        if (DEBUGLEVEL) fprintferr("regulator is zero\n");
         nlze=min(KC,nrelsup); goto EXTRAREL;
       }
       desalloc(mat); goto INCREASE;
     }
-    c_1 = divrr(gmul2n(gmul(h,reg),1), cst);
-  }
-  else
-  {
-    reg = gun;
-    c_1 = divrr(gmul(h,mppi(DEFAULTPREC)), cst);
   }
 
-  if (gcmpgs(gmul2n(c_1,2),3)<0) { c_1=stoi(10); nrelsup=7; }
-  if (gcmpgs(gmul2n(c_1,1),3)>0)
+  c_1 = gtodouble(gdiv(gmul(h,R), z));
+  if (c_1 < 0.75) { c_1 = 10.; nrelsup = 7; }
+  if (c_1 > 1.5)
   {
-    nrelsup++;
-    if (nrelsup<=7)
+    if (++nrelsup <= 7)
     {
-      if (DEBUGLEVEL)
-        { fprintferr("***** check = %f\n\n",gtodouble(c_1)); flusherr(); }
+      if (DEBUGLEVEL) fprintferr("***** check = %f\n\n",c_1);
       nlze=min(KC,nrelsup); goto EXTRAREL;
     }
     if (cbach < 5.99) { desalloc(mat); goto INCREASE; }
     err(warner,"suspicious check. Suggest increasing extra relations.");
   }
   basecl = get_clgp(Disc,W,&met,PRECREG);
-  s = lg(basecl); desalloc(mat); tetpil=avma;
+  s = lg(basecl); desalloc(mat);
 
   res=cgetg(6,t_VEC);
-  res[1]=lcopy(h); p1=cgetg(s,t_VEC);
-  for (i=1; i<s; i++) p1[i] = (long)icopy(gcoeff(met,i,i));
-  res[2]=(long)p1;
-  res[3]=lcopy(basecl);
-  res[4]=lcopy(reg);
-  res[5]=lcopy(c_1); return gerepile(av0,tetpil,res);
+  res[1]=(long)h;
+  res[2]=(long)met;
+  res[3]=(long)basecl;
+  res[4]=(long)R;
+  res[5]=(long)dbltor(c_1); return gerepilecopy(av0,res);
 }
 
 GEN
