@@ -1,4 +1,4 @@
-/* $Id: buch2.c,v 1.97 2001/10/17 12:50:41 karim Exp $
+/* $Id: buch2.c,v 1.120 2001/12/18 21:41:26 karim Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -21,6 +21,7 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
 /*******************************************************************/
 #include "pari.h"
 #include "parinf.h"
+extern GEN famat_to_arch(GEN nf, GEN fa, long prec);
 extern GEN to_famat_all(GEN x, GEN y);
 extern int addcolumntomatrix(GEN V, GEN invp, GEN L);
 extern double check_bach(double cbach, double B);
@@ -78,7 +79,7 @@ static long
 ccontent(GEN x)
 {
   long i, l = lg(x), s=labs(x[1]);
-  for (i=2; i<l && s>1; i++) s = cgcd(x[i],s);
+  for (i=2; i<l && s!=1; i++) s = cgcd(x[i],s);
   return s;
 }
 
@@ -304,14 +305,12 @@ FBgen(GEN nf,long n2,long n)
   return Res;
 }
 
-/* can we factor I / m ? (m pseudo minimum, computed in ideallllredpart1) */
+/* can we factor I / m ? (m pseudo minimum, computed in pseudomin) */
 static long
-factorgen(GEN nf,GEN idealvec,long kcz,long limp)
+factorgen(GEN nf,GEN I,GEN m,long kcz,long limp)
 {
   long i,j,n1,ip,v,p,k,lo,ifinal;
   GEN x,q,r,P,p1,listexpo;
-  GEN I = (GEN)idealvec[1];
-  GEN m = (GEN)idealvec[2];
   GEN Nm= absi( subres(gmul((GEN)nf[7],m), (GEN)nf[1]) ); /* |Nm| */
 
   x = divii(Nm, dethnf_i(I)); /* m in I, so NI | Nm */
@@ -771,12 +770,11 @@ split_ideal(GEN nf, GEN x0, long prec, GEN vperm)
   if (flag && factorgensimple(nf,y)) return y;
 
   y = ideallllred(nf,x0,NULL,prec);
-  if (flag &&  ((!x0[2] && gegal((GEN)y[1], (GEN)x[1]))
-             ||  (x0[2] && gcmp0((GEN)y[2])))) flag = 0; /* y == x0 */
+  if (flag && gcmp0((GEN)y[2])) flag = 0; /* y == x0 */
   if (flag && factorgensimple(nf,y)) return y;
 
   z = init_idele(nf); ru = lg(z[2]);
-  if (!x0[2]) { z[2] = 0; x0 = x; } /* stop cheating */
+  z[2] = lgetg(1, t_MAT);
   vdir = cgetg(ru,t_VEC);
   for (i=2; i<ru; i++) vdir[i]=zero;
   for (i=1; i<ru; i++)
@@ -998,13 +996,14 @@ isprincipalarch(GEN bnf, GEN col, GEN kNx, GEN e, GEN dx, long *pe)
 static int
 fact_ok(GEN nf, GEN y, GEN C, GEN g, GEN e)
 {
+  ulong av = avma;
   long i, c = lg(e);
   GEN z = C? C: gun;
   for (i=1; i<c; i++)
     if (signe(e[i])) z = idealmul(nf, z, idealpow(nf, (GEN)g[i], (GEN)e[i]));
   if (typ(z) != t_MAT) z = idealhermite(nf,z);
   if (typ(y) != t_MAT) y = idealhermite(nf,y);
-  return gegal(y, z);
+  i = gegal(y, z); avma = av; return i;
 }
 
 /* assume x in HNF. cf class_group_gen for notations */
@@ -1030,9 +1029,10 @@ isprincipalall0(GEN bnf, GEN x, long *ptprec, long flag)
   /* factor x */
   xc = content(x); if (!gcmp1(xc)) x = gdiv(x,xc);
 
-  p1 = init_idele(nf); p1[1] = (long)x;
-  if (!(flag & nf_GEN)) p1[2] = 0; /* don't compute archimedean part */
-  xar = split_ideal(nf,p1,prec,vperm);
+  p1 = init_idele(nf);
+  p1[1] = (long)x;
+  p1[2] = lgetg(1,t_MAT); /* compute archimedean part as a famat */
+  xar= split_ideal(nf,p1,prec,vperm);
 
   lW = lg(W)-1; Wex = zerocol(lW);
   lB = lg(B)-1; Bex = zerocol(lB); get_split_expo(Wex,Bex,vperm);
@@ -1063,11 +1063,11 @@ isprincipalall0(GEN bnf, GEN x, long *ptprec, long flag)
   { /* g A = G Ur A + [ga]A, Ur A = D Q + R as above (R = ex)
            = G R + [GD]Q + [ga]A */
     GEN ga = (GEN)clg2[2], GD = (GEN)clg2[3];
-    col = act_arch(Bex, WB_C + lW);
+    if (lB) col = act_arch(Bex, WB_C + lW); else col = zerovec(1); /* nf=Q ! */
     if (lW) col = gadd(col, act_arch(A, ga));
     if (c)  col = gadd(col, act_arch(Q, GD));
   }
-  col = gsub(col, (GEN)xar[2]);
+  col = gadd(col, famat_to_arch(nf, (GEN)xar[2], prec));
 
   /* find coords on Zk; Q = N (x / \prod gj^ej) = N(alpha), denom(alpha) | d */
   Q = gdiv(dethnf_i(x), get_norm_fact(gen, ex, &d));
@@ -1115,14 +1115,14 @@ isprincipalall(GEN bnf,GEN x,long flag)
       err(talker,"not the same number field in isprincipal");
     x = (GEN)x[2]; tx = t_POL;
   }
-  if (tx == t_POL || tx == t_COL)
+  if (tx == t_POL || tx == t_COL || tx == t_INT || tx == t_FRAC)
   {
     if (gcmp0(x)) err(talker,"zero ideal in isprincipal");
     return triv_gen(nf, x, lg(mael3(bnf,8,1,2))-1, flag);
   }
   x = idealhermite(nf,x);
   if (lg(x)==1) err(talker,"zero ideal in isprincipal");
-  if (lgef(nf[1])==4)
+  if (degpol(nf[1]) == 1)
     return gerepileupto(av, triv_gen(nf, gcoeff(x,1,1), 0, flag));
 
   pr = prec_arch(bnf); /* precision of unit matrix */
@@ -1396,6 +1396,21 @@ unset_fact(GEN c)
   for (i=1; i<=primfact[0]; i++) c[primfact[i]] = 0;
 }
 
+/* as small_to_mat, with explicit dimension d */
+static GEN
+small_to_mat_i(GEN z, long d)
+{
+  long i,j, l = lg(z);
+  GEN x = cgetg(l,t_MAT);
+  for (j=1; j<l; j++)
+  {
+    GEN c = cgetg(d+1, t_COL), cz = (GEN)z[j];
+    x[j] = (long)c;
+    for (i=1; i<=d; i++) c[i] = lstoi(cz[i]);
+  }
+  return x;
+}
+
 #define MAXTRY 20
 
 /* return -1 in case of precision problems. t = current number of relations */
@@ -1406,7 +1421,7 @@ small_norm_for_buchall(long cglob,GEN *mat,GEN matarch,long LIMC, long PRECREG,
   const double eps = 0.000001;
   double *y,*z,**q,*v, MINKOVSKI_BOUND,BOUND;
   ulong av = avma, av1,av2,limpile;
-  long i,j,k,noideal, nbrel = lg(mat)-1;
+  long j,k,noideal, nbrel = lg(mat)-1;
   long alldep = 0, nbsmallnorm,nbfact,R1, N = degpol(nf[1]);
   GEN x,xembed,M,T2,r,cbase,invcbase,T2vec,prvec;
 
@@ -1432,7 +1447,8 @@ small_norm_for_buchall(long cglob,GEN *mat,GEN matarch,long LIMC, long PRECREG,
   MINKOVSKI_BOUND = get_minkovski(N,R1,(GEN)nf[3],gborne);
   for (noideal=KC; noideal; noideal--)
   {
-    long nbrelideal=0, dependent = 0;
+    ulong av0 = avma;
+    long nbrelideal=0, dependent = 0, oldcglob = cglob;
     GEN IDEAL, ideal = (GEN)vectbase[noideal];
     GEN normideal = idealnorm(nf,ideal);
 
@@ -1449,14 +1465,14 @@ small_norm_for_buchall(long cglob,GEN *mat,GEN matarch,long LIMC, long PRECREG,
     {
       v[k]=gtodouble(gcoeff(r,k,k));
       for (j=1; j<k; j++) q[j][k]=gtodouble(gcoeff(r,j,k));
-      if (DEBUGLEVEL>3) fprintferr("v[%ld]=%.0f ",k,v[k]);
+      if (DEBUGLEVEL>3) fprintferr("v[%ld]=%.4g ",k,v[k]);
     }
 
     BOUND = MINKOVSKI_BOUND * pow(gtodouble(normideal),2./(double)N);
     if (DEBUGLEVEL>1)
     {
       if (DEBUGLEVEL>3) fprintferr("\n");
-      fprintferr("BOUND = %.0f\n",BOUND); flusherr();
+      fprintferr("BOUND = %.4g\n",BOUND); flusherr();
     }
     BOUND += eps; av2=avma; limpile = stack_lim(av2,1);
     k = N; y[N]=z[N]=0; x[N]= (long) sqrt(BOUND/v[N]);
@@ -1486,7 +1502,7 @@ small_norm_for_buchall(long cglob,GEN *mat,GEN matarch,long LIMC, long PRECREG,
 	}
 	if (k==1) /* element complete */
 	{
-	  if (!x[1] && y[1]<=eps) goto ENDIDEAL;
+	  if (y[1]<=eps) goto ENDIDEAL; /* skip all scalars: [*,0...0] */
 	  if (ccontent(x)==1) /* primitive */
 	  {
             GEN Nx, gx = gmul_mati_smallvec(IDEAL,x);
@@ -1505,6 +1521,7 @@ small_norm_for_buchall(long cglob,GEN *mat,GEN matarch,long LIMC, long PRECREG,
       }
       cglob++; col = mat[cglob];
       set_fact(col);
+      /* make sure we get maximal rank first, then allow all relations */
       if (cglob > 1 && cglob <= KC && ! addcolumntomatrix(col,invp,L))
       { /* Q-dependent from previous ones: forget it */
         cglob--; unset_fact(col);
@@ -1532,56 +1549,35 @@ small_norm_for_buchall(long cglob,GEN *mat,GEN matarch,long LIMC, long PRECREG,
       }
     }
 ENDIDEAL:
-    invp = gerepilecopy(av1, invp);
+    if (cglob == oldcglob) avma = av0; else invp = gerepilecopy(av1, invp);
     if (DEBUGLEVEL>1) msgtimer("for this ideal");
   }
 END:
   if (DEBUGLEVEL)
   {
-    fprintferr("\n");
-    msgtimer("small norm relations");
-    if (DEBUGLEVEL>1)
-    {
-      GEN p1,tmp=cgetg(cglob+1,t_MAT);
-
-      fprintferr("Elements of small norm gave %ld relations.\n",cglob);
-      fprintferr("Computing rank: "); flusherr();
-      for(j=1;j<=cglob;j++)
-      {
-	p1=cgetg(KC+1,t_COL); tmp[j]=(long)p1;
-	for(i=1;i<=KC;i++) p1[i]=lstoi(mat[j][i]);
-      }
-      tmp = (GEN)sindexrank(tmp)[2]; k=lg(tmp)-1;
-      fprintferr("%ld; independent columns: %Z\n",k,tmp);
-    }
-    if(nbsmallnorm)
-      fprintferr("\nnb. fact./nb. small norm = %ld/%ld = %.3f\n",
+    fprintferr("\n"); msgtimer("small norm relations");
+    fprintferr("  small norms gave %ld relations, rank = %ld.\n",
+               cglob, rank(small_to_mat_i((GEN)mat, KC)));
+    if (nbsmallnorm)
+      fprintferr("  nb. fact./nb. small norm = %ld/%ld = %.3f\n",
                   nbfact,nbsmallnorm,((double)nbfact)/nbsmallnorm);
-    else
-      fprintferr("\nnb. small norm = 0\n");
   }
   avma = av; return cglob;
 }
 #undef MAXTRY
 
-/* I assumed to be integral HNF, T2 a weighted T2 matrix */
+/* I assumed to be integral HNF, T2 a weighted T2 matrix. Return an
+ * irrational m in I with T2(m) small */
 static GEN
-ideallllredpart1(GEN I, GEN T2, long prec)
+pseudomin(GEN I, GEN T2)
 {
-  GEN y,m,idealpro;
-
-  y = lllgramintern(qf_base_change(T2,I,1),100,1,prec+1);
+  GEN m, y = lllgramintern(qf_base_change(T2,I,1),100,1, 0);
   if (!y) return NULL;
 
-  /* I, m, y integral */
   m = gmul(I,(GEN)y[1]);
   if (isnfscalar(m)) m = gmul(I,(GEN)y[2]);
-
-  idealpro = cgetg(3,t_VEC);
-  idealpro[1] = (long)I;
-  idealpro[2] = (long)m; /* irrational element of small T2 norm in I */
-  if (DEBUGLEVEL>5) fprintferr("\nidealpro = %Z\n",idealpro);
-  return idealpro;
+  if (DEBUGLEVEL>5) fprintferr("\nm = %Z\n",m);
+  return m;
 }
 
 static void
@@ -1673,18 +1669,18 @@ static GEN
 remove_content(GEN I)
 {
   long N = lg(I)-1;
-  if (!gcmp1(gcoeff(I,N,N))) { GEN y=content(I); if (!gcmp1(y)) I=gdiv(I,y); }
+  if (!gcmp1(gcoeff(I,N,N))) I = primpart(I);
   return I;
 }
 
 /* if phase != 1 re-initialize static variables. If <0 return immediately */
 static long
-random_relation(long phase,long cglob,long LIMC,long PRECLLL,long PRECREG,
+random_relation(long phase,long cglob,long LIMC,long PRECREG,
                 GEN nf,GEN subFB,GEN vecT2,GEN *mat,GEN matarch,GEN list_jideal)
 {
   static long jideal, jdir;
   long lim,i,av,av1,cptzer,nbT2,lgsub,r1, jlist = 1;
-  GEN arch,col,colarch,ideal,idealpro,P,ex;
+  GEN arch,col,colarch,ideal,m,P,ex;
 
   if (phase != 1) { jideal=jdir=1; if (phase<0) return 0; }
 
@@ -1722,9 +1718,9 @@ random_relation(long phase,long cglob,long LIMC,long PRECLLL,long PRECREG,
       if (DEBUGLEVEL>2)
         fprintferr("phase=%ld,jideal=%ld,jdir=%ld,rand=%ld\n",
                    phase,jideal,jdir,getrand());
-      idealpro = ideallllredpart1(ideal,(GEN)vecT2[jdir],PRECLLL);
-      if (!idealpro) return -2;
-      if (!factorgen(nf,idealpro,KCZ,LIMC))
+      m = pseudomin(ideal,(GEN)vecT2[jdir]);
+      if (!m) return -2;
+      if (!factorgen(nf,ideal,m,KCZ,LIMC))
       {
         if (DEBUGLEVEL>1) { fprintferr("."); flusherr(); }
         continue;
@@ -1757,7 +1753,7 @@ random_relation(long phase,long cglob,long LIMC,long PRECLLL,long PRECREG,
         }
       colarch = (GEN)matarch[cglob];
       /* arch = archimedean component (MULTIPLICATIVE form) of ideal */
-      arch = vecdiv(arch, gmul(gmael(nf,5,1), (GEN)idealpro[2]));
+      arch = vecdiv(arch, gmul(gmael(nf,5,1), m));
       set_log_embed(colarch, arch, r1, PRECREG);
       if (DEBUGLEVEL) dbg_newrel(jideal,jdir,phase,cglob,col,colarch,lim);
 
@@ -1785,15 +1781,21 @@ static long
 be_honest(GEN nf,GEN subFB,long PRECLLL)
 {
   ulong av;
-  GEN MC = gmael(nf,5,2), M = gmael(nf,5,1), D = (GEN)nf[3];
-  long ex,i,j,J,k,iz,nbtest, lgsub = lg(subFB), ru = lg(MC);
-  GEN P,ideal,idealpro, exu = cgetg(ru, t_VECSMALL), MCtw= cgetg(ru, t_MAT);
+  long ex,i,j,J,k,iz,nbtest,ru, lgsub = lg(subFB);
+  GEN MCtw,MC,M,P,ideal,m,exu, D = (GEN)nf[3];
 
+  if (KCZ2 <= KCZ) return 1;
   if (DEBUGLEVEL)
   {
     fprintferr("Be honest for primes from %ld to %ld\n", FB[KCZ+1],FB[KCZ2]);
     flusherr();
   }
+  if (!powsubFB) powsubFBgen(nf,subFB,CBUCHG+1,0);
+  M = gprec_w(gmael(nf,5,1), PRECLLL);
+  MC= gprec_w(gmael(nf,5,2), PRECLLL);
+  ru = lg(MC);
+  exu = cgetg(ru, t_VECSMALL);
+  MCtw= cgetg(ru, t_MAT);
   av = avma;
   for (iz=KCZ+1; iz<=KCZ2; iz++, avma = av)
   {
@@ -1825,9 +1827,9 @@ be_honest(GEN nf,GEN subFB,long PRECLLL)
 	  }
           for (i=1; i<ru; i++)
             MCtw[i] = exu[i]? lmul2n((GEN)MC[i],exu[i]<<1): MC[i];
-          idealpro = ideallllredpart1(ideal, mulmat_real(MCtw,M), PRECLLL);
-          if (idealpro && factorgen(nf,idealpro,iz-1,FB[iz-1])) break;
-	  nbtest++; if (nbtest==200) return 0;
+          m = pseudomin(ideal, mulmat_real(MCtw,M));
+          if (m && factorgen(nf,ideal,m,iz-1,FB[iz-1])) break;
+	  if (++nbtest==200) return 0;
 	}
 	avma = av2; if (k < ru) break;
       }
@@ -1852,8 +1854,8 @@ trunc_error(GEN x)
 static GEN
 compute_multiple_of_R(GEN xarch,long RU,long N,GEN *ptlambda)
 {
-  GEN T,v,mdet,mdet_t,Im_mdet,kR,lambda,xreal, *gptr[2];
-  long av = avma, i,j, zc = lg(xarch)-1, R1 = 2*RU - N;
+  GEN T,v,mdet,mdet_t,Im_mdet,kR,xreal,lambda, *gptr[2];
+  long av = avma, i, zc = lg(xarch)-1, R1 = 2*RU - N;
 
   if (DEBUGLEVEL) fprintferr("\n#### Computing regulator multiple\n");
   xreal = greal(xarch); /* = (log |sigma_i(u_j)|) */
@@ -1877,32 +1879,54 @@ compute_multiple_of_R(GEN xarch,long RU,long N,GEN *ptlambda)
 
   kR = mpabs(kR);
   lambda = gauss(Im_mdet,xreal); /* approximate rational entries */
-  for (i=1; i<=zc; i++)
-  {
-    GEN p1 = (GEN)lambda[i]; setlg(p1, RU);
-    for (j=1; j<RU; j++)
-      if (trunc_error((GEN)p1[j])) { *ptlambda = NULL; return gzero; }
-  }
-  *ptlambda = lambda;
-  gptr[0]=ptlambda; gptr[1]=&kR;
-  gerepilemany(av,gptr,2); return kR;
+  for (i=1; i<=zc; i++) setlg(lambda[i], RU);
+  gptr[0]=&lambda; gptr[1]=&kR; gerepilemany(av,gptr,2); 
+  *ptlambda = lambda; return kR;
 }
 
-extern GEN hnflll_i(GEN A, GEN *ptB, int remove);
+static GEN
+bestappr_noer(GEN x, GEN k)
+{
+  VOLATILE GEN y;
+  jmp_buf env;
+  void *c;
+  if (setjmp(env)) return NULL;
+  else
+  {
+    c = err_catch(precer, env, NULL);
+    y = bestappr(x,k);
+  }
+  err_leave(&c);
+  return y;
+}
 
-/* c = Rz = 2n, according to Dirichlet's formula. Compute a tentative
- * regulator (not a multiple this time). *ptkR = multiple of regulator */
+/* Input:
+ * lambda = approximate rational entries: coords of units found so far on a
+ * sublattice of maximal rank (sublambda)
+ * *ptkR = regulator of sublambda = multiple of regulator of lambda
+ * Compute R = true regulator of lambda.
+ *
+ * If c := Rz = 2, by Dirichlet's formula, then lambda is the full group of
+ * units AND the full set of relations for the class group has been computed.
+ *
+ * Output: *ptkR = R, *ptU = basis of fundamental units (in terms lambda) */
 static int
-compute_R(GEN lambda, GEN z, GEN *ptU, GEN *ptkR)
+compute_R(GEN lambda, GEN z, GEN *ptL, GEN *ptkR)
 {
   ulong av = avma;
   long r;
-  GEN U,L,H,gc,den,R;
+  GEN L,H,gc,den,R;
   double c;
 
   if (DEBUGLEVEL) { fprintferr("\n#### Computing check\n"); flusherr(); }
   gc = gmul(*ptkR,z);
-  lambda = bestappr(lambda,gc); den = denom(lambda);
+  lambda = bestappr_noer(lambda,gc);
+  if (!lambda)
+  {
+    if (DEBUGLEVEL) fprintferr("truncation error in bestappr\n");
+    return PRECI;
+  }
+  den = denom(lambda);
   if (gcmp(den,gc) > 0)
   {
     if (DEBUGLEVEL) fprintferr("c = %Z\nden = %Z\n",gc,den);
@@ -1920,10 +1944,7 @@ compute_R(GEN lambda, GEN z, GEN *ptU, GEN *ptkR)
   }
   if (c < 1.5) return PRECI;
   if (c > 3.) { avma = av; return RELAT; }
-  /* *pU = coords. of fundamental units on sublambda */
-  H = hnflll_i(L,&U,0); /* try hard to get a SMALL base change */
-  U += (lg(U)-1 - r); U[0] = evaltyp(t_MAT)|evallg(r+1);
-  *ptkR = R; *ptU = U; return LARGE;
+  *ptkR = R; *ptL = L; return LARGE;
 }
 
 /* find the smallest (wrt norm) among I, I^-1 and red(I^-1) */
@@ -1931,18 +1952,15 @@ static GEN
 inverse_if_smaller(GEN nf, GEN I, long prec)
 {
   GEN J, d, dmin, I1;
-  int inv;
+
   J = (GEN)I[1];
   dmin = dethnf_i(J); I1 = idealinv(nf,I);
   J = (GEN)I1[1]; J = gmul(J,denom(J)); I1[1] = (long)J;
-  d = dethnf_i(J);
-  if (cmpii(d,dmin) < 0) {inv=1; I=I1; dmin=d;}
-  else                    inv=0;
+  d = dethnf_i(J); if (cmpii(d,dmin) < 0) {I=I1; dmin=d;}
   /* try reducing (often _increases_ the norm) */
   I1 = ideallllred(nf,I1,NULL,prec);
   J = (GEN)I1[1];
-  d = dethnf_i(J);
-  if (cmpii(d,dmin) < 0) {inv=1; I=I1;}
+  d = dethnf_i(J); if (cmpii(d,dmin) < 0) I=I1;
   return I;
 }
 
@@ -1961,19 +1979,30 @@ setlg_col(GEN U, long l)
   for (; c>U; c--) setlg(*c, l);
 }
 
+GEN
+get_Vbase(GEN W, GEN vectbase, GEN vperm)
+{
+  long i, l = lg(W);
+  GEN Vbase = cgetg(l,t_VEC);
+  if (typ(vperm) == t_VECSMALL)
+    for (i=1; i<l; i++) Vbase[i] = vectbase[vperm[i]];
+  else
+    for (i=1; i<l; i++) Vbase[i] = vectbase[itos((GEN)vperm[i])];
+  return Vbase;
+}
+
 /* compute class group (clg1) + data for isprincipal (clg2) */
 static void
-class_group_gen(GEN nf,GEN W,GEN C,GEN vperm,GEN *ptclg1,GEN *ptclg2,long prec)
+class_group_gen(GEN nf,GEN W,GEN C,GEN Vbase,long prec, GEN nf0,
+                GEN *ptclg1,GEN *ptclg2)
 {
-  GEN z,G,Ga,ga,GD,cyc,X,Y,D,U,V,Ur,Ui,Uir,I,J,Vbase;
+  GEN z,G,Ga,ga,GD,cyc,X,Y,D,U,V,Ur,Ui,Uir,I,J;
   long i,j,s,lo,lo0;
 
   if (DEBUGLEVEL)
     { fprintferr("\n#### Computing class group generators\n"); timer2(); }
-  z = smith2(W); /* U W V = D, D diagonal, G = g Ui (G=new gens, g=old)  */
-  U = (GEN)z[1]; Ui = ginv(U);
-  V = (GEN)z[2];
-  D = (GEN)z[3];
+  D = smithall(W,&U,&V); /* UWV = D, D diagonal, G = g Ui (G=new gens, g=old) */
+  Ui = ginv(U);
   lo0 = lo = lg(D);
  /* we could set lo = lg(cyc) and truncate all matrices below
   *   setlg_col(D && U && Y, lo) + setlg(D && V && X && Ui, lo)
@@ -1988,39 +2017,34 @@ class_group_gen(GEN nf,GEN W,GEN C,GEN vperm,GEN *ptclg1,GEN *ptclg2,long prec)
   * P invertible diagonal matrix (\pm 1) which is only implicitly defined
   * G = g Uir P + [Ga],  Uir = Ui + WX
   * g = G P Ur  + [ga],  Ur  = U + DY */
-  Vbase = cgetg(lo0,t_VEC);
-  if (typ(vperm) == t_VECSMALL)
-    for (i=1; i<lo0; i++) Vbase[i] = vectbase[vperm[i]];
-  else
-    for (i=1; i<lo0; i++) Vbase[i] = vectbase[itos((GEN)vperm[i])];
-
   G = cgetg(lo,t_VEC);
   Ga= cgetg(lo,t_VEC);
-  z = init_idele(nf);
+  z = init_idele(nf); z[2] = lgetg(1, t_MAT);
+  if (!nf0) nf0 = nf;
   for (j=1; j<lo; j++)
   {
     GEN p1 = gcoeff(Uir,1,j);
-    z[1]=Vbase[1]; I = idealpowred(nf,z,p1,prec);
+    z[1]=Vbase[1]; I = idealpowred(nf0,z,p1,prec);
     if (signe(p1)<0) I[1] = lmul((GEN)I[1],denom((GEN)I[1]));
     for (i=2; i<lo0; i++)
     {
       p1 = gcoeff(Uir,i,j); s = signe(p1);
       if (s)
       {
-	z[1]=Vbase[i]; J = idealpowred(nf,z,p1,prec);
+	z[1]=Vbase[i]; J = idealpowred(nf0,z,p1,prec);
         if (s<0) J[1] = lmul((GEN)J[1],denom((GEN)J[1]));
-	I = idealmulh(nf,I,J);
-	I = ideallllred(nf,I,NULL,prec);
+	I = idealmulh(nf0,I,J);
+	I = ideallllred(nf0,I,NULL,prec);
       }
     }
-    J = inverse_if_smaller(nf, I, prec);
+    J = inverse_if_smaller(nf0, I, prec);
     if (J != I)
     { /* update wrt P */
       neg_row(Y ,j); V[j] = lneg((GEN)V[j]);
       neg_row(Ur,j); X[j] = lneg((GEN)X[j]);
     }
     G[j] = (long)J[1]; /* generator, order cyc[j] */
-    Ga[j]= (long)J[2];
+    Ga[j]= lneg(famat_to_arch(nf, (GEN)J[2], prec));
   }
   /* at this point Y = PY, Ur = PUr, V = VP, X = XP */
 
@@ -2077,13 +2101,19 @@ shift_t2(GEN T2, GEN M, GEN MC, long a, long b)
 }
 
 static GEN
-compute_vecT2(long RU,GEN nf)
+compute_vecT2(GEN nf,long prec)
 {
   GEN vecT2, M = gmael(nf,5,1), MC = gmael(nf,5,2), T2 = gmael(nf,5,3);
-  long i,j,n = min(RU,9), ind = 1;
+  long i,j,ind, n = min(lg(M[1])-1, 9);
 
   vecT2=cgetg(1 + n*(n+1)/2,t_VEC);
-  for (j=1; j<=n; j++)
+  if (nfgetprec(nf) > prec)
+  {
+    M = gprec_w(M,prec);
+    MC= gprec_w(MC,prec);
+    T2= gprec_w(T2,prec);
+  }
+  for (ind=j=1; j<=n; j++)
     for (i=1; i<=j; i++) vecT2[ind++] = (long)shift_t2(T2,M,MC,i,j);
   if (DEBUGLEVEL) msgtimer("weighted T2 matrices");
   return vecT2;
@@ -2130,7 +2160,7 @@ addcolumntomatrix(GEN V, GEN invp, GEN L)
  * Starting from the identity (canonical basis of Q^n), we obtain a base
  * change matrix P by taking the independant columns of A and vectors from
  * the canonical basis. Update invp = 1/P, and L in {0,1}^n, with L[i] = 0
- * of P[i] = e_i, and 1 if P[i] = A_i (i-th column of A)
+ * if P[i] = e_i, and 1 if P[i] = A_i (i-th column of A)
  */
 static GEN
 relationrank(GEN *A, long r, GEN L)
@@ -2440,20 +2470,20 @@ get_arch2(GEN nf, GEN a, long prec, int units)
 }
 
 static void
-my_class_group_gen(GEN bnf, GEN *ptcl, GEN *ptcl2, long prec)
+my_class_group_gen(GEN bnf, long prec, GEN nf0, GEN *ptcl, GEN *ptcl2)
 {
-  GEN W=(GEN)bnf[1], C=(GEN)bnf[4], vperm=(GEN)bnf[6], nf=(GEN)bnf[7], *gptr[2];
-  long av = avma;
+  ulong av = avma;
+  GEN W=(GEN)bnf[1], C=(GEN)bnf[4], nf=(GEN)bnf[7], *gptr[2];
+  GEN Vbase = get_Vbase(W, (GEN)bnf[5], (GEN)bnf[6]);
 
-  vectbase = (GEN)bnf[5]; /* needed by class_group_gen */
-  class_group_gen(nf,W,C,vperm,ptcl,ptcl2, prec);
+  class_group_gen(nf,W,C,Vbase,prec,nf0, ptcl,ptcl2);
   gptr[0]=ptcl; gptr[1]=ptcl2; gerepilemany(av,gptr,2);
 }
 
 GEN
 bnfnewprec(GEN bnf, long prec)
 {
-  GEN nf,ro,res,p1,funits,mun,matal,clgp,clgp2,y;
+  GEN nf0 = (GEN)bnf[7], nf,ro,res,p1,funits,mun,matal,clgp,clgp2,y;
   long r1,r2,ru,pl1,pl2,prec1;
 
   bnf = checkbnf(bnf);
@@ -2462,13 +2492,12 @@ bnfnewprec(GEN bnf, long prec)
   funits = check_units(bnf,"bnfnewprec");
   nf = (GEN)bnf[7];
   ro = (GEN)nf[6]; ru = lg(ro)-1;
-  r1 = nf_get_r1(nf);
-  r2 = (r1 + ru) >> 1;
+  nf_get_sign(nf, &r1, &r2);
   pl1 = (ru == 1 && r1 == 0)? 0: gexpo(funits);
   pl2 = gexpo(ro);
   prec1 = prec;
   prec += ((ru + r2 - 1) * (pl1 + (ru + r2) * pl2)) >> TWOPOTBITS_IN_LONG;
-  nf = nfnewprec((GEN)bnf[7],prec);
+  nf = nfnewprec(nf0,prec);
   res = cgetg(7,t_VEC);
   ro = (GEN)nf[6];
   mun = get_arch2(nf,funits,prec,1);
@@ -2489,7 +2518,7 @@ bnfnewprec(GEN bnf, long prec)
   y[6]=lcopy((GEN)bnf[6]);
   y[7]=(long)nf;
   y[8]=(long)res;
-  my_class_group_gen(y,&clgp,&clgp2,prec);
+  my_class_group_gen(y,prec,nf0, &clgp,&clgp2);
   res[1]=(long)clgp;
   y[9]=(long)clgp2;
   y[10]=lcopy((GEN)bnf[10]); return y;
@@ -2511,7 +2540,7 @@ bnfmake(GEN sbnf, long prec)
 {
   long av = avma, j,k,n,r1,r2,ru,lpf;
   GEN p1,x,bas,ro,nf,mun,funits,index;
-  GEN pfc,vp,mc,clgp,clgp2,res,y,W,racu,reg,matal;
+  GEN pfc,vp,mc,clgp,clgp2,res,y,W,racu,reg,matal,vectbase,Vbase;
 
   if (typ(sbnf)!=t_VEC || lg(sbnf)!=13)
     err(talker,"incorrect sbnf in bnfmake");
@@ -2549,7 +2578,8 @@ bnfmake(GEN sbnf, long prec)
     vectbase[j]=(long)decodeprime(nf,(GEN)pfc[j]);
   }
   W = (GEN)sbnf[7];
-  class_group_gen(nf,W,mc,vp,&clgp,&clgp2, prec); /* uses vectbase */
+  Vbase = get_Vbase(W,vectbase,vp);
+  class_group_gen(nf,W,mc,Vbase,prec,NULL, &clgp,&clgp2);
 
   reg = get_regulator(mun,prec);
   p1=cgetg(3,t_VEC); racu=(GEN)sbnf[10];
@@ -2708,7 +2738,7 @@ buchall_end(GEN nf,GEN CHANGE,long fl,long k, GEN fu, GEN clg1, GEN clg2,
     RES[2]=nf[2]; p1=cgetg(3,t_VEC); p1[1]=nf[3]; p1[2]=nf[4];
     RES[3]=(long)p1;
     RES[4]=nf[7];
-    z=cgetg(2,t_MAT); z[1]=lcopy(RES); return z;
+    z=cgetg(2,t_MAT); z[1]=(long)RES; return z;
   }
   z=cgetg(11,t_VEC);
   z[1]=(long)W;
@@ -2724,7 +2754,7 @@ buchall_end(GEN nf,GEN CHANGE,long fl,long k, GEN fu, GEN clg1, GEN clg2,
   z[9]=(long)clg2;
   z[10]=zero; /* dummy: we MUST have lg(bnf) != lg(nf) */
   if (CHANGE) { p1=cgetg(3,t_VEC); p1[1]=(long)z; p1[2]=(long)CHANGE; z=p1; }
-  return gcopy(z);
+  return z;
 }
 
 static GEN
@@ -2735,13 +2765,13 @@ buchall_for_degree_one_pol(GEN nf, GEN CHANGE, long flun)
   GEN fu=cgetg(1,t_VEC), reg=gun, c_1=gun, zu=cgetg(3,t_VEC);
   GEN clg1=cgetg(4,t_VEC), clg2=cgetg(4,t_VEC);
 
-  clg1[1]=un; clg1[2]=clg1[3]=clg2[3]=lgetg(1,t_VEC);
-  clg2[1]=clg2[2]=lgetg(1,t_MAT);
+  clg1[1]=un; clg1[2]=clg1[3]=clg2[2]=clg2[3]=lgetg(1,t_VEC);
+  clg2[1]=lgetg(1,t_MAT);
   zu[1]=deux; zu[2]=lnegi(gun);
   W=B=xarch=matarch=cgetg(1,t_MAT);
   vectbase=cgetg(1,t_COL); vperm=cgetg(1,t_VEC);
 
-  return gerepileupto(av, buchall_end(nf,CHANGE,flun,k,fu,clg1,clg2,reg,c_1,zu,W,B,xarch,matarch,vectbase,vperm));
+  return gerepilecopy(av, buchall_end(nf,CHANGE,flun,k,fu,clg1,clg2,reg,c_1,zu,W,B,xarch,matarch,vectbase,vperm));
 }
 
 static GEN
@@ -2764,12 +2794,12 @@ buchall(GEN P,GEN gcbach,GEN gcbach2,GEN gRELSUP,GEN gborne,long nbrelpid,
         long minsFB,long flun,long prec)
 {
   ulong av = avma,av0,av1,limpile;
-  long N,R1,R2,RU,PRECREG,PRECLLL,KCCO,RELSUP,LIMC,LIMC2,lim;
+  long N,R1,R2,RU,PRECREG,PRECLLL,PRECLLLadd,KCCO,RELSUP,LIMC,LIMC2,lim;
   long nlze,zc,nrelsup,nreldep,phase,matmax,i,j,k,ss,cglob;
-  long first=1, sfb_increase=0, sfb_trials=0, precdouble=0, precadd=0;
+  long sfb_increase=0, sfb_trials=0, precdouble=0, precadd=0;
   double cbach,cbach2,drc,LOGD2;
   GEN p1,vecT2,fu,zu,nf,LLLnf,D,xarch,W,R,Res,z,h,vperm,subFB;
-  GEN resc,B,C,c1,lambda,pdep,parch,liste,invp,clg1,clg2, *mat;
+  GEN L,resc,B,C,c1,lambda,pdep,liste,invp,clg1,clg2,Vbase, *mat;
   GEN CHANGE=NULL, extramat=NULL, extraC=NULL, list_jideal=NULL;
   char *precpb = NULL;
 
@@ -2782,7 +2812,9 @@ buchall(GEN P,GEN gcbach,GEN gcbach2,GEN gRELSUP,GEN gborne,long nbrelpid,
 
   /* Initializations */
   fu = NULL; /* gcc -Wall */
-  N = degpol(P); PRECREG = max(BIGDEFAULTPREC,prec);
+  N = degpol(P);
+  PRECREG = max(BIGDEFAULTPREC,prec);
+  PRECLLLadd = DEFAULTPREC;
   if (!nf)
   {
     nf = initalgall0(P, nf_REGULAR, PRECREG);
@@ -2794,7 +2826,7 @@ buchall(GEN P,GEN gcbach,GEN gcbach2,GEN gRELSUP,GEN gborne,long nbrelpid,
   zu[2] = lmul((GEN)nf[7],(GEN)zu[2]);
   if (DEBUGLEVEL) msgtimer("initalg & rootsof1");
 
-  R1 = itos(gmael(nf,2,1)); R2 = (N-R1)>>1; RU = R1+R2;
+  nf_get_sign(nf, &R1, &R2); RU = R1+R2;
   D = (GEN)nf[3]; drc = fabs(gtodouble(D));
   LOGD2 = log(drc); LOGD2 = LOGD2*LOGD2;
   lim = (long) (exp(-(double)N) * sqrt(2*PI*N*drc) * pow(4/PI,(double)R2));
@@ -2813,8 +2845,7 @@ buchall(GEN P,GEN gcbach,GEN gcbach2,GEN gRELSUP,GEN gborne,long nbrelpid,
   av0 = avma; mat = NULL; FB = NULL;
 
 START:
-  avma = av0;
-  if (first) first = 0; else desallocate(&mat);
+  avma = av0; desallocate(&mat);
   if (precpb)
   {
     precdouble++;
@@ -2854,10 +2885,14 @@ START:
 
   /* PRECLLL = prec for LLL-reductions (idealred)
    * PRECREG = prec for archimedean components */
-  PRECLLL = DEFAULTPREC
-          + ((expi(D)*(lg(subFB)-2)+((N*N)>>2))>>TWOPOTBITS_IN_LONG);
+  PRECLLL = PRECLLLadd
+          + ((expi(D)*(lg(subFB)-2) + ((N*N)>>2)) >> TWOPOTBITS_IN_LONG);
   if (!precdouble) PRECREG = prec+1;
-  if (PRECREG < PRECLLL) PRECREG = PRECLLL;
+  if (PRECREG < PRECLLL)
+  { /* very rare */
+    PRECREG = PRECLLL;
+    nf = nfnewprec(nf,PRECREG); av0 = avma;
+  }
   KCCO = KC+RU-1 + max(ss,RELSUP); /* expected number of needed relations */
   if (DEBUGLEVEL)
     fprintferr("relsup = %ld, ss = %ld, KCZ = %ld, KC = %ld, KCCO = %ld\n",
@@ -2902,6 +2937,7 @@ START:
   phase = 0;
   nlze = matmax = 0; /* for lint */
   vecT2 = NULL;
+  list_jideal = NULL;
 
   /* random relations */
   if (cglob == KCCO) /* enough rels, but init random_relations just in case */
@@ -2916,7 +2952,7 @@ MORE:
     {
       if (DEBUGLEVEL) fprintferr("*** Increasing sub factor base\n");
       sfb_increase = 0;
-      if (++sfb_trials >= SFB_MAX) goto START;
+      if (++sfb_trials > SFB_MAX) goto START;
       subFB = subFBgen(N,min(lim,LIMC2), lg(subFB)-1+SFB_STEP,NULL,&ss);
       if (!subFB) goto START;
       nreldep = nrelsup = 0;
@@ -2933,7 +2969,7 @@ MORE:
       if (slim > matmax)
       {
         matmax = 2 * slim;
-        mat = (long**)gprealloc(mat, (matmax+1) * sizeof(long*));
+        mat = (GEN*)gprealloc(mat, (matmax+1) * sizeof(GEN));
       }
       setlg(mat, slim+1);
       if (DEBUGLEVEL)
@@ -2943,7 +2979,7 @@ MORE:
     }
     if (!vecT2)
     {
-      vecT2 = compute_vecT2(RU,nf);
+      vecT2 = compute_vecT2(nf,PRECLLL);
       av1 = avma;
     }
     if (!powsubFB)
@@ -2951,11 +2987,15 @@ MORE:
       powsubFBgen(nf,subFB,CBUCHG+1,PRECREG);
       av1 = avma;
     }
-    ss = random_relation(phase,cglob,(long)LIMC,PRECLLL,PRECREG,
+    ss = random_relation(phase,cglob,(long)LIMC,PRECREG,
                          nf,subFB,vecT2,mat,matarch,list_jideal);
     if (ss < 0)
     { /* could not find relations */
-      if (ss != -1) precpb = "random_relation"; /* precision pb */
+      if (ss != -1)
+      {
+        precpb = "random_relation"; /* precision pb */
+        PRECLLLadd = (PRECLLLadd<<1) - 2;
+      }
       goto START;
     }
     if (DEBUGLEVEL > 2) dbg_outrel(phase,cglob,vperm,mat,matarch);
@@ -3025,10 +3065,8 @@ MORE:
   if (DEBUGLEVEL) fprintferr("\n#### Tentative class number: %Z\n", h);
 
   /* z ~ h R if enough relations, a multiple otherwise */
-  z = mulrr(Res,resc);
-  p1 = gmul2n(divir(h,z), 1);
-  i = compute_R(lambda,p1,&parch,&R);
-  switch(i)
+  z = mulrr(Res,resc); p1 = gmul2n(divir(h,z), 1);
+  switch (compute_R(lambda,p1,&L,&R))
   {
     case PRECI: /* precision problem unless we cheat on Bach constant */
       if (!precdouble) precpb = "compute_R";
@@ -3040,16 +3078,12 @@ MORE:
   }
   /* DONE! */
 
-  if (KCZ2 > KCZ)
-  { /* "be honest" */
-    if (!powsubFB) powsubFBgen(nf,subFB,CBUCHG+1,PRECREG);
-    if (!be_honest(nf,subFB,PRECLLL)) goto START;
-  }
+  if (!be_honest(nf,subFB,PRECLLL)) goto START;
 
   /* fundamental units */
   if (flun < 0 || flun > 1)
   {
-    xarch = gmul(xarch, parch); /* arch. components of fund. units */
+    xarch = gmul(xarch, lllint(L)); /* arch. components of fund. units */
     xarch = cleancol(xarch,N,PRECREG);
     if (DEBUGLEVEL) msgtimer("cleancol");
   }
@@ -3058,11 +3092,7 @@ MORE:
     fu = getfu(nf,&xarch,R,flun,&k,PRECREG);
     if (k <= 0 && labs(flun) > 2)
     {
-      if (k < 0)
-      {
-        precadd = (DEFAULTPREC-2) + ((-k) >> TWOPOTBITS_IN_LONG);
-        if (precadd <= 0) precadd = (DEFAULTPREC-2);
-      }
+      if (k < 0) precadd = (DEFAULTPREC-2) + ((-k) >> TWOPOTBITS_IN_LONG);
       precpb = "getfu"; goto START;
     }
   }
@@ -3070,10 +3100,11 @@ MORE:
   /* class group generators */
   i = lg(C)-zc; C += zc; C[0] = evaltyp(t_MAT)|evallg(i);
   C = cleancol(C,N,PRECREG);
-  class_group_gen(nf,W,C,vperm, &clg1, &clg2, PRECREG);
+  Vbase = get_Vbase(W,vectbase,vperm);
+  class_group_gen(nf,W,C,Vbase,PRECREG,NULL, &clg1, &clg2);
 
   c1 = gdiv(gmul(R,h), z);
   desallocate(&mat);
 
-  return gerepileupto(av, buchall_end(nf,CHANGE,flun,k,fu,clg1,clg2,R,c1,zu,W,B,xarch,C,vectbase,vperm));
+  return gerepilecopy(av, buchall_end(nf,CHANGE,flun,k,fu,clg1,clg2,R,c1,zu,W,B,xarch,C,vectbase,vperm));
 }
