@@ -1,4 +1,4 @@
-/* $Id: buch3.c,v 1.48 2001/10/13 22:11:57 karim Exp $
+/* $Id: buch3.c,v 1.58 2002/01/10 21:58:16 karim Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -35,19 +35,6 @@ extern GEN isprincipalfact(GEN bnf,GEN P, GEN e, GEN C, long flag);
 extern GEN idealaddtoone_i(GEN nf, GEN x, GEN y);
 extern GEN ideallllred_elt(GEN nf, GEN I);
 
-/* U W V = D, Ui = U^(-1) */
-GEN
-compute_class_number(GEN W, GEN *D,GEN *Ui,GEN *V)
-{
-  GEN S = smith2(W);
-
-  *Ui= ginv((GEN)S[1]);
-  if (V) *V = (GEN)S[2];
-  *D = (GEN)S[3];
-  if (DEBUGLEVEL>=4) msgtimer("smith/class group");
-  return dethnf_i(*D);
-}
-
 /* FIXME: obsolete, see zarchstar (which is much slower unfortunately). */
 static GEN
 get_full_rank(GEN nf, GEN v, GEN _0, GEN _1, GEN gen, long ngen, long rankmax)
@@ -91,7 +78,7 @@ GEN
 buchnarrow(GEN bnf)
 {
   GEN nf,_0,_1,cyc,gen,v,matsign,arch,cycgen,logs;
-  GEN dataunit,p1,p2,h,clh,basecl,met,u1;
+  GEN dataunit,p1,p2,h,basecl,met,u1;
   long r1,R,i,j,ngen,sizeh,t,lo,c;
   ulong av = avma;
 
@@ -134,12 +121,12 @@ buchnarrow(GEN bnf)
     vconcat(diagonal(cyc), logs),
     vconcat(zeromat(ngen, r1-t), gscalmat(gdeux,r1-t))
   );
-  clh = compute_class_number(h,&met,&u1,NULL);
-  lo = lg(met)-1;
-  for (c=1; c<=lo; c++)
-    if (gcmp1(gcoeff(met,c,c))) break;
+ 
+  lo = lg(h)-1;
+  met = smithrel(h,NULL,&u1);
+  c = lg(met);
+  if (DEBUGLEVEL>3) msgtimer("smith/class group");
 
-  u1 = reducemodmatrix(u1,h);
   basecl = cgetg(c,t_VEC);
   for (j=1; j<c; j++)
   {
@@ -152,30 +139,31 @@ buchnarrow(GEN bnf)
       if (signe(p1))
       {
 	p2 = idealmul(nf,p2, idealpow(nf,(GEN)gen[i],p1));
-        p1 = content(p2); if (!gcmp1(p1)) p2 = gdiv(p2,p1);
+        p2 = primpart(p2);
       }
     }
     basecl[j] = (long)p2;
   }
   v = cgetg(4,t_VEC);
-  v[1] = lcopy(clh); setlg(met,c);
-  v[2] = (long)mattodiagonal(met);
-  v[3] = lcopy(basecl); return gerepileupto(av, v);
+  v[1] = (long)dethnf_i(h);
+  v[2] = (long)met;
+  v[3] = (long)basecl; return gerepilecopy(av, v);
 }
 
-/* given two coprime ideals x (integral) and id, compute alpha in x,
- * alpha = 1 mod (id), with x/alpha nearly reduced.
- */
+/* given two coprime integral ideals x and f (f HNF), compute "small" a in x,
+ * such that a = 1 mod (f). */
 static GEN
-findalpha(GEN nf,GEN x,GEN id)
+findalpha(GEN nf,GEN x,GEN f)
 {
-  GEN p1, y;
-  GEN alp = idealaddtoone_i(nf,x,id);
+  GEN p1,y,b;
 
-  y = ideallllred_elt(nf, idealmullll(nf,x,id));
-  p1 = ground(element_div(nf,alp,y));
-  alp = gsub(alp, element_mul(nf,p1,y));
-  return gcmp0(alp)? y: alp;
+  if (gcmp1(gcoeff(f,1,1)))
+    return ideallllred_elt(nf, x); /* f = 1 */
+
+  b = idealaddtoone_i(nf,x,f); /* a = b mod (x f) */
+  y = ideallllred_elt(nf, idealmullll(nf,x,f));
+  p1 = ground(element_div(nf,b,y));
+  return gsub(b, element_mul(nf,p1,y)); /* != 0 since = 1 mod f */
 }
 
 static int
@@ -192,21 +180,22 @@ too_big(GEN nf, GEN bet)
 }
 
 static GEN
-idealmodidele(GEN nf, GEN x, GEN ideal, GEN sarch, GEN arch)
+idealmodidele(GEN nf, GEN x, GEN f, GEN sarch, GEN arch)
 {
   long av = avma,i,l;
   GEN p1,p2,alp,bet,b;
 
-  nf=checknf(nf); alp=findalpha(nf,x,ideal);
-  p1=idealdiv(nf,alp,x);
-  bet = element_div(nf,findalpha(nf,p1,ideal),alp);
+  nf = checknf(nf);
+  alp= findalpha(nf,x,f);
+  p1 = idealdiv(nf,alp,x); /* small, integral, = 1/x in Cl_f */
+  bet = element_div(nf,findalpha(nf,p1,f),alp);
   if (too_big(nf,bet) > 0) { avma=av; return x; }
   p1=(GEN)sarch[2]; l=lg(p1);
   if (l > 1)
   {
     b=bet; p2=lift_intern(gmul((GEN)sarch[3],zsigne(nf,bet,arch)));
     for (i=1; i<l; i++)
-    if (signe(p2[i])) bet = element_mul(nf,bet,(GEN)p1[i]);
+      if (signe(p2[i])) bet = element_mul(nf,bet,(GEN)p1[i]);
     if (b != bet && too_big(nf,bet) > 0) { avma=av; return x; }
   }
   return idealmul(nf,bet,x);
@@ -218,39 +207,38 @@ idealmulmodidele(GEN nf,GEN x,GEN y, GEN ideal,GEN sarch,GEN arch)
   return idealmodidele(nf,idealmul(nf,x,y),ideal,sarch,arch);
 }
 
+struct muldata {
+  GEN nf, ideal, sarch, arch;
+};
+
+static GEN
+_mul(void *data, GEN x, GEN y)
+{
+  struct muldata *D = (struct muldata *)data;
+  GEN t = idealmul(D->nf,x,y);
+  return idealmodidele(D->nf, t, D->ideal, D->sarch, D->arch);
+}
+static GEN
+_sqr(void *data, GEN x) {
+  return _mul(data,x,x);
+}
+
 /* assume n > 0 */
 /* FIXME: should compute x^n = a I using idealred, then reduce a mod idele */
 static GEN
 idealpowmodidele(GEN nf,GEN x,GEN n, GEN ideal,GEN sarch,GEN arch)
 {
-  long i,m,av=avma;
+  ulong av = avma;
+  struct muldata D;
   GEN y;
-  ulong j;
 
-  if (cmpis(n, 16) < 0)
-  {
-    if (gcmp1(n)) return x;
-    x = idealpow(nf,x,n);
-    x = idealmodidele(nf,x,ideal,sarch,arch);
-    return gerepileupto(av,x);
-  }
-
-  i = lgefint(n)-1; m=n[i]; j=HIGHBIT;
-  while ((m&j)==0) j>>=1;
-  y = x;
-  for (j>>=1; j; j>>=1)
-  {
-    y = idealmul(nf,y,y);
-    if (m&j) y = idealmul(nf,y,x);
-    y = idealmodidele(nf,y,ideal,sarch,arch);
-  }
-  for (i--; i>=2; i--)
-    for (m=n[i],j=HIGHBIT; j; j>>=1)
-    {
-      y = idealmul(nf,y,y);
-      if (m&j) y = idealmul(nf,y,x);
-      y = idealmodidele(nf,y,ideal,sarch,arch);
-    }
+  if (!signe(n)) return idealpow(nf,x,n);
+  if (gcmp1(n)) return x;
+  D.nf = nf;
+  D.ideal= ideal;
+  D.arch = arch;
+  D.sarch= sarch;
+  y = leftright_pow(x, n, (void*)&D, &_sqr, &_mul);
   return gerepileupto(av,y);
 }
 
@@ -258,7 +246,7 @@ static GEN
 buchrayall(GEN bnf,GEN module,long flag)
 {
   GEN nf,cyc,gen,genplus,fa2,sarch,hmatu,u,clg,logs;
-  GEN dataunit,p1,p2,h,clh,genray,met,u1,u2,u1old,cycgen;
+  GEN dataunit,p1,p2,h,genray,met,u1,u2,U,cycgen;
   GEN racunit,bigres,bid,cycbid,genbid,x,y,funits,hmat,vecel;
   long RU,Ri,i,j,ngen,lh,lo,c,av=avma;
 
@@ -340,15 +328,19 @@ buchrayall(GEN bnf,GEN module,long flag)
     vconcat(diagonal(cyc), gneg_i(logs)),
     vconcat(zeromat(ngen, Ri), hmat)
   );
-  clh = compute_class_number(h,&met,&u1,NULL);
-  u1old = u1; lo = lg(met)-1;
+  h = hnf(h);
+ 
+  met = smithall(h, &U, NULL); /* cf smithrel */
+  lo = lg(met)-1;
   for (c=1; c<=lo; c++)
     if (gcmp1(gcoeff(met,c,c))) break;
-
+  setlg(met,c);
+ 
   if (flag & nf_GEN)
   {
     GEN Id = idmat(degpol(nf[1])), arch = (GEN)module[2];
-    u1 = reducemodmatrix(u1,h);
+    u1 = ginv(U); setlg(u1,c);
+    u1 = reducemodHNF(u1, h, NULL);
     genray = cgetg(c,t_VEC);
     for (j=1; j<c; j++)
     {
@@ -375,8 +367,9 @@ buchrayall(GEN bnf,GEN module,long flag)
     }
     clg = cgetg(4,t_VEC); clg[3] = lcopy(genray);
   } else clg = cgetg(3,t_VEC);
-  clg[1] = licopy(clh); setlg(met,c);
-  clg[2] = (long)mattodiagonal(met);
+  met = mattodiagonal(met);
+  clg[1] = (long)dethnf_i(h);
+  clg[2] = (long)met;
   if (!(flag & nf_INIT)) return gerepileupto(av,clg);
 
   u2 = cgetg(Ri+1,t_MAT);
@@ -384,17 +377,17 @@ buchrayall(GEN bnf,GEN module,long flag)
   for (j=1; j<=RU; j++) { u1[j]=u[j]; setlg(u[j],RU+1); }
   u += RU;
   for (j=1; j<=Ri; j++) { u2[j]=u[j]; setlg(u[j],RU+1); }
-  p1 = lllint(u1); p2 = ginv(hmat);
+ 
   y = cgetg(7,t_VEC);
-  y[1] = lcopy(bnf);
-  y[2] = lcopy(bid);
-  y[3] = lcopy(vecel);
-  y[4] = linv(u1old);
-  y[5] = lcopy(clg); u = cgetg(3,t_VEC);
+  y[1] = (long)bnf;
+  y[2] = (long)bid;
+  y[3] = (long)vecel;
+  y[4] = (long)U;
+  y[5] = (long)clg; u = cgetg(3,t_VEC);
   y[6] = (long)u;
-    u[1] = lmul(u2,p2);
-    u[2] = lmul(u1,p1);
-  return gerepileupto(av,y);
+    u[1] = lmul(u2, ginv(hmat));
+    u[2] = lmul(u1, lllint(u1));
+  return gerepilecopy(av,y);
 }
 
 GEN
@@ -796,7 +789,7 @@ minimforunits(GEN nf, long BORNE, long stockmax)
   GEN u,r,S,a,M,p1;
   double p;
   double **q,*v,*y,*z;
-  double eps=0.000001, BOUND = BORNE + eps;
+  double eps=0.000001, BOUND = BORNE * 1.00001;
 
   if (DEBUGLEVEL>=2)
   {
@@ -2517,7 +2510,7 @@ hnflistdivise(GEN list,GEN h)
 }
 
 static GEN
-subgroupcond(GEN bnr, long indexbound)
+subgroupcond(GEN bnr, GEN indexbound)
 {
   ulong av = avma;
   long i,j,lgi,lp;
@@ -2554,7 +2547,7 @@ subgroupcond(GEN bnr, long indexbound)
 }
 
 GEN
-subgrouplist0(GEN bnr, long indexbound, long all)
+subgrouplist0(GEN bnr, GEN indexbound, long all)
 {
   if (typ(bnr)!=t_VEC) err(typeer,"subgrouplist");
   if (lg(bnr)!=1 && typ(bnr[1])!=t_INT)
