@@ -1,4 +1,4 @@
-/* $Id: buch4.c,v 1.15 2001/10/01 12:11:30 karim Exp $
+/* $Id: buch4.c,v 1.19 2001/11/23 22:36:01 karim Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -361,7 +361,11 @@ hilb2nf(GEN nf,GEN a,GEN b,GEN p)
 {
   ulong av = avma;
   long rep;
-  GEN pol = coefs_to_pol(3, lift(a), zero, lift(b));
+  GEN pol;
+
+  if (typ(a) != t_POLMOD) a = basistoalg(nf, a);
+  if (typ(b) != t_POLMOD) b = basistoalg(nf, b);
+  pol = coefs_to_pol(3, lift(a), zero, lift(b));
   /* varn(nf.pol) = 0, pol is not a valid GEN  [as in Pol([x,x], x)].
    * But it is only used as a placeholder, hence it is not a problem */
 
@@ -457,7 +461,7 @@ nfhilbert0(GEN nf,GEN a,GEN b,GEN p)
 extern GEN isprincipalfact(GEN bnf,GEN P, GEN e, GEN C, long flag);
 extern GEN vconcat(GEN Q1, GEN Q2);
 extern GEN mathnfspec(GEN x, GEN *ptperm, GEN *ptdep, GEN *ptB, GEN *ptC);
-extern GEN factorback_i(GEN fa, GEN nf, int red);
+extern GEN factorback_i(GEN fa, GEN e, GEN nf, int red);
 /* S a list of prime ideal in primedec format. Return res:
  * res[1] = generators of (S-units / units), as polynomials
  * res[2] = [perm, HB, den], for bnfissunit
@@ -472,7 +476,7 @@ bnfsunit(GEN bnf,GEN S,long prec)
   ulong ltop = avma;
   long i,j,ls;
   GEN p1,nf,classgp,gen,M,U,H;
-  GEN sunit,card,sreg,res,pow,fa = cgetg(3, t_MAT);
+  GEN sunit,card,sreg,res,pow;
 
   if (typ(S) != t_VEC) err(typeer,"bnfsunit");
   bnf = checkbnf(bnf); nf=(GEN)bnf[7];
@@ -504,23 +508,19 @@ bnfsunit(GEN bnf,GEN S,long prec)
   card = gun;
   if (lg(H) > 1)
   { /* non trivial (rare!) */
-    GEN SNF, ClS = cgetg(4,t_VEC);
+    GEN D,U, ClS = cgetg(4,t_VEC);
 
-    SNF = smith2(H); p1 = (GEN)SNF[3];
-    card = dethnf_i(p1);
+    D = smithall(H, &U, NULL);
+    card = dethnf_i(D);
     ClS[1] = (long)card; /* h */
-    for(i=1; i<lg(p1); i++)
-      if (gcmp1((GEN)p1[i])) break;
-    setlg(p1,i);
-    ClS[2]=(long)p1; /* cyc */
+    for(i=1; i<lg(D); i++)
+      if (gcmp1((GEN)D[i])) break;
+    setlg(D,i);
+    ClS[2]=(long)D; /* cyc */
 
-    p1=cgetg(i,t_VEC); pow=ZM_inv((GEN)SNF[1],gun);
-    fa[1] = (long)gen;
+    p1=cgetg(i,t_VEC); pow=ZM_inv(U,gun);
     for(i--; i; i--)
-    {
-      fa[2] = pow[i];
-      p1[i] = (long)factorback_i(fa, nf, 1);
-    }
+      p1[i] = (long)factorback_i(gen, (GEN)pow[i], nf, 1);
     ClS[3]=(long)p1; /* gen */
     res[5]=(long) ClS;
   }
@@ -546,7 +546,7 @@ bnfsunit(GEN bnf,GEN S,long prec)
     Sperm = cgetg(ls, t_VEC); sunit = cgetg(ls, t_VEC);
     for (i=1; i<ls; i++) Sperm[i] = S[perm[i]]; /* S o perm */
 
-    setlg(Sperm, lH); fa[1] = (long)Sperm;
+    setlg(Sperm, lH);
     for (i=1; i<lH; i++)
       sunit[i] = isprincipalfact(bnf,Sperm,(GEN)H[i],NULL,fl)[2];
     for (j=1; j<lB; j++,i++)
