@@ -1,4 +1,4 @@
-/* $Id: gen1.c,v 1.36 2001/10/13 15:44:32 karim Exp $
+/* $Id: gen1.c,v 1.41 2001/11/27 22:06:01 karim Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -235,84 +235,101 @@ gsub(GEN x, GEN y)
 static GEN
 addpadic(GEN x, GEN y)
 {
-  long c,e,r,d,r1,r2,av,tetpil;
-  GEN z,p1,p2, p = (GEN)x[2];
+  ulong av = avma;
+  long c,d,e,r,rx,ry;
+  GEN u,z,p,mod;
 
-  z=cgetg(5,t_PADIC); icopyifstack(p, z[2]); av=avma;
-  e=valp(x); r=valp(y); d = r-e;
-  if (d<0) { p1=x; x=y; y=p1; e=r; d = -d; }
-  r1=precp(x); r2=precp(y);
-  if (d)
+  (void)new_chunk(5+lgefint(x[3])+lgefint(y[3]));
+  e = valp(x);
+  r = valp(y); d = r-e;
+  if (d < 0) { GEN p1=x; x=y; y=p1; e = r; d = -d; }
+  rx = precp(x); p = (GEN)x[2];
+  ry = precp(y);
+  if (d) /* v(x) < v(y) */
   {
-    r = d+r2;
-    p1 = (d==1)? p: gclone(gpuigs(p,d));
-    avma=av;
-    if (r<r1) z[3]=lmulii(p1,(GEN)y[3]);
-    else
+    r = d+ry; z = gpowgs(p,d);
+    if (r < rx) mod = mulii(z,(GEN)y[3]); else { r = rx; mod = (GEN)x[3]; }
+    u = addii((GEN)x[4], mulii(z,(GEN)y[4]));
+  }
+  else
+  {
+    if (ry < rx) { r=ry; mod = (GEN)x[3]; } else { r=rx; mod = (GEN)y[3]; }
+    u = addii((GEN)x[4], (GEN)y[4]);
+    if (!signe(u) || (c = pvaluation(u,p,&u)) >= r)
     {
-      r=r1; z[3]=licopy((GEN)x[3]);
+      avma = av; return padiczero(p, e+r);
     }
-    av=avma; p2=mulii(p1,(GEN)y[4]);
-    if (d!=1) gunclone(p1);
-    p1=addii(p2,(GEN)x[4]); tetpil=avma;
-    z[4]=lpile(av,tetpil, modii(p1,(GEN)z[3]));
-    z[1]=evalprecp(r) | evalvalp(e); return z;
-  }
-  if (r2<r1) { r=r2; p1=x; x=y; y=p1; } else r=r1;
-  p1 = addii((GEN)x[4],(GEN)y[4]);
-  if (!signe(p1) || (c = pvaluation(p1,p,&p2)) >=r)
-  {
-    avma=av; z[4]=zero; z[3]=un;
-    z[1]=evalvalp(e+r); return z;
-  }
-  if (c)
-  {
-    p2=gclone(p2); avma=av;
-    if (c==1)
-      z[3] = ldivii((GEN)x[3], p);
-    else
+    if (c)
     {
-      p1 = gpuigs(p,c); tetpil=avma;
-      z[3] = lpile(av,tetpil, divii((GEN)x[3], p1));
+      mod = divii(mod, gpowgs(p,c));
+      r -= c;
+      e += c;
     }
-    z[4]=lmodii(p2,(GEN)z[3]); gunclone(p2);
-    z[1]=evalprecp(r-c) | evalvalp(e+c); return z;
   }
-  tetpil=avma;
-  z[4]=lpile(av,tetpil,modii(p1,(GEN)x[3]));
-  z[3]=licopy((GEN)x[3]);
-  z[1]=evalprecp(r) | evalvalp(e); return z;
+  avma = av; z = cgetg(5,t_PADIC);
+  z[1] = evalprecp(r) | evalvalp(e);
+  z[3] = licopy(mod);
+  z[4] = lmodii(u,(GEN)z[3]);
+  icopyifstack(p, z[2]); return z;
 }
 
 /* return x + y, where x is t_INT or t_FRAC(N), y t_PADIC */
 static GEN
 gaddpex(GEN x, GEN y)
 {
-  long tx,e1,e2,e3,av,tetpil;
-  GEN z,p,p1,p2;
+  ulong av;
+  long tx,d,r,e;
+  GEN z,q,p,p1,p2,mod,u;
 
   if (gcmp0(x)) return gcopy(y);
 
-  av=avma; p=(GEN)y[2]; tx=typ(x);
-  z=cgetg(5,t_PADIC); z[2]=(long)p;
-  e3 = (tx == t_INT)? pvaluation(x,p,&p1)
+  av = avma; p = (GEN)y[2]; tx = typ(x);
+  e = (tx == t_INT)? pvaluation(x,p,&p1)
                     : pvaluation((GEN)x[1],p,&p1) -
                       pvaluation((GEN)x[2],p,&p2);
-  e1 = valp(y)-e3; e2 = signe(y[4])? e1+precp(y): e1;
-  if (e2<=0)
+  d = valp(y)-e; r = d+precp(y);
+  if (r<=0) { avma = av; return gcopy(y); }
+  mod = (GEN)y[3];
+  u   = (GEN)y[4];
+  (void)new_chunk(5 + lgefint(mod) + lgefint(p)*labs(d));
+
+  if (d > 0)
   {
-    z[1] = evalprecp(0) | evalvalp(e3);
-    z[3] = un;
-    z[4] = zero;
+    q = gpowgs(p,d);
+    mod = mulii(mod, q);
+    u   = mulii(u, q);
+    if (tx != t_INT && !is_pm1(p2)) p1 = mulii(p1, mpinvmod(p2,mod));
+    u = addii(u, p1);
+  }
+  else if (d < 0)
+  {
+    q = gpowgs(p,-d);
+    mod = divii(mod, q);
+    u   = modii(u, mod);
+    if (tx != t_INT && !is_pm1(p2)) p1 = mulii(p1, mpinvmod(p2,mod));
+    p1 = mulii(p1, q);
+    u = addii(u, p1);
   }
   else
   {
-    if (tx != t_INT && !is_pm1(p2)) p1 = gdiv(p1,p2);
-    z[1] = evalprecp(e2) | evalvalp(e3);
-    z[3] = e1? lmul((GEN)y[3], gpuigs(p,e1)): y[3];
-    z[4] = lmod(p1,(GEN)z[3]);
+    long c;
+    u = addii(u, p1);
+    if (!signe(u) || (c = pvaluation(u,p,&u)) >= r)
+    {
+      avma = av; return padiczero(p,e+r);
+    }
+    if (c)
+    {
+      mod = divii(mod, gpowgs(p,c));
+      r -= c;
+      e += c;
+    }
   }
-  tetpil=avma; return gerepile(av,tetpil,addpadic(z,y));
+  avma = av; z = cgetg(5,t_PADIC);
+  z[1] = evalprecp(r) | evalvalp(e);
+  z[3] = licopy(mod);
+  z[4] = lmodii(u,(GEN)z[3]);
+  icopyifstack(p, z[2]); return z;
 }
 
 static long
@@ -386,10 +403,10 @@ addrfrac(GEN x, GEN y)
   x2 = gdeuc(x2,delta);
   y2 = gdeuc(y2,delta);
   n = gadd(gmul(x1,y2), gmul(y1,x2));
-  if (!signe(n)) return gerepileupto((long)(z+3), n);
+  if (gcmp0(n)) return gerepileupto((long)(z+3), n);
   tetpil = avma; d = gmul(x2, y2);
   p1 = poldivres(n, delta, &p2); /* we want gcd(n,delta) */
-  if (!signe(p2))
+  if (gcmp0(p2))
   {
     if (lgef(d) == 3) /* "constant" denominator */
     {
@@ -529,7 +546,7 @@ gadd(GEN x, GEN y)
 	    p1=co8(y,lg(x)+i); tetpil=avma;
 	    return gerepile(av,tetpil,gadd(p1,x));
 	
-	  case t_INTMOD: case t_PADIC: err(operf,"+",tx,ty);
+	  case t_INTMOD: case t_PADIC: err(operf,"+",x,y);
 	}
 	
       case t_INTMOD:
@@ -616,7 +633,7 @@ gadd(GEN x, GEN y)
             return gerepile(av,tetpil,gadd(p1,y));
 
 	  case t_QUAD:
-	    lx=precision(x); if (!lx) err(operi,"+",tx,ty);
+	    lx=precision(x); if (!lx) err(operi,"+",x,y);
 	    if (gcmp0(y)) return gcopy(x);
 
 	    av=avma; i=gexpo(y)-gexpo(x);
@@ -629,7 +646,7 @@ gadd(GEN x, GEN y)
         switch(ty)
 	{
 	  case t_PADIC:
-            if (!egalii((GEN)x[2],(GEN)y[2])) err(operi,"+",tx,ty);
+            if (!egalii((GEN)x[2],(GEN)y[2])) err(operi,"+",x,y);
             return addpadic(x,y);
 
 	  case t_QUAD:
@@ -646,7 +663,7 @@ gadd(GEN x, GEN y)
 	}
 
       case t_QUAD: z=cgetg(4,t_QUAD); k=x[1]; l=y[1];
-        if (!gegal((GEN)k,(GEN)l)) err(operi,"+",tx,ty);
+        if (!gegal((GEN)k,(GEN)l)) err(operi,"+",x,y);
         copyifstack(l, z[1]);
         z[2]=ladd((GEN)x[2],(GEN)y[2]);
         z[3]=ladd((GEN)x[3],(GEN)y[3]); return z;
@@ -733,9 +750,9 @@ gadd(GEN x, GEN y)
 	if (isexactzero(x)) return gcopy(y);
 	if (ty == t_MAT) return gaddmat(x,y);
         /* fall through */
-      case t_QFR: case t_QFI: err(operf,"+",tx,ty);
+      case t_QFR: case t_QFI: err(operf,"+",x,y);
     }
-    err(operf,"+",tx,ty);
+    err(operf,"+",x,y);
   }
 
   /* here !isscalar(x) && isscalar(y) && (vx=vy || ismatvec(x and y)) */
@@ -769,7 +786,7 @@ gadd(GEN x, GEN y)
           z[1]=lpile(av,tetpil, gadd(p1,(GEN)y[1]));
           z[2]=lcopy((GEN)y[2]); return z;
 
-	default: err(operf,"+",tx,ty);
+	default: err(operf,"+",x,y);
       }
 
     case t_SER:
@@ -827,14 +844,14 @@ gadd(GEN x, GEN y)
           p1 = gdiv((GEN)y[1], p1); tetpil=avma;
           return gerepile(av,tetpil,gadd(p1,x));
 
-	default: err(operf,"+",tx,ty);
+	default: err(operf,"+",x,y);
       }
 
     case t_RFRAC:
-      if (!is_rfrac_t(ty)) err(operi,"+",tx,ty);
+      if (!is_rfrac_t(ty)) err(operi,"+",x,y);
       return addrfrac(x,y);
     case t_RFRACN:
-      if (!is_rfrac_t(ty)) err(operi,"+",tx,ty);
+      if (!is_rfrac_t(ty)) err(operi,"+",x,y);
       z=cgetg(3,t_RFRACN); av=avma;
       p1=gmul((GEN)x[1],(GEN)y[2]);
       p2=gmul((GEN)x[2],(GEN)y[1]); tetpil=avma;
@@ -843,13 +860,13 @@ gadd(GEN x, GEN y)
 
     case t_VEC: case t_COL: case t_MAT:
       lx = lg(x); ly = lg(y);
-      if (lx!=ly || tx!=ty) err(operi,"+",tx,ty);
+      if (lx!=ly || tx!=ty) err(operi,"+",x,y);
       z=cgetg(ly,ty);
       for (i=1; i<ly; i++)
 	z[i]=ladd((GEN)x[i],(GEN)y[i]);
       return z;
   }
-  err(operf,"+",tx,ty);
+  err(operf,"+",x,y);
   return NULL; /* not reached */
 }
 
@@ -1040,11 +1057,16 @@ gmul_err(GEN x, GEN y, long tx, long ty)
       case t_QFR: return compreal(x,y);
       case t_VECSMALL:
         l = lg(x); z = cgetg(l, t_VECSMALL);
-        if (l != lg(y)) err(operf,"*",tx,ty);
-        for (i=1; i<l; i++) z[i]=x[y[i]];
+        if (l != lg(y)) err(operf,"*",x,y);
+        for (i=1; i<l; i++) 
+	{
+	  long yi=y[i];
+	  if (yi<1 || yi>=l) err(operf,"*",x,y);
+	  z[i]=x[y[i]];
+	}
         return z;
     }
-  err(operf,"*",tx,ty);
+  err(operf,"*",x,y);
   return NULL; /* not reached */
 }
 
@@ -1129,7 +1151,7 @@ gmul(GEN x, GEN y)
 	    l=avma; p1=co8(y,lg(x)); tetpil=avma;
 	    return gerepile(l,tetpil,gmul(p1,x));
 	
-	  default: err(operf,"*",tx,ty);
+	  default: err(operf,"*",x,y);
 	}
 	
       case t_INTMOD:
@@ -1239,7 +1261,7 @@ gmul(GEN x, GEN y)
             return gerepile(av,tetpil,gmul(p1,y));
 	
 	  case t_QUAD:
-	    lx=precision(x); if (!lx) err(operi,"*",tx,ty);
+	    lx=precision(x); if (!lx) err(operi,"*",x,y);
 	    l=avma; p1=co8(y,lx); tetpil=avma;
 	    return gerepile(l,tetpil,gmul(p1,x));
 	}
@@ -1248,7 +1270,7 @@ gmul(GEN x, GEN y)
         switch(ty)
 	{
 	  case t_PADIC:
-	    if (!egalii((GEN)x[2],(GEN)y[2])) err(operi,"*",tx,ty);
+	    if (!egalii((GEN)x[2],(GEN)y[2])) err(operi,"*",x,y);
             l = valp(x)+valp(y);
 	    if (!signe(x[4])) { z=gcopy(x); setvalp(z,l); return z; }
 	    if (!signe(y[4])) { z=gcopy(y); setvalp(z,l); return z; }
@@ -1273,7 +1295,7 @@ gmul(GEN x, GEN y)
 	
       case t_QUAD: z=cgetg(4,t_QUAD);
         p1=(GEN)x[1]; p2=(GEN)y[1];
-        if (!gegal(p1,p2)) err(operi,"*",tx,ty);
+        if (!gegal(p1,p2)) err(operi,"*",x,y);
 
         copyifstack(p2, z[1]); l=avma;
         p2=gmul((GEN)x[2],(GEN)y[2]);
@@ -1327,7 +1349,7 @@ gmul(GEN x, GEN y)
         switch(ty)
         {
           case t_COL:
-            if (lx!=ly) err(operi,"*",tx,ty);
+            if (lx!=ly) err(operi,"*",x,y);
             z=gzero; l=avma;
             for (i=1; i<lx; i++)
             {
@@ -1338,7 +1360,7 @@ gmul(GEN x, GEN y)
 
           case t_MAT:
             if (ly==1) return cgetg(1,t_VEC);
-            l=lg(y[1]); if (lx!=l) err(operi,"*",tx,ty);
+            l=lg(y[1]); if (lx!=l) err(operi,"*",x,y);
 
             z=cgetg(ly,tx);
             for (i=1; i<ly; i++)
@@ -1353,7 +1375,7 @@ gmul(GEN x, GEN y)
             }
             return z;
 
-          default: err(operf,"*",tx,ty);
+          default: err(operf,"*",x,y);
         }
 
       case t_COL:
@@ -1364,32 +1386,32 @@ gmul(GEN x, GEN y)
             for (i=1; i<ly; i++)
             {
               p1 = gmul((GEN)y[i],x);
-              if (typ(p1) != t_COL) err(operi,"*",tx,ty);
+              if (typ(p1) != t_COL) err(operi,"*",x,y);
               z[i]=(long)p1;
             }
             return z;
 
           case t_MAT:
-            if (ly!=1 && lg(y[1])!=2) err(operi,"*",tx,ty);
+            if (ly!=1 && lg(y[1])!=2) err(operi,"*",x,y);
 
             z=cgetg(ly,t_MAT);
             for (i=1; i<ly; i++) z[i]=lmul(gcoeff(y,1,i),x);
             return z;
 
-          default: err(operf,"*",tx,ty);
+          default: err(operf,"*",x,y);
         }
 
       case t_MAT:
         switch(ty)
         {
           case t_VEC:
-            if (lx!=2) err(operi,"*",tx,ty);
+            if (lx!=2) err(operi,"*",x,y);
             z=cgetg(ly,t_MAT);
             for (i=1; i<ly; i++) z[i]=lmul((GEN)y[i],(GEN)x[1]);
             return z;
 
           case t_COL:
-            if (lx!=ly) err(operi,"*",tx,ty);
+            if (lx!=ly) err(operi,"*",x,y);
             if (lx==1) return gcopy(y);
 
             lx=lg(x[1]); z=cgetg(lx,t_COL);
@@ -1407,7 +1429,7 @@ gmul(GEN x, GEN y)
 
           case t_MAT:
             if (ly==1) return cgetg(ly,t_MAT);
-            if (lx != lg(y[1])) err(operi,"*",tx,ty);
+            if (lx != lg(y[1])) err(operi,"*",x,y);
             z=cgetg(ly,t_MAT);
             if (lx==1)
             {
@@ -1440,7 +1462,7 @@ gmul(GEN x, GEN y)
   {
     if (isexactzero(x)) 
     {
-      if (vy == BIGINT) err(operf,"*",tx,ty);
+      if (vy == BIGINT) err(operf,"*",x,y);
       return zeropol(vy);
     }
     if (tx == t_INT && is_pm1(x))
@@ -1468,7 +1490,7 @@ gmul(GEN x, GEN y)
       case t_RFRACN: av=avma; z=cgetg(3,t_RFRACN);
         z[1]=lmul(x,(GEN)y[1]);
         z[2]=lcopy((GEN)y[2]); return z;
-      default: err(operf,"*",tx,ty);
+      default: err(operf,"*",x,y);
     }
   }
 
@@ -1522,7 +1544,7 @@ gmul(GEN x, GEN y)
           z[1]=lmul(x,(GEN)y[1]);
           z[2]=lcopy((GEN)y[2]); return z;
 	
-	default: err(operf,"*",tx,ty);
+	default: err(operf,"*",x,y);
       }
 	
     case t_SER:
@@ -1554,13 +1576,13 @@ gmul(GEN x, GEN y)
 	  l=avma; p1=gmul((GEN)y[1],x); tetpil=avma;
           return gerepile(l,tetpil,gdiv(p1,(GEN)y[2]));
 	
-	default: err(operf,"*",tx,ty);
+	default: err(operf,"*",x,y);
       }
 	
     /* (tx,ty) == t_RFRAC <==> ty == t_RFRAC */
     case t_RFRAC: return mulrfrac(x,y);
     case t_RFRACN:
-      if (!is_rfrac_t(ty)) err(operf,"*",tx,ty);
+      if (!is_rfrac_t(ty)) err(operf,"*",x,y);
       av=avma; z=cgetg(3,ty);
       z[1]=lmul((GEN)x[1],(GEN)y[1]);
       z[2]=lmul((GEN)x[2],(GEN)y[2]); return z;
@@ -1691,7 +1713,7 @@ gsqr(GEN x)
     case t_MAT:
       lx=lg(x);
       if (lx==1) return cgetg(1,tx);
-      if (lx != lg(x[1])) err(operi,"*",tx,tx);
+      if (lx != lg(x[1])) err(operi,"*",x,x);
       z=cgetg(lx,tx);
       for (j=1; j<lx; j++)
       {
@@ -1843,7 +1865,7 @@ gdiv(GEN x, GEN y)
 	    l=avma; p1=co8(y,lg(x)); tetpil=avma;
 	    return gerepile(l,tetpil,gdiv(x,p1));
 
-	  case t_INTMOD: case t_PADIC: err(operf,"/",tx,ty);
+	  case t_INTMOD: case t_PADIC: err(operf,"/",x,y);
 	}
 
       case t_INTMOD:
@@ -1881,7 +1903,7 @@ gdiv(GEN x, GEN y)
 	    l=avma; p1=cgetg(3,t_INTMOD); p1[1]=x[1]; p1[2]=lgeti(lg(x[1]));
 	    gaffect(y,p1); tetpil=avma; return gerepile(l,tetpil,gdiv(x,p1));
 
-	  case t_REAL: err(operf,"/",tx,ty);
+	  case t_REAL: err(operf,"/",x,y);
 	}
 
       case t_FRAC: case t_FRACN:
@@ -1991,7 +2013,7 @@ gdiv(GEN x, GEN y)
 	    return gerepile(av,tetpil,gdiv(p1,y));
 
 	  case t_QUAD:
-	    lx=precision(x); if (!lx) err(operi,"/",tx,ty);
+	    lx=precision(x); if (!lx) err(operi,"/",x,y);
 	    l=avma; p1=co8(y,lx); tetpil=avma;
 	    return gerepile(l,tetpil,gdiv(x,p1));
 	}
@@ -2012,7 +2034,7 @@ gdiv(GEN x, GEN y)
 	    return gerepile(l,tetpil,gdiv(p1,y));
 
 	  case t_PADIC:
-	    if (!egalii((GEN)x[2],(GEN)y[2])) err(operi,"/",tx,ty);
+	    if (!egalii((GEN)x[2],(GEN)y[2])) err(operi,"/",x,y);
 	    if (!signe(x[4]))
 	    {
 	      z=gcopy(x); setvalp(z,valp(x)-valp(y));
@@ -2031,7 +2053,7 @@ gdiv(GEN x, GEN y)
 	    return gerepile(l,tetpil,gdiv(p1,p2));
 
 	  case t_REAL:
-	    err(operf,"/",tx,ty);
+	    err(operf,"/",x,y);
 	}
 
       case t_QUAD:
@@ -2052,13 +2074,13 @@ gdiv(GEN x, GEN y)
 	    tetpil=avma; return gerepile(l,tetpil,gdiv(p1,y));
 
 	  case t_COMPLEX:
-	    ly=precision(y); if (!ly) err(operi,"/",tx,ty);
+	    ly=precision(y); if (!ly) err(operi,"/",x,y);
 	    l=avma; p1=co8(x,ly); tetpil=avma;
 	    return gerepile(l,tetpil,gdiv(p1,y));
 
 	  case t_QUAD:
 	    k=x[1]; l=y[1];
-	    if (!gegal((GEN)k,(GEN)l)) err(operi,"/",tx,ty);
+	    if (!gegal((GEN)k,(GEN)l)) err(operi,"/",x,y);
 	    l=avma; p1=gnorm(y); p2=gmul(x,gconj(y)); tetpil=avma;
 	    return gerepile(l,tetpil,gdiv(p2,p1));
 	}
@@ -2085,7 +2107,7 @@ gdiv(GEN x, GEN y)
     }
     return gerepileupto(av, gmul(x, ginv(y)));
   }
-  if (is_noncalc_t(tx) || is_noncalc_t(ty)) err(operf,"/",tx, ty);
+  if (is_noncalc_t(tx) || is_noncalc_t(ty)) err(operf,"/",x,y);
   /* now x and y are not both is_scalar_t */
 
   lx = lg(x);
@@ -2107,7 +2129,7 @@ gdiv(GEN x, GEN y)
         for (i=lontyp[tx]; i<lx; i++) z[i]=ldiv((GEN)x[i],y);
         return z;
     }
-    err(operf,"/",tx,ty);
+    err(operf,"/",x,y);
   }
 
   ly=lg(y); 
@@ -2141,9 +2163,9 @@ gdiv(GEN x, GEN y)
 	l=avma; p1=invmat(y); tetpil=avma;
 	return gerepile(l,tetpil,gmul(x,p1));
 
-      case t_VEC: case t_COL: err(operf,"/",tx,ty);
+      case t_VEC: case t_COL: err(operf,"/",x,y);
     }
-    err(operf,"/",tx,ty);
+    err(operf,"/",x,y);
   }
 
   /* ici vx=vy et tx>=10 et ty>=10*/
@@ -2167,7 +2189,7 @@ gdiv(GEN x, GEN y)
 	  z[1]=lmul(x,(GEN)y[2]);
 	  z[2]=lcopy((GEN)y[1]); return z;
 
-	default: err(operf,"/",tx,ty);
+	default: err(operf,"/",x,y);
       }
 
     case t_SER:
@@ -2230,7 +2252,7 @@ gdiv(GEN x, GEN y)
 	  l=avma; p2=gmul(x,(GEN)y[2]); tetpil=avma;
 	  return gerepile(l,tetpil,gdiv(p2,(GEN)y[1]));
 
-	default: err(operf,"/",tx,ty);
+	default: err(operf,"/",x,y);
       }
 
     case t_RFRAC: case t_RFRACN:
@@ -2253,7 +2275,7 @@ gdiv(GEN x, GEN y)
 	  z[1]=lmul((GEN)x[1],(GEN)y[2]);
           z[2]=lmul((GEN)x[2],(GEN)y[1]); return z;
 
-	default: err(operf,"/",tx,ty);
+	default: err(operf,"/",x,y);
       }
 
     case t_VEC: case t_COL: case t_MAT:
@@ -2263,12 +2285,12 @@ gdiv(GEN x, GEN y)
 	for (i=1; i<lx; i++) z[i]=ldiv((GEN)x[i],y);
 	return z;
       }
-      if (ty!=t_MAT || ly==1 || lg(y[1])!=ly) err(operi,"/",tx,ty);
+      if (ty!=t_MAT || ly==1 || lg(y[1])!=ly) err(operi,"/",x,y);
       l=avma; p1=invmat(y); tetpil=avma;
       return gerepile(l,tetpil,gmul(x,p1));
      case t_QFI:case t_QFR:
        break;
-     default: err(operf,"/",tx,ty);
+     default: err(operf,"/",x,y);
   }
   /*Here tx==t_QFI || tx==t_QFR*/
   if (tx==ty)
@@ -2283,6 +2305,6 @@ gdiv(GEN x, GEN y)
         setsigne(y[2],l); setsigne(y[4],k); return z;
     }
   }
-  err(operf,"/",tx,ty);
+  err(operf,"/",x,y);
   return NULL; /* not reached */
 }
