@@ -1,4 +1,4 @@
-/* $Id: gen2.c,v 1.35 2001/10/17 22:20:11 karim Exp $
+/* $Id: gen2.c,v 1.41 2002/01/10 19:41:57 karim Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -403,6 +403,23 @@ gcmp(GEN x, GEN y)
   av=avma; y=gneg_i(y); f=gsigne(gadd(x,y)); avma=av; return f;
 }
 
+static int
+lexcmp_scal_vec(GEN x, GEN y)
+{
+  long fl;
+  if (lg(y)==1) return 1;
+  fl = lexcmp(x,(GEN)y[1]);
+  if (fl) return fl;
+  return -1;
+}
+
+static int
+lexcmp_vec_mat(GEN x, GEN y)
+{
+  if (lg(x)==1) return -1;
+  return lexcmp_scal_vec(x,y);
+}
+
 /* as gcmp for vector/matrices, using lexicographic ordering on components */
 int
 lexcmp(GEN x, GEN y)
@@ -410,52 +427,33 @@ lexcmp(GEN x, GEN y)
   const long tx=typ(x), ty=typ(y);
   long lx,ly,l,fl,i;
 
-  ly=lg(y);
   if (!is_matvec_t(tx))
   {
     if (!is_matvec_t(ty)) return gcmp(x,y);
-    if (ly==1) return 1;
-    fl = lexcmp(x,(GEN)y[1]);
-    if (fl) return fl;
-    return (ly>2)? -1:0;
+    return  lexcmp_scal_vec(x,y);
   }
-
-  lx=lg(x);
   if (!is_matvec_t(ty))
-  {
-    if (lx==1) return -1;
-    fl = lexcmp(y,(GEN)x[1]);
-    if (fl) return -fl;
-    return (lx>2)? 1:0;
-  }
+    return -lexcmp_scal_vec(y,x);
 
   /* x and y are matvec_t */
-  if (ly==1) return (lx==1)?0:1;
-  if (lx==1) return -1;
   if (ty==t_MAT)
   {
     if (tx != t_MAT)
-    {
-      fl = lexcmp(x,(GEN)y[1]);
-      if (fl) return fl;
-      return (ly>2)?-1:0;
-    }
+      return lexcmp_vec_mat(x,y);
   }
   else if (tx==t_MAT)
-  {
-    fl = lexcmp(y,(GEN)x[1]);
-    if (fl) return -fl;
-    return (ly>2)?1:0;
-  }
-
+    return -lexcmp_vec_mat(y,x);
+   
   /* tx = ty = t_MAT, or x and y are both vect_t */
-  l=min(lx,ly);
+  lx = lg(x);
+  ly = lg(y); l = min(lx,ly);
   for (i=1; i<l; i++)
   {
     fl = lexcmp((GEN)x[i],(GEN)y[i]);
     if (fl) return fl;
   }
-  return (ly != lx)? -1 : 0;
+  if (lx == ly) return 0;
+  return (lx < ly)? -1 : 1;
 }
 
 /*****************************************************************/
@@ -587,6 +585,29 @@ minval(GEN x, GEN p, long first, long lx)
   return val;
 }
 
+long 
+polvaluation(GEN x, GEN *Z)
+{
+  long v;
+
+  if (!signe(x)) { if (Z) *Z = zeropol(varn(x)); return VERYBIGINT; }
+  for (v = 0;; v++)
+    if (!isexactzero((GEN)x[2+v])) break;
+  if (Z)
+  {
+    if (!v) *Z = x;
+    else
+    {
+      long i, lz = lgef(x)-v;
+      GEN z = cgetg(lz, t_POL);
+      z[1] = x[1]; setlgef(z,lz); x += v;
+      for (i=2; i<lz; i++) z[i] = x[i];
+      *Z = z;
+    }
+  }
+  return v;
+}
+
 long
 ggval(GEN x, GEN p)
 {
@@ -630,10 +651,7 @@ ggval(GEN x, GEN p)
 	if (vx == v)
 	{
 	  if ((p>=(GEN)polx && p <= (GEN)(polx+MAXVARN)) || ismonome(p))
-          {
-            i=2; while (isexactzero((GEN)x[i])) i++;
-            return i-2;
-          }
+            return polvaluation(x, NULL);
 	  av = avma; limit=stack_lim(av,1);
 	  for (val=0; ; val++)
 	  {
@@ -1060,10 +1078,10 @@ gaffsg(long s, GEN x)
       gaffsg(s,(GEN)x[1]); gaffsg(1,(GEN)x[2]); break;
 
     case t_VEC: case t_COL: case t_MAT:
-      if (lg(x)!=2) err(operi,"",t_INT,typ(x));
+      if (lg(x)!=2) err(operi,"",stoi(s),x);
       gaffsg(s,(GEN)x[1]); break;
 
-    default: err(operf,"",t_INT,typ(x));
+    default: err(operf,"",stoi(s),x);
   }
 }
 
@@ -1131,7 +1149,7 @@ gaffect(GEN x, GEN y)
 	    case t_POLMOD:
 	      gaffect(x,(GEN)y[2]); break;
 
-	    default: err(operf,"",tx,ty);
+	    default: err(operf,"",x,y);
 	  }
 	  break;
 	
@@ -1149,9 +1167,9 @@ gaffect(GEN x, GEN y)
 
 	    case t_INT: case t_INTMOD: case t_FRAC:
 	    case t_FRACN: case t_PADIC: case t_QUAD:
-              err(operf,"",tx,ty);
+              err(operf,"",x,y);
 
-	    default: err(operf,"",tx,ty);
+	    default: err(operf,"",x,y);
 	  }
 	  break;
 	
@@ -1160,13 +1178,13 @@ gaffect(GEN x, GEN y)
 	  {
 	    case t_INTMOD:
 	      if (!divise((GEN)x[1],(GEN)y[1]))
-                err(operi,"",tx,ty);
+                err(operi,"",x,y);
 	      modiiz((GEN)x[2],(GEN)y[1],(GEN)y[2]); break;
 
 	    case t_POLMOD:
 	      gaffect(x,(GEN)y[2]); break;
 
-	    default: err(operf,"",tx,ty);
+	    default: err(operf,"",x,y);
 	  }
 	  break;
 
@@ -1175,7 +1193,7 @@ gaffect(GEN x, GEN y)
 	  {
 	    case t_INT:
 	      if (! mpdivis((GEN)x[1],(GEN)x[2],y))
-                err(operi,"",tx,ty);
+                err(operi,"",x,y);
 	      break;
 
 	    case t_REAL:
@@ -1207,7 +1225,7 @@ gaffect(GEN x, GEN y)
 	      gaffect(x,(GEN)y[2]); gaffsg(0,(GEN)y[3]); break;
 	    case t_POLMOD:
 	      gaffect(x,(GEN)y[2]); break;
-	    default: err(operf,"",tx,ty);
+	    default: err(operf,"",x,y);
 	  }
 	  break;
 
@@ -1215,7 +1233,7 @@ gaffect(GEN x, GEN y)
 	  switch(ty)
 	  {
 	    case t_INT:
-	      if (! mpdivis((GEN)x[1],(GEN)x[2],y)) err(operi,"",tx,ty);
+	      if (! mpdivis((GEN)x[1],(GEN)x[2],y)) err(operi,"",x,y);
 	      break;
 
 	    case t_REAL:
@@ -1242,7 +1260,7 @@ gaffect(GEN x, GEN y)
 	      gaffect(x,(GEN)y[2]); gaffsg(0,(GEN)y[3]); break;
 	    case t_POLMOD:
 	      gaffect(x,(GEN)y[2]); break;
-	    default: err(operf,"",tx,ty);
+	    default: err(operf,"",x,y);
 	  }
 	  break;
 	
@@ -1251,7 +1269,7 @@ gaffect(GEN x, GEN y)
 	  {
 	    case t_INT: case t_REAL: case t_INTMOD:
 	    case t_FRAC: case t_FRACN: case t_PADIC: case t_QUAD:
-	      if (!gcmp0((GEN)x[2])) err(operi,"",tx,ty);
+	      if (!gcmp0((GEN)x[2])) err(operi,"",x,y);
 	      gaffect((GEN)x[1],y); break;
 	
 	    case t_COMPLEX:
@@ -1262,7 +1280,7 @@ gaffect(GEN x, GEN y)
 	    case t_POLMOD:
 	      gaffect(x,(GEN)y[2]); break;
 
-	    default: err(operf,"",tx,ty);
+	    default: err(operf,"",x,y);
 	  }
 	  break;
 	
@@ -1270,22 +1288,22 @@ gaffect(GEN x, GEN y)
 	  switch(ty)
 	  {
 	    case t_INTMOD:
-	      if (valp(x)<0) err(operi,"",tx,ty);
+	      if (valp(x)<0) err(operi,"",x,y);
 	      av=avma;
 	      v = pvaluation((GEN)y[1],(GEN)x[2],&p1);
               k = signe(x[4])? (precp(x)+valp(x)): valp(x)+1;
-	      if (!gcmp1(p1) || v > k) err(operi,"",tx,ty);
+	      if (!gcmp1(p1) || v > k) err(operi,"",x,y);
 	      modiiz(gtrunc(x),(GEN)y[1],(GEN)y[2]); avma=av; break;
 	
 	    case t_PADIC:
-	      if (!egalii((GEN)x[2],(GEN)y[2])) err(operi,"",tx,ty);
+	      if (!egalii((GEN)x[2],(GEN)y[2])) err(operi,"",x,y);
 	      modiiz((GEN)x[4],(GEN)y[3],(GEN)y[4]);
 	      setvalp(y,valp(x)); break;
 	
 	    case t_POLMOD:
 	      gaffect(x,(GEN)y[2]); break;
 
-	    default: err(operf,"",tx,ty);
+	    default: err(operf,"",x,y);
 	  }
 	  break;
 	
@@ -1294,7 +1312,7 @@ gaffect(GEN x, GEN y)
 	  {
 	    case t_INT: case t_INTMOD: case t_FRAC:
 	    case t_FRACN: case t_PADIC:
-	      if (!gcmp0((GEN)x[3])) err(operi,"",tx,ty);
+	      if (!gcmp0((GEN)x[3])) err(operi,"",x,y);
 	      gaffect((GEN)x[2],y); break;
 
 	    case t_REAL:
@@ -1307,28 +1325,28 @@ gaffect(GEN x, GEN y)
 	      }
 	      else
 	      {
-                if (!gcmp0((GEN)x[3])) err(operi,"",tx,ty);
+                if (!gcmp0((GEN)x[3])) err(operi,"",x,y);
 		gaffect((GEN)x[2],y);
 	      }
 	      break;
 
 	    case t_QUAD:
-	      if (! gegal((GEN)x[1],(GEN)y[1])) err(operi,"",tx,ty);
+	      if (! gegal((GEN)x[1],(GEN)y[1])) err(operi,"",x,y);
 	      affii((GEN)x[2],(GEN)y[2]);
 	      affii((GEN)x[3],(GEN)y[3]);
 	      break;
 	    case t_POLMOD:
 	      gaffect(x,(GEN)y[2]); break;
-	    default: err(operf,"",tx,ty);
+	    default: err(operf,"",x,y);
 	  }
 	  break;
 
 	case t_POLMOD:
-	  if (ty!=t_POLMOD) err(operf,"",tx,ty);
-	  if (! gdivise((GEN)x[1],(GEN)y[1])) err(operi,"",tx,ty);
+	  if (ty!=t_POLMOD) err(operf,"",x,y);
+	  if (! gdivise((GEN)x[1],(GEN)y[1])) err(operi,"",x,y);
 	  gmodz((GEN)x[2],(GEN)y[1],(GEN)y[2]); break;
 
-        default: err(operf,"",tx,ty);
+        default: err(operf,"",x,y);
       }
       return;
     }
@@ -1358,7 +1376,7 @@ gaffect(GEN x, GEN y)
       case t_RFRAC: case t_RFRACN:
 	gaffect(x,(GEN)y[1]); gaffsg(1,(GEN)y[2]); break;
 
-      default: err(operf,"",tx,ty);
+      default: err(operf,"",x,y);
     }
     return;
   }
@@ -1385,7 +1403,7 @@ gaffect(GEN x, GEN y)
 	  }
       }
 #undef initial_value
-      err(operf,"",tx,ty);
+      err(operf,"",x,y);
   }
   
   switch(tx)
@@ -1395,10 +1413,10 @@ gaffect(GEN x, GEN y)
       switch(ty)
       {
 	case t_POL:
-	  vy=varn(y); if (vy>v) err(operf,"",tx,ty);
+	  vy=varn(y); if (vy>v) err(operf,"",x,y);
 	  if (vy==v)
 	  {
-	    l=lgef(x); if (l>ly) err(operi,"",tx,ty);
+	    l=lgef(x); if (l>ly) err(operi,"",x,y);
 	    y[1]=x[1]; for (i=2; i<l; i++) gaffect((GEN)x[i],(GEN)y[i]);
 	  }
 	  else
@@ -1411,7 +1429,7 @@ gaffect(GEN x, GEN y)
 	  break;
 	
 	case t_SER:
-	  vy=varn(y); if (vy>v) err(operf,"",tx,ty);
+	  vy=varn(y); if (vy>v) err(operf,"",x,y);
 	  if (!signe(x)) { gaffsg(0,y); return; }
 	  if (vy==v)
 	  {
@@ -1437,13 +1455,13 @@ gaffect(GEN x, GEN y)
 	case t_RFRAC: case t_RFRACN:
 	  gaffect(x,(GEN)y[1]); gaffsg(1,(GEN)y[2]); break;
 
-        default: err(operf,"",tx,ty);
+        default: err(operf,"",x,y);
       }
       break;
 
     case t_SER:
-      if (ty!=t_SER) err(operf,"",tx,ty);
-      v=varn(x); vy=varn(y); if (vy>v) err(operf,"",tx,ty);
+      if (ty!=t_SER) err(operf,"",x,y);
+      v=varn(x); vy=varn(y); if (vy>v) err(operf,"",x,y);
       if (vy==v)
       {
 	y[1]=x[1]; k=lx; if (k>ly) k=ly;
@@ -1465,7 +1483,7 @@ gaffect(GEN x, GEN y)
       switch(ty)
       {
 	case t_POL: case t_VEC: case t_COL: case t_MAT:
-	  err(operf,"",tx,ty);
+	  err(operf,"",x,y);
 
 	case t_POLMOD:
 	  av=avma; p1=ginvmod((GEN)x[2],(GEN)y[1]);
@@ -1488,16 +1506,16 @@ gaffect(GEN x, GEN y)
 	  gaffect((GEN)x[1],(GEN)y[1]);
 	  gaffect((GEN)x[2],(GEN)y[2]); break;
 	
-        default: err(operf,"",tx,ty);
+        default: err(operf,"",x,y);
       }
       break;
 
     case t_QFR: case t_QFI: case t_VEC: case t_COL: case t_MAT:
-      if (ty != tx || ly != lx) err(operi,"",tx,ty);
+      if (ty != tx || ly != lx) err(operi,"",x,y);
       for (i=1; i<lx; i++) gaffect((GEN)x[i],(GEN)y[i]);
       break;
 
-    default: err(operf,"",tx,ty);
+    default: err(operf,"",x,y);
   }
 }
 
@@ -1574,7 +1592,7 @@ cvtop(GEN x, GEN p, long l)
 	  n = ggval((GEN)p2[1],p) - ggval((GEN)p2[2],p);
 	  p2=gadd(p2,ggrandocp(p,n+l)); break;
 
-	default: err(operi,"",t_QUAD,t_QUAD);
+	default: err(operi,"",x,x);
       }
       p2=gsqrt(p2,0); p1=gmul((GEN)x[3],gsub(p2,p3)); tetpil=avma;
       return gerepile(av,tetpil,gadd((GEN)x[2],p1));
@@ -1658,6 +1676,7 @@ sizedigit(GEN x)
   return gcmp0(x)? 0: (long) ((gexpo(x)+1) * L2SL10) + 1;
 }
 
+#if 0
 /* Normalize series x in place.
  * Assumption: x,x[2],...,x[lg(x)-1] have been created in that order.
  * All intermediate objects will be destroyed.
@@ -1682,6 +1701,29 @@ normalize(GEN x)
     }
   avma = (long) (x+lx); return zeroser(varn(x),lx-2);
 }
+#else
+
+/* normalize series. avma is not updated */
+GEN
+normalize(GEN x)
+{
+  long i, lx = lg(x);
+  GEN y;
+
+  if (typ(x) != t_SER) err(typeer,"normalize");
+  if (lx==2) { setsigne(x,0); return x; }
+  if (! isexactzero((GEN)x[2])) { setsigne(x,1); return x; }
+  for (i=3; i<lx; i++)
+    if (! isexactzero((GEN)x[i]))
+    {
+      i -= 2; y = x + i;
+      y[1] = evalsigne(1) | evalvalp(valp(x)+i) | evalvarn(varn(x));
+      y[0] = evaltyp(t_SER) | evallg(lx-i); /* don't swap these lines ! */
+      stackdummy(x, i); return y;
+    }
+  return zeroser(varn(x),lx-2);
+}
+#endif
 
 GEN
 normalizepol_i(GEN x, long lx)
