@@ -1,4 +1,4 @@
-/* $Id: gen3.c,v 1.43 2001/10/13 22:53:23 karim Exp $
+/* $Id: gen3.c,v 1.53 2001/12/14 17:42:08 karim Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -197,14 +197,13 @@ padicprec(GEN x, GEN p)
   return 0; /* not reached */
 }
 
-/* Degre de x par rapport a la variable v si v>=0, par rapport a la variable
- * principale si v<0. On suppose x fraction rationnelle ou polynome.
- * Convention deg(0)=-1.
+/* Degree of x (scalar, t_POL, t_RFRAC) wrt variable v if v >= 0,
+ * wrt to main variable if v < 0.  Convention: deg(0) = -1.
  */
 long
 poldegree(GEN x, long v)
 {
-  long tx=typ(x), av, w, d;
+  long tx = typ(x), lx,w,i,d;
 
   if (is_scalar_t(tx)) return gcmp0(x)? -1: 0;
   switch(tx)
@@ -213,9 +212,13 @@ poldegree(GEN x, long v)
       w = varn(x);
       if (v < 0 || v == w) return degpol(x);
       if (v < w) return signe(x)? 0: -1;
-      av = avma; x = gsubst(gsubst(x,w,polx[MAXVARN]),v,polx[0]);
-      if (gvar(x)) { d = gcmp0(x)? -1: 0; } else d = degpol(x);
-      avma = av; return d;
+      lx = lgef(x); d = -1;
+      for (i=2; i<lx; i++)
+      {
+        long e = poldegree((GEN)x[i], v);
+        if (e > d) d = e;
+      }
+      return d;
 
     case t_RFRAC: case t_RFRACN:
       if (gcmp0((GEN)x[1])) return -1;
@@ -584,22 +587,37 @@ gdivgs(GEN x, long s)
 
 /*******************************************************************/
 /*                                                                 */
-/*                    MODULO GENERAL                               */
+/*                    GENERIC REMAINDER                            */
 /*                                                                 */
 /*******************************************************************/
+/* euclidean quotient for scalars of admissible types */
+static GEN
+_quot(GEN x, GEN y)
+{
+  GEN q = gdiv(x,y), f = gfloor(q);
+  if (gsigne(y) < 0 && !gegal(f,q)) f = gadd(f,gun);
+  return f;
+}
+static GEN
+quot(GEN x, GEN y)
+{
+  ulong av = avma; return gerepileupto(av, _quot(x,y));
+}
 
 GEN
 gmod(GEN x, GEN y)
 {
-  long av,tetpil,i, tx=typ(x), ty = typ(y);
+  ulong av,tetpil;
+  long i,lx,ty, tx = typ(x);
   GEN z,p1;
 
   if (is_matvec_t(tx))
   {
-    i=lg(x); z=cgetg(i,tx);
-    for (i--; i; i--) z[i]=lmod((GEN)x[i],y);
+    lx=lg(x); z=cgetg(lx,tx);
+    for (i=1; i<lx; i++) z[i] = lmod((GEN)x[i],y);
     return z;
   }
+  ty = typ(y);
   switch(ty)
   {
     case t_INT:
@@ -634,28 +652,30 @@ gmod(GEN x, GEN y)
         }
 	case t_POLMOD: case t_POL:
 	  return gzero;
+      /* case t_REAL could be defined as below, but conlicting semantic
+       * with lift(x * Mod(1,y)). */
 
-	default: err(operf,"%",tx,ty);
+	default: err(operf,"%",x,y);
       }
 
     case t_REAL: case t_FRAC: case t_FRACN:
       switch(tx)
       {
 	case t_INT: case t_REAL: case t_FRAC: case t_FRACN:
-	  av=avma; p1 = gfloor(gdiv(x,y)); p1 = gneg_i(gmul(p1,y));
-          tetpil=avma; return gerepile(av,tetpil,gadd(x,p1));
+	  av = avma;
+          return gerepileupto(av, gsub(x, gmul(quot(x,y),y)));
 
 	case t_POLMOD: case t_POL:
 	  return gzero;
 
-	default: err(operf,"%",tx,ty);
+	default: err(operf,"%",x,y);
       }
 
     case t_POL:
       if (is_scalar_t(tx))
       {
         if (tx!=t_POLMOD || varn(x[1]) > varn(y))
-          return (lgef(y) < 4)? gzero: gcopy(x);
+          return degpol(y)? gcopy(x): gzero;
 	if (varn(x[1])!=varn(y)) return gzero;
         z=cgetg(3,t_POLMOD);
         z[1]=lgcd((GEN)x[1],y);
@@ -677,14 +697,14 @@ gmod(GEN x, GEN y)
           if (ismonome(y) && varn(x) == varn(y))
           {
             long d = degpol(y);
-            if (lg(x)-2 + valp(x) < d) err(operi,"%",tx,ty);
+            if (lg(x)-2 + valp(x) < d) err(operi,"%",x,y);
             av = avma; 
             return gerepileupto(av, gmod(gtrunc(x), y));
           }
-	default: err(operf,"%",tx,ty);
+	default: err(operf,"%",x,y);
       }
   }
-  err(operf,"%",tx,ty);
+  err(operf,"%",x,y);
   return NULL; /* not reached */
 }
 
@@ -701,7 +721,7 @@ gmodulsg(long x, GEN y)
     case t_POL: z=cgetg(3,t_POLMOD);
       z[1]=lcopy(y); z[2]=lstoi(x); return z;
   }
-  err(operf,"%",t_INT,typ(y)); return NULL; /* not reached */
+  err(operf,"%",stoi(x),y); return NULL; /* not reached */
 }
 
 GEN
@@ -748,7 +768,7 @@ gmodulo(GEN x,GEN y)
       if (tx!=t_POL && !is_rfrac_t(tx) && tx!=t_SER) break;
       z[2]=(long)specialmod(x,y); return z;
   }
-  err(operf,"%",tx,typ(y)); return NULL; /* not reached */
+  err(operf,"%",x,y); return NULL; /* not reached */
 }
 
 GEN
@@ -781,7 +801,7 @@ gmodulcp(GEN x,GEN y)
       if (tx!=t_POL && !is_rfrac_t(tx) && tx!=t_SER) break;
       z[2]=(long)specialmod(x,y); return z;
   }
-  err(operf,"%",tx,typ(y)); return NULL; /* not reached */
+  err(operf,"%",x,y); return NULL; /* not reached */
 }
 
 GEN
@@ -798,52 +818,194 @@ Mod0(GEN x,GEN y,long flag)
 
 /*******************************************************************/
 /*                                                                 */
-/*                 DIVISION ENTIERE GENERALE                       */
-/*            DIVISION ENTIERE AVEC RESTE GENERALE                 */
+/*                 GENERIC EUCLIDEAN DIVISION                      */
 /*                                                                 */
 /*******************************************************************/
 
 GEN
 gdivent(GEN x, GEN y)
 {
-  long tx=typ(x),ty=typ(y);
+  long tx = typ(x);
 
-  if (tx == t_INT)
+  if (is_matvec_t(tx))
   {
-    if (ty == t_INT) return truedvmdii(x,y,NULL);
-    if (ty!=t_POL) err(typeer,"gdivent");
-    return gzero;
+    long i, lx = lg(x);
+    GEN z = cgetg(lx,tx);
+    for (i=1; i<lx; i++) z[i] = (long)gdivent((GEN)x[i],y);
+    return z;
   }
-  if (ty != t_POL) err(typeer,"gdivent");
-  if (tx == t_POL) return gdeuc(x,y);
-  if (! is_scalar_t(tx)) err(typeer,"gdivent");
-  return gzero;
+  switch(typ(y))
+  {
+    case t_INT:
+      switch(tx)
+      { /* equal to, but more efficient than, quot(x,y) */
+        case t_INT: return truedvmdii(x,y,NULL);
+        case t_REAL:
+        case t_FRAC:
+        case t_FRACN: return quot(x,y);
+        case t_POL: return gdiv(x,y);
+      }
+      break;
+    case t_REAL:
+    case t_FRAC:
+    case t_FRACN: return quot(x,y);
+    case t_POL:
+      if (is_scalar_t(tx))
+      {
+        if (tx == t_POLMOD) break;
+        return degpol(y)? gzero: gdiv(x,y);
+      }
+      if (tx == t_POL) return gdeuc(x,y);
+  }
+  err(operf,"\\",x,y);
+  return NULL; /* not reached */
+}
+
+/* with remainder */
+static GEN
+quotrem(GEN x, GEN y, GEN *r)
+{
+  ulong av;
+  GEN q = quot(x,y);
+  av = avma;
+  *r = gerepileupto(av, gsub(x, gmul(q,y)));
+  return q;
 }
 
 GEN
 gdiventres(GEN x, GEN y)
 {
-  long tx=typ(x),ty=typ(y);
-  GEN z = cgetg(3,t_COL);
+  long tx = typ(x);
+  GEN z,q,r;
 
-  if (tx==t_INT)
+  if (is_matvec_t(tx))
   {
-    if (ty==t_INT)
-    {
-      z[1]=(long)truedvmdii(x,y,(GEN*)(z+2));
-      return z;
-    }
-    if (ty!=t_POL) err(typeer,"gdiventres");
-    z[1]=zero; z[2]=licopy(x); return z;
-  }
-  if (ty != t_POL) err(typeer,"gdiventres");
-  if (tx == t_POL)
-  {
-    z[1]=ldivres(x,y,(GEN *)(z+2));
+    long i, lx = lg(x);
+    z = cgetg(lx,tx);
+    for (i=1; i<lx; i++) z[i] = (long)gdiventres((GEN)x[i],y);
     return z;
   }
-  if (! is_scalar_t(tx)) err(typeer,"gdiventres");
-  z[1]=zero; z[2]=lcopy(x); return z;
+  z = cgetg(3,t_COL);
+  switch(typ(y))
+  {
+    case t_INT:
+      switch(tx)
+      { /* equal to, but more efficient than next case */
+        case t_INT:
+          z[1] = (long)truedvmdii(x,y,(GEN*)(z+2));
+          return z;
+        case t_REAL:
+        case t_FRAC:
+        case t_FRACN:
+          q = quotrem(x,y,&r);
+          z[1] = (long)q;
+          z[2] = (long)r; return z;
+        case t_POL: return gdiv(x,y);
+      }
+      break;
+    case t_REAL:
+    case t_FRAC:
+    case t_FRACN:
+          q = quotrem(x,y,&r);
+          z[1] = (long)q;
+          z[2] = (long)r; return z;
+    case t_POL:
+      if (is_scalar_t(tx))
+      {
+        if (tx == t_POLMOD) break;
+        if (degpol(y))
+        {
+          q = gzero;
+          r = gcopy(x);
+        }
+        else
+        {
+          q = gdiv(x,y);
+          r = gzero;
+        }
+        return q;
+      }
+      if (tx == t_POL)
+      {
+        z[1]=ldivres(x,y,(GEN *)(z+2));
+        return z;
+      }
+  }
+  err(operf,"\\",x,y);
+  return NULL; /* not reached */
+}
+
+extern GEN swap_vars(GEN b0, long v);
+
+GEN
+divrem(GEN x, GEN y, long v)
+{
+  ulong av = avma;
+  long vx,vy;
+  GEN z,q,r;
+  if (v < 0 || typ(y) != t_POL || typ(x) != t_POL) return gdiventres(x,y);
+  vx = varn(x); if (vx != v) x = swap_vars(x,v);
+  vy = varn(y); if (vy != v) y = swap_vars(y,v);
+  q = poldivres(x,y, &r);
+  if (v && (vx != v || vy != v))
+  {
+    q = poleval(q, polx[v]);
+    r = poleval(r, polx[v]);
+  }
+  z = cgetg(3,t_COL);
+  z[1] = (long)q;
+  z[2] = (long)r; return gerepilecopy(av, z);
+}
+
+static int
+is_scal(long t) { return t==t_INT || t==t_FRAC || t==t_FRACN; }
+
+/* If x and y are not both scalars, same as gdivent.
+ * Otherwise, compute the quotient x/y, rounded to the nearest integer
+ * (towards +oo in case of tie). */
+GEN
+gdivround(GEN x, GEN y)
+{
+  ulong av1,av = avma;
+  long tx=typ(x),ty=typ(y);
+  GEN q,r;
+  int fl;
+
+  if (tx==t_INT && ty==t_INT)
+  {
+    q = dvmdii(x,y,&r); /* q = x/y rounded towards 0, sgn(r)=sgn(x) */
+    if (r==gzero) return q;
+    av1 = avma;
+    fl = absi_cmp(shifti(r,1),y);
+    avma = av1; cgiv(r);
+    if (fl >= 0) /* If 2*|r| >= |y| */
+    {
+      long sz = signe(x)*signe(y);
+      if (fl || sz > 0) q = gerepileuptoint(av, addis(q,sz));
+    }
+    return q;
+  }
+  if (is_scal(tx) && is_scal(ty))
+  { /* same as above but less efficient */
+    q = quotrem(x,y,&r);
+    av1 = avma;
+    fl = gcmp(gmul2n(gabs(r,0),1), gabs(y,0));
+    avma = av1; cgiv(r);
+    if (fl >= 0) /* If 2*|r| >= |y| */
+    {
+      long sz = signe(x)*signe(y);
+      if (fl || sz > 0) q = gerepileupto(av, gaddgs(q, sz));
+    }
+    return q;
+  }
+  if (is_matvec_t(tx))
+  {
+    long i,lx = lg(x);
+    GEN z = cgetg(lx,tx);
+    for (i=1; i<lx; i++) z[i] = (long)gdivround((GEN)x[i],y);
+    return z;
+  }
+  return gdivent(x,y);
 }
 
 GEN
@@ -862,43 +1024,6 @@ gdivmod(GEN x, GEN y, GEN *pr)
   return poldivres(x,y,pr);
 }
 
-/* When x and y are integers, compute the quotient x/y, rounded to the
- * nearest integer. If there is a tie, the quotient is rounded towards zero.
- * If x and y are not both integers, same as gdivent.
- */
-GEN
-gdivround(GEN x, GEN y)
-{
-  long tx=typ(x),ty=typ(y);
-
-  if (tx==t_INT)
-  {
-    if (ty==t_INT)
-    {
-      long av = avma, av1,fl;
-      GEN r, q=dvmdii(x,y,&r); /* q = x/y rounded towards 0, sqn(r)=sgn(x) */
-
-      if (r==gzero) return q;
-      av1 = avma;
-      fl = absi_cmp(shifti(r,1),y);
-      avma = av1; cgiv(r);
-      if (fl >= 0) /* If 2*|r| >= |y| */
-      {
-        long sz = signe(x)*signe(y);
-	if (fl || sz > 0)
-          { av1=avma; q = gerepile(av,av1,addis(q,sz)); }
-      }
-      return q;
-    }
-    if (ty!=t_POL) err(typeer,"gdivround");
-    return gzero;
-  }
-  if (ty != t_POL) err(typeer,"gdivround");
-  if (tx == t_POL) return gdeuc(x,y);
-  if (! is_scalar_t(tx)) err(typeer,"gdivround");
-  return gzero;
-}
-
 /*******************************************************************/
 /*                                                                 */
 /*                       SHIFT D'UN GEN                            */
@@ -1109,6 +1234,13 @@ ginv(GEN x)
       return y;
     case t_MAT:
       return (lg(x)==1)? cgetg(1,t_MAT): invmat(x);
+    case t_VECSMALL:
+    {
+      long i,lx = lg(x);
+      y = cgetg(lx,t_VECSMALL);
+      for (i=1; i<lx; i++) y[x[i]] = i;
+      return y;
+    }
   }
   err(typeer,"inverse");
   return NULL; /* not reached */
@@ -1189,21 +1321,29 @@ gsubst(GEN x, long v, GEN y)
   switch(tx)
   {
     case t_POL:
-      l=lgef(x);
+      l = lgef(x);
       if (l==2)
         return (ty==t_MAT)? gscalmat(gzero,ly-1): gzero;
 
-      vx=varn(x);
-      if (vx<v)
+      vx = varn(x);
+      if (vx < v)
       {
+        if (gvar(y) > vx)
+        { /* easy special case */
+          z = cgetg(l, t_POL); z[1] = x[1];
+          for (i=2; i<l; i++) z[i] = lsubst((GEN)x[i],v,y);
+          return normalizepol_i(z,l);
+        }
+        /* general case */
 	av=avma; p1=polx[vx]; z= gsubst((GEN)x[l-1],v,y);
 	for (i=l-1; i>2; i--) z=gadd(gmul(z,p1),gsubst((GEN)x[i-1],v,y));
 	return gerepileupto(av,z);
       }
+      /* v <= vx */
       if (ty!=t_MAT)
-        return (vx>v)? gcopy(x): poleval(x,y);
+        return (vx > v)? gcopy(x): poleval(x,y);
 
-      if (vx>v) return gscalmat(x,ly-1);
+      if (vx > v) return gscalmat(x,ly-1);
       if (l==3) return gscalmat((GEN)x[2],ly-1);
       av=avma; z=(GEN)x[l-1];
       for (i=l-1; i>2; i--) z=gaddmat((GEN)x[i-1],gmul(z,y));
@@ -1371,7 +1511,7 @@ derivpol(GEN x)
   long i,lx = lgef(x)-1;
   GEN y;
 
-  if (lx<3) return gzero;
+  if (lx<3) return zeropol(varn(x));
   y = cgetg(lx,t_POL);
   for (i=2; i<lx ; i++) y[i] = lmulsg(i-1,(GEN)x[i+1]);
   y[1] = x[1]; return normalizepol_i(y,i);
@@ -2124,12 +2264,7 @@ gtovec(GEN x)
     for (i=1; i<=lx-2; i++) y[i]=lcopy((GEN)x[i]);
     return y;
   }
-  if (tx == t_VECSMALL)
-  {
-    lx=lg(x); y=cgetg(lx,t_VEC);
-    for (i=1; i<lx; i++) y[i]=lstoi(x[i]);
-    return y;
-  }
+  if (tx == t_VECSMALL) return small_to_vec(x);
   if (!signe(x)) { y=cgetg(2,t_VEC); y[1]=zero; return y; }
   lx=lg(x); y=cgetg(lx-1,t_VEC); x++;
   for (i=1; i<=lx-2; i++) y[i]=lcopy((GEN)x[i]);
@@ -2172,74 +2307,100 @@ compo(GEN x, long n)
   return gcopy((GEN)x[l]);
 }
 
+/* assume x a t_POL */
+static GEN
+_polcoeff(GEN x, long n, long v)
+{
+  long i, w, dx;
+  GEN z;
+  dx = degpol(x);
+  if (dx < 0) return gzero;
+  if (v < 0 || v == (w=varn(x)))
+    return (n < 0 || n > dx)? gzero: (GEN)x[n+2];
+  if (w > v) return n? gzero: x;
+  /* w < v */
+  if (dx == 0) return polcoeff_i((GEN)x[2], n, v);
+  z = cgetg(dx+3, t_POL);
+  z[1] = x[1];
+  for (i=2; i<dx+3; i++) z[i] = (long)polcoeff_i((GEN)x[i], n, v);
+  return normalizepol(z);
+}
+
+/* assume x a t_SER */
+static GEN
+_sercoeff(GEN x, long n, long v)
+{
+  long i, w, dx, ex;
+  GEN z;
+ 
+  if (!signe(x)) return gzero;
+  dx = lg(x)-3; ex = valp(x); n -= ex;
+  if (v < 0 || v == (w=varn(x)))
+  {
+    if (n > dx) err(talker,"non existent component in truecoeff");
+    return (n < 0)? gzero: (GEN)x[n+2];
+  }
+  if (w > v) return n? gzero: x;
+  /* w < v */
+  if (dx == 0) return polcoeff_i((GEN)x[2], n, v);
+  z = cgetg(dx+3, t_POL);
+  z[1] = x[1];
+  for (i=2; i<dx+3; i++) z[i] = (long)polcoeff_i((GEN)x[i], n, v);
+  return normalizepol(z);
+}
+
+/* assume x a t_RFRAC(n) */
+static GEN
+_rfraccoeff(GEN x, long n, long v)
+{
+  long ex, w = precdl;
+  if (v<0) v = gvar(x);
+  ex = ggval((GEN)x[2], polx[v]);
+  precdl = n + ex + 1; x = gtoser(x, v);
+  precdl = w;
+  return _sercoeff(x,n,v);
+}
+
+
+GEN
+polcoeff_i(GEN x, long n, long v)
+{
+  switch(typ(x))
+  {
+    case t_POL: return _polcoeff(x,n,v);
+    case t_SER: return _sercoeff(x,n,v);
+    case t_RFRAC:
+    case t_RFRACN: return _rfraccoeff(x,n,v);
+    default: return n? gzero: x;
+  }
+}
+
 /* with respect to the main variable if v<0, with respect to the variable v
    otherwise. v ignored if x is not a polynomial/series. */
-
 GEN
 polcoeff0(GEN x, long n, long v)
 {
-  long tx=typ(x),lx,ex,w,av,tetpil;
-  GEN xinit;
+  long tx=typ(x),av;
 
   if (is_scalar_t(tx)) return n? gzero: gcopy(x);
 
+  av = avma;
   switch(tx)
   {
+    case t_POL: x = _polcoeff(x,n,v); break;
+    case t_SER: x = _sercoeff(x,n,v); break;
+    case t_RFRAC:
+    case t_RFRACN: x = _rfraccoeff(x,n,v); break;
+   
     case t_QFR: case t_QFI: case t_VEC: case t_COL: case t_MAT:
-      if (n<1 || n>=lg(x)) break;
-      return gcopy((GEN)x[n]);
+      if (n>=1 && n<lg(x)) return gcopy((GEN)x[n]);
+    /* fall through */
 
-    case t_POL:
-      if (n<0) return gzero;
-      w=varn(x);
-      if (v < 0 || v == w)
-	return (n>=lgef(x)-2)? gzero: gcopy((GEN)x[n+2]);
-      if (v < w) return n? gzero: gcopy(x);
-      av=avma; xinit=x;
-      x=gsubst(gsubst(x,w,polx[MAXVARN]),v,polx[0]);
-      if (gvar(x)) { avma=av; return n? gzero: gcopy(xinit); }
-      if (typ(x) == t_POL)
-      {
-        if (n>=lgef(x)-2) { avma=av; return gzero; }
-        x = (GEN)x[n+2];
-      }
-      else
-        x = polcoeff0(x, n, 0);
-      tetpil=avma; return gerepile(av,tetpil,gsubst(x,MAXVARN,polx[w]));
-
-    case t_SER:
-      w=varn(x);
-      if (v < 0 || v == w)
-      {
-	if (!signe(x)) return gzero;
-	lx=lg(x); ex=valp(x); if (n<ex) return gzero;
-	if (n>=ex+lx-2) break;
-	return gcopy((GEN)x[n-ex+2]);
-      }
-      if (v < w) return n?  gzero: gcopy(x);
-      av=avma; xinit=x;
-      x=gsubst(gsubst(x,w,polx[MAXVARN]),v,polx[0]);
-      if (gvar(x)) { avma=av; return n? gzero: gcopy(xinit); }
-      if (gcmp0(x)) { avma=av; return gzero; }
-      if (typ(x) == t_SER)
-      {
-        lx=lg(x); ex=valp(x); if (n<ex) return gzero;
-        if (n>=ex+lx-2) break;
-        x = (GEN)x[n-ex+2];
-      }
-      else
-        x = polcoeff0(x, n, 0);
-      tetpil=avma; return gerepile(av,tetpil,gsubst(x,MAXVARN,polx[w]));
-
-    case t_RFRAC: case t_RFRACN:
-      w = precdl; av = avma;
-      if (v<0) v = gvar(x);
-      ex = ggval((GEN)x[2], polx[v]);
-      precdl = n + ex + 1; x = gtoser(x, v); precdl = w;
-      return gerepileupto(av, polcoeff0(x, n, v));
+    default: err(talker,"nonexistent component in truecoeff");
   }
-  err(talker,"nonexistent component in truecoeff");
-  return NULL; /* not reached */
+  if (x == gzero) return x;
+  if (avma == av) return gcopy(x);
+  return gerepilecopy(av, x);
 }
 
 GEN
