@@ -1,4 +1,4 @@
-/* $Id: rootpol.c,v 1.36 2001/10/15 13:37:55 karim Exp $
+/* $Id: rootpol.c,v 1.44 2002/01/10 21:32:59 karim Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -358,11 +358,13 @@ mylog2(GEN z)
   return x+0.5*log2( 1 + exp2(2*(y-x)));
 }
 
-long
+/* find s such that  A_h <= 2^s <= 2 A_i  for one h and all i < n = deg(p),
+ * with  A_i := (binom(n,i) lc(p) / p_i) ^ 1/(n-i), and  p = sum p_i X^i */
+static long
 findpower(GEN p)
 {
-  double x, logbinomial,pente, pentemax = -pariINFINITY;
-  long n=degpol(p),i;
+  double x, logbinomial, mins = pariINFINITY;
+  long n = degpol(p),i;
 
   logbinomial = mylog2((GEN)p[n+2]); /* log2(lc * binom(n,i)) */
   for (i=n-1; i>=0; i--)
@@ -371,11 +373,13 @@ findpower(GEN p)
     x = mylog2((GEN)p[i+2]);
     if (x != -pariINFINITY)
     {
-      pente = (x - logbinomial) / (double) (n-i);
-      if (pente > pentemax) pentemax = pente;
+      double s = (logbinomial - x) / (double) (n-i);
+      if (s < mins) mins = s;
     }
   }
-  return (long) -floor(pentemax);
+  i = (long)ceil(mins);
+  if (i - mins > 1 - 1e-12) i--;
+  return i;
 }
 
 /* returns the exponent for the procedure modulus, from the newton diagram */
@@ -439,10 +443,8 @@ myshiftic(GEN z, long e)
 static GEN
 myrealun(long bitprec)
 {
-  GEN x;
   if (bitprec < 0) bitprec = 0;
-  x = cgetr(bitprec/BITS_IN_LONG+3);
-  affsr(1,x); return x;
+  return realun(bitprec/BITS_IN_LONG+3);
 }
 
 static GEN
@@ -542,7 +544,7 @@ homothetie(GEN p, GEN R, long bitprec)
   return r;
 }
 
-/* change q in 2^(n*e) p(x*2^(-e)), n=deg(q) */
+/* change q in 2^(n*e) p(x*2^(-e)), n=deg(q)  [ ~as above with R = 2^-e ]*/
 static void
 homothetie2n(GEN p, long e)
 {
@@ -564,15 +566,6 @@ homothetie_gauss(GEN p, long e,long f)
   }
 }
 
-static long
-valuation(GEN p)
-{
-  long j=0,n=degpol(p);
-
-  while ((j<=n) && isexactzero((GEN)p[j+2])) j++;
-  return j;
-}
-
 /* provides usually a good lower bound on the largest modulus of the roots,
 puts in k an upper bound of the number of roots near the largest root
 at a distance eps */
@@ -580,82 +573,83 @@ static double
 lower_bound(GEN p, long *k, double eps)
 {
   long n=degpol(p),i,j,ltop=avma;
-  GEN a,s,icd;
+  GEN a,s,S,icd;
   double r,*rho;
 
   if (n<4) { *k=n; return 0.; }
-  a=cgetg(6,t_POL); s=cgetg(6,t_POL);
+  S = cgetg(5,t_VEC);
+  a = cgetg(5,t_VEC);
   rho=(double *) gpmalloc(4*sizeof(double));
   icd = gdiv(realun(DEFAULTPREC), (GEN) p[n+2]);
-  for (i=1; i<=4; i++) a[i+1]=lmul(icd,(GEN)p[n+2-i]);
+  for (i=1; i<=4; i++) a[i] = lmul(icd,(GEN)p[n+2-i]);
   for (i=1; i<=4; i++)
   {
-    s[i+1]=lmulsg(i,(GEN)a[i+1]);
+    s = gmulsg(i,(GEN)a[i]);
     for (j=1; j<i; j++)
-      s[i+1]=ladd((GEN)s[i+1],gmul((GEN)s[j+1],(GEN)a[i+1-j]));
-    s[i+1]=lneg((GEN)s[i+1]);
-    r=gtodouble(gabs((GEN) s[i+1],3));
-    if (r<=0.)  /* should not be strictly negative */
-      rho[i-1]=0.;
+      s = gadd(s, gmul((GEN)S[j],(GEN)a[i-j]));
+    S[i] = lneg(s); /* Newton sum S_i */
+    r = gtodouble(gabs(s,3));
+    if (r == 0.)
+      rho[i-1] = 0.;
     else
-      rho[i-1]=exp(log(r/(double)n)/(double) i);
+      rho[i-1] = exp(log(r/(double)n)/(double) i);
   }
-  r=0.;
+  r = 0.;
   for (i=0; i<4; i++) if (r<rho[i]) r=rho[i];
   if (r>0. && eps<1.2)
-    *k=(long) floor((n*rho[0]/r+n)/(1+exp(-eps)*cos(eps)));
+    *k = (long) floor((n*rho[0]/r+n)/(1+exp(-eps)*cos(eps)));
   else
-    *k=n;
-  free(rho); avma=ltop; return r;
+    *k = n;
+  free(rho); avma = ltop; return r;
 }
 
 /* returns the maximum of the modulus of p with a relative error tau */
 static GEN
 max_modulus(GEN p, double tau)
 {
-  GEN q,aux,gunr;
-  long i,j,k,valuat,n=degpol(p),nn,ltop=avma,bitprec,imax,e;
-  double r,rho,eps, tau2 = (tau > 3.0)? 0.5: tau/6.;
+  GEN r,q,aux,gunr;
+  ulong av, ltop = avma;
+  long i,k,n=degpol(p),nn,bitprec,M,e;
+  double rho,eps, tau2 = (tau > 3.0)? 0.5: tau/6.;
+
+  r = cgeti(BIGDEFAULTPREC);
+  av = avma;
 
   eps = - 1/log(1.5*tau2); /* > 0 */
-  bitprec=(long) ((double) n*log2(1./tau2)+3*log2((double) n))+1;
-  gunr=myrealun(bitprec+2*n);
-  aux=gdiv(gunr,(GEN) p[2+n]);
-  q=gmul(aux,p); q[2+n]=lcopy(gunr);
-  k=nn=n;
-  e=findpower(q); homothetie2n(q,e); r=-(double) e;
-  q=mygprec(q,bitprec+(n<<1));
+  bitprec = (long) ((double) n*log2(1./tau2)+3*log2((double) n))+1;
+  gunr = myrealun(bitprec+2*n);
+  aux = gdiv(gunr,(GEN) p[2+n]);
+  q = gmul(aux,p); q[2+n] = (long)gunr;
+  e = findpower(q);
+  homothetie2n(q,e);
+  affsi(e, r);
+  q = mygprec(q,bitprec+(n<<1));
   pol_to_gaussint(q,bitprec);
-  imax = (long) (log2( log(4.*n) / (2*tau2) )) + 2;
+  M = (long) (log2( log(4.*n) / (2*tau2) )) + 2;
+  nn = n;
   for (i=0,e=0;;)
   { /* nn = deg(q) */
     rho = lower_bound(q,&k,eps);
     if (rho > exp2(-(double) e)) e = (long) -floor(log2(rho));
-    r -= e / exp2((double)i);
-    if (++i == imax) {
-      avma=ltop; 
-      return gpui(dbltor(2.),dbltor(r),DEFAULTPREC);
-    }
+    affii(shifti(addis(r,e), 1), r);
+    if (++i == M) break;
 
-    if (k<nn)
-      bitprec=(long) ((double) k* log2(1./tau2)+
-                      (double) (nn-k)*log2(1./eps)+
-                      3*log2((double) nn))+1;
-    else /* k = nn */
-      bitprec=(long) ((double) k* log2(1./tau2)+
-                      3.*log2((double) nn))+1;
+    bitprec = (long) ((double) k* log2(1./tau2) +
+                     (double) (nn-k)*log2(1./eps) + 3*log2((double) nn)) + 1;
     homothetie_gauss(q,e,bitprec-(long)floor(mylog2((GEN) q[2+nn])+0.5));
-    valuat=valuation(q);
-    if (valuat>0)
-    {
-      nn-=valuat; setlgef(q,nn+3);
-      for (j=0; j<=nn; j++) q[2+j]=q[2+valuat+j];
-    }
+    nn -= polvaluation(q, &q);
     set_karasquare_limit(gexpo(q));
-    q = gerepileupto(ltop, graeffe(q));
-    tau2 *= 1.5; eps = 1/log(1./tau2);
+    q = gerepileupto(av, graeffe(q));
+    tau2 *= 1.5; eps = -1/log(tau2);
     e = findpower(q);
   }
+  if (!signe(r)) { avma = ltop; return realun(DEFAULTPREC); } 
+  aux = cgetr(DEFAULTPREC);
+  affir(r,aux); r = aux; setexpo(r, expo(r) - M);
+  rho = rtodbl(r);
+  /* rho = sum e_i 2^-i */
+  avma = ltop;
+  return mpexp(dbltor(-rho * LOG2)); /* 2^-r */
 }
 
 /* return the k-th modulus (in ascending order) of p, rel. error tau*/
@@ -663,32 +657,25 @@ static GEN
 modulus(GEN p, long k, double tau)
 {
   GEN q,gunr;
-  long i,j,kk=k,imax,n=degpol(p),nn,nnn,valuat,av,ltop=avma,bitprec,decprec,e;
+  long i,kk=k,imax,n=degpol(p),nn,av,ltop=avma,bitprec,decprec,e;
   double tau2,r;
 
-  tau2=tau/6; nn=n;
+  tau2 = tau/6;
   bitprec= (long) ((double) n*(2.+log2(3.*(double) n)+log2(1./tau2)));
   decprec=(long) ((double) bitprec * L2SL10)+1;
   gunr=myrealun(bitprec);
   av = avma;
-  q=gprec(p,decprec);
-  q=gmul(gunr,q);
-  e=newton_polygon(q,k);
+  q = gprec(p,decprec);
+  q = gmul(gunr,q);
+  e = newton_polygon(q,k);
   homothetie2n(q,e);
-  r=(double) e;
+  r = (double) e;
   imax=(long) ((log2(3./tau)+log2(log(4.*(double) n)) ))+1;
   for (i=1; i<imax; i++)
   {
-    q=eval_rel_pol(q,bitprec);
-
-    nnn=degpol(q); valuat=valuation(q);
-    if (valuat>0)
-    {
-      kk-=valuat;
-      for (j=0; j<=nnn-valuat; j++) q[2+j]=q[2+valuat+j];
-      setlgef(q,nnn-valuat+3);
-    }
-    nn=nnn-valuat;
+    q = eval_rel_pol(q,bitprec);
+    kk -= polvaluation(q, &q);
+    nn = degpol(q);
 
     set_karasquare_limit(bitprec);
     q = gerepileupto(av, graeffe(q));
@@ -697,7 +684,7 @@ modulus(GEN p, long k, double tau)
     q=gmul(gunr,q);
     homothetie2n(q,e);
 
-    tau2=1.5*tau2; if (tau2>1.) tau2=1.;
+    tau2 *= 1.5; if (tau2 > 1.) tau2 = 1.;
     bitprec= 1+(long) ((double) nn*(2.+log2(3.*(double) nn)+log2(1./tau2)));
   }
   avma=ltop; return mpexp(dbltor(-r * LOG2));
@@ -759,7 +746,7 @@ static long
 dual_modulus(GEN p, GEN R, double tau, long l)
 {
   GEN q;
-  long i,j,imax,k,delta_k=0,n=degpol(p),nn,nnn,valuat,ltop=avma,bitprec,ll=l;
+  long i,imax,k,delta_k=0,n=degpol(p),nn,nnn,valuat,ltop=avma,bitprec,ll=l;
   double logmax,aux,tau2;
 
   tau2=7.*tau/8.;
@@ -773,24 +760,18 @@ dual_modulus(GEN p, GEN R, double tau, long l)
     bitprec=6*nn-5*ll+(long) ((double) nn*(log2(1/tau2)+8.*tau2/7.));
 
     q=eval_rel_pol(q,bitprec);
-    nnn=degpol(q); valuat=valuation(q);
-    if (valuat>0)
-    {
-      delta_k+=valuat;
-      for (j=0; j<=nnn-valuat; j++) q[2+j]=q[2+valuat+j];
-      setlgef(q,nnn-valuat+3);
-    }
+    nnn = degpol(q); valuat = polvaluation(q, &q);
     ll= (-valuat<nnn-n)? ll-valuat: ll+nnn-n; /* min(ll-valuat,ll+nnn-n) */
     if (ll<0) ll=0;
 
-    nn=nnn-valuat;
+    nn = nnn-valuat; delta_k += valuat;
     if (nn==0) return delta_k;
 
     set_karasquare_limit(bitprec);
     q = gerepileupto(ltop, graeffe(q));
     tau2=tau2*7./4.;
   }
-  k=-1; logmax=- (double) pariINFINITY;
+  k = -1; logmax = - (double)pariINFINITY;
   for (i=0; i<=degpol(q); i++)
   {
     aux=mylog2((GEN)q[2+i]);
@@ -887,7 +868,7 @@ static GEN
 exp_i(GEN x)
 {
   GEN v;
-  
+
   if (!signe(x)) return realun(lg(x)); /* should not happen */
   v = cgetg(3,t_COMPLEX);
   mpsincos(x, (GEN*)(v+2), (GEN*)(v+1));
@@ -1374,7 +1355,7 @@ compute_radius(GEN* radii, GEN p, long k, double aux, double *delta)
   {
     if (!signe(radii[i]) || cmprr(radii[i], p1) < 0)
       affrr(p1, radii[i]);
-    else      
+    else
       p1 = radii[i];
   }
   *delta = rtodbl(gmul2n(mplog(divrr(rmax,rmin)), -1));
@@ -1429,7 +1410,7 @@ conformal_mapping(GEN *radii, GEN ctr, GEN p, long k, long bitprec,
 
   rho = compute_radius(radii, q,k,aux/10., &delta);
   invrho = update_radius(radii, rho, &param, &param2);
-  
+
   bitprec2 += (long) (((double)n) * fabs(log2ir(rho)) + 1.);
   R = mygprec(invrho,bitprec2);
   q = scalepol(q,R,bitprec2);
@@ -1457,7 +1438,7 @@ conformal_mapping(GEN *radii, GEN ctr, GEN p, long k, long bitprec,
 }
 
 static GEN
-myrealzero()
+myrealzero(void)
 {
   GEN x = cgetr(3);
   x[1] = evalexpo(-bit_accuracy(3));
@@ -1474,7 +1455,7 @@ split_2(GEN p, long bitprec, GEN ctr, double thickness, GEN *F, GEN *G)
   long n=degpol(p),i,j,k,bitprec2;
   GEN q,FF,GG,R;
   GEN *radii = (GEN*) cgetg(n+1, t_VEC);
-  for (i=2; i<n; i++) radii[i]=myrealzero(3);
+  for (i=2; i<n; i++) radii[i] = myrealzero();
   aux = thickness/(double) n/4.;
   radii[1] = rmin = min_modulus(p, aux);
   radii[n] = rmax = max_modulus(p, aux);
@@ -1836,7 +1817,7 @@ split_complete(GEN p, long bitprec, GEN roots_pol)
 }
 
 /* compute a bound on the maximum modulus of roots of p */
-static GEN
+GEN
 cauchy_bound(GEN p)
 {
   long i,n=degpol(p);
@@ -1928,10 +1909,11 @@ all_roots(GEN p, long bitprec)
   pd = poldeflate(p, &h); lc = leading_term(pd);
   e = 2*gexpo(cauchy_bound(pd)); if (e<0) e=0;
   bitprec0=bitprec + gexpo(pd) - gexpo(lc) + (long)log2(n/h)+1+e;
+  bitprec2 = bitprec0; e = 0;
   for (av=avma,i=1;; i++,avma=av)
   {
-    roots_pol = cgetg(n+1,t_VEC); setlg(roots_pol,1); 
-    bitprec2 = bitprec0 + (1<<i)*n;
+    roots_pol = cgetg(n+1,t_VEC); setlg(roots_pol,1);
+    bitprec2 += e + (1<<i)*n;
     q = gmul(myrealun(bitprec2), mygprec(pd,bitprec2));
     m = split_complete(q,bitprec2,roots_pol);
     roots_pol = fix_roots(roots_pol, &m, h, bitprec2);
@@ -1942,8 +1924,8 @@ all_roots(GEN p, long bitprec)
     if (e<-2*bitprec2) e=-2*bitprec2; /* to avoid e=-pariINFINITY */
     if (e < 0)
     {
-      e = a_posteriori_errors(p,roots_pol,e);
-      if (e < -bitprec) return roots_pol;
+      e = bitprec + a_posteriori_errors(p,roots_pol,e);
+      if (e < 0) return roots_pol;
     }
     if (DEBUGLEVEL > 7)
       fprintferr("all_roots: restarting, i = %ld, e = %ld\n", i,e);
