@@ -1,4 +1,4 @@
-/* $Id: trans1.c,v 1.36 2001/10/12 11:22:44 karim Exp $
+/* $Id: trans1.c,v 1.46 2001/12/15 14:14:59 karim Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -254,6 +254,23 @@ puiss0(GEN x)
   return y;
 }
 
+static GEN
+_sqr(void *data /* ignored */, GEN x) {
+  (void)data; return gsqr(x);
+}
+static GEN
+_mul(void *data /* ignored */, GEN x, GEN y) {
+  (void)data; return gmul(x,y);
+}
+static GEN
+_sqri(void *data /* ignored */, GEN x) {
+  (void)data; return sqri(x);
+}
+static GEN
+_muli(void *data /* ignored */, GEN x, GEN y) {
+  (void)data; return mulii(x,y);
+}
+
 /* INTEGER POWERING (a^|n| for integer a and integer |n| > 1)
  *
  * Nonpositive powers and exponent one should be handled by gpow() and
@@ -269,7 +286,7 @@ puiss0(GEN x)
 static GEN
 puissii(GEN a, GEN n, long s)
 {
-  long av,*p,m,k,i,lim;
+  ulong av;
   GEN y;
 
   if (!signe(a)) return gzero; /* a==0 */
@@ -283,71 +300,56 @@ puissii(GEN a, GEN n, long s)
     if (n[2] == 1) { a = icopy(a); setsigne(a,s); return a; }
     if (n[2] == 2) return sqri(a);
   }
-  /* be paranoid about memory consumption */
-  av=avma; lim=stack_lim(av,1);
-  y = a; p = n+2; m = *p;
-  /* normalize, i.e set highest bit to 1 (we know m != 0) */
-  k = 1+bfffo(m); m<<=k; k = BITS_IN_LONG-k;
-  /* first bit is now implicit */
-  for (i=lgefint(n)-2;;)
-  {
-    for (; k; m<<=1,k--)
-    {
-      y = sqri(y);
-      if (m < 0) y = mulii(y,a); /* first bit is set: multiply by base */
-      if (low_stack(lim, stack_lim(av,1)))
-      {
-        if (DEBUGMEM>1) err(warnmem,"puissii");
-        y = gerepileuptoint(av,y);
-      }
-    }
-    if (--i == 0) break;
-    m = *++p; k = BITS_IN_LONG;
-  }
+  av = avma;
+  y = leftright_pow(a, n, NULL, &_sqri, &_muli);
   setsigne(y,s); return gerepileuptoint(av,y);
 }
 
+struct muldata {
+  long prec, a;
+  GEN (*sqr)(GEN);
+  GEN (*mulsg)(long,GEN);
+  GEN unr;
+};
+
+static GEN
+_rpowsi_mul(void *data, GEN x, GEN y/* base; ignored */)
+{
+  struct muldata *D = (struct muldata *)data;
+  return D->mulsg(D->a, x);
+}
+
+static GEN
+_rpowsi_sqr(void *data, GEN x)
+{
+  struct muldata *D = (struct muldata *)data;
+  if (lgefint(x) >= D->prec && typ(x) == t_INT)
+  { /* switch to t_REAL */
+    D->sqr   = &gsqr;
+    D->mulsg = &mulsr;
+    affir(x, D->unr); x = D->unr;
+  }
+  return D->sqr(x);
+}
+
 /* return a^n as a t_REAL of precision prec. Adapted from puissii().
  * Assume a > 0, n > 0 */
 GEN
 rpowsi(ulong a, GEN n, long prec)
 {
-  long av,*p,m,k,i,lim;
+  ulong av = avma;
   GEN y, unr = realun(prec);
-  GEN (*sq)(GEN);
-  GEN (*mulsg)(long,GEN);
+  struct muldata D;
 
   if (a == 1) return unr;
   if (a == 2) { setexpo(unr, itos(n)); return unr; }
   if (is_pm1(n)) { affsr(a, unr); return unr; }
-  /* be paranoid about memory consumption */
-  av=avma; lim=stack_lim(av,1);
-  y = stoi(a); p = n+2; m = *p;
-  /* normalize, i.e set highest bit to 1 (we know m != 0) */
-  k = 1+bfffo(m); m<<=k; k = BITS_IN_LONG-k;
-  /* first bit is now implicit */
-  sq = &sqri; mulsg = &mulsi;
-  for (i=lgefint(n)-2;;)
-  {
-    for (; k; m<<=1,k--)
-    {
-      y = sq(y);
-      if (m < 0) y = mulsg(a,y);
-      if (lgefint(y) >= prec && typ(y) == t_INT) /* switch to t_REAL */
-      { 
-        sq = &gsqr; mulsg = &mulsr;
-        affir(y, unr); y = unr;
-      }
-
-      if (low_stack(lim, stack_lim(av,1)))
-      {
-        if (DEBUGMEM>1) err(warnmem,"rpuisssi");
-        y = gerepileuptoleaf(av,y);
-      }
-    }
-    if (--i == 0) break;
-    m = *++p, k = BITS_IN_LONG;
-  }
+  D.sqr   = &sqri; 
+  D.mulsg = &mulsi;
+  D.prec = prec;
+  D.unr = unr;
+  D.a = a;
+  y = leftright_pow(stoi(a), n, (void*)&D, &_rpowsi_sqr, &_rpowsi_mul);
   if (typ(y) == t_INT) { affir(y, unr); y = unr ; }
   return gerepileuptoleaf(av,y);
 }
@@ -462,9 +464,8 @@ extern GEN powrealform(GEN x, GEN n);
 GEN
 powgi(GEN x, GEN n)
 {
-  long i,j,m,tx, sn=signe(n);
-  ulong lim,av;
-  GEN y, p1;
+  long tx, sn = signe(n);
+  GEN y;
 
   if (typ(n) != t_INT) err(talker,"not an integral exponent in powgi");
   if (!sn) return puiss0(x);
@@ -511,25 +512,26 @@ powgi(GEN x, GEN n)
     }
     case t_PADIC:
     {
-      GEN p, pe;
+      long e = itos(n)*valp(x), v;
+      GEN mod, p = (GEN)x[2];
+      
       if (!signe(x[4]))
       {
         if (sn < 0) err(talker, "division by 0 p-adic in powgi");
-        y=gcopy(x); setvalp(y, itos(n)*valp(x)); return y;
+        return padiczero(p, e);
       }
       y = cgetg(5,t_PADIC);
-      p = (GEN)x[2];
-      pe= (GEN)x[3]; i = ggval(n, p);
-      if (i == 0) pe = icopy(pe);
+      mod = (GEN)x[3]; v = ggval(n, p);
+      if (v == 0) mod = icopy(mod);
       else
       {
-        pe = mulii(pe, gpowgs(p,i));
-        pe = gerepileuptoint((long)y, pe);
+        mod = mulii(mod, gpowgs(p,v));
+        mod = gerepileuptoint((long)y, mod);
       }
-      y[1] = evalprecp(precp(x)+i) | evalvalp(itos(n) * valp(x));
+      y[1] = evalprecp(precp(x)+v) | evalvalp(e);
       icopyifstack(p, y[2]);
-      y[3] = (long)pe;
-      y[4] = (long)powmodulo((GEN)x[4], n, pe);
+      y[3] = (long)mod;
+      y[4] = (long)powmodulo((GEN)x[4], n, mod);
       return y;
     }
     case t_QFR:
@@ -537,31 +539,12 @@ powgi(GEN x, GEN n)
     case t_POL:
       if (ismonome(x)) return pow_monome(x,n);
     default:
-      av=avma; lim=stack_lim(av,1);
-      p1 = n+2; m = *p1;
-      y=x; j=1+bfffo(m); m<<=j; j = BITS_IN_LONG-j;
-      for (i=lgefint(n)-2;;)
-      {
-        for (; j; m<<=1,j--)
-        {
-          y=gsqr(y);
-          if (low_stack(lim, stack_lim(av,1)))
-          {
-            if(DEBUGMEM>1) err(warnmem,"[1]: powgi");
-            y = gerepileupto(av, y);
-          }
-          if (m<0) y=gmul(y,x);
-          if (low_stack(lim, stack_lim(av,1)))
-          {
-            if(DEBUGMEM>1) err(warnmem,"[2]: powgi");
-            y = gerepileupto(av, y);
-          }
-        }
-        if (--i == 0) break;
-        m = *++p1, j = BITS_IN_LONG;
-      }
+    {
+      ulong av = avma;
+      y = leftright_pow(x, n, NULL, &_sqr, &_mul);
       if (sn < 0) y = ginv(y);
       return av==avma? gcopy(y): gerepileupto(av,y);
+    }
   }
 }
 
@@ -664,113 +647,174 @@ gpow(GEN x, GEN n, long prec)
 GEN
 mpsqrt(GEN x)
 {
-  long l,l0,l1,l2,s,eps,n,i,ex,av;
+  ulong av, av0;
+  long l,l0,l1,l2,s,eps,n,i,ex;
   double beta;
   GEN y,p1,p2,p3;
 
-  if (typ(x)!=t_REAL) err(typeer,"mpsqrt");
-  s=signe(x); if (s<0) err(talker,"negative argument in mpsqrt");
+  if (typ(x) != t_REAL) err(typeer,"mpsqrt");
+  s = signe(x); if (s < 0) err(talker,"negative argument in mpsqrt");
   if (!s) return realzero_bit(expo(x) >> 1);
-  l=lg(x); y=cgetr(l); av=avma;
+ 
+  l = lg(x); y = cgetr(l); av0 = avma;
 
-  p1=cgetr(l+1); affrr(x,p1);
-  ex=expo(x); eps = ex & 1; ex >>= 1; 
-  setexpo(p1,eps); setlg(p1,3);
+  p1 = cgetr(l+1); affrr(x,p1);
+  ex = expo(x); eps = ex & 1; ex >>= 1;
+  setexpo(p1,eps); /* x = 2^(2 ex) p1 */
 
-  n=(long)(2+log((double) (l-2))/LOG2);
-  p2=cgetr(l+1); p2[1]=evalexpo(0) | evalsigne(1);
-  beta=sqrt((eps+1) * (2 + p1[2]/C31));
-  p2[2]=(long)((beta-2)*C31);
-  if (!p2[2]) { p2[2]=HIGHBIT; setexpo(p2,1); }
-  for (i=3; i<=l; i++) p2[i]=0;
-  setlg(p2,3);
+  n = (long)(2 + log((double) (l-2))/LOG2);
+  beta = sqrt((eps+1) * (2 + p1[2]/C31));
+  p2 = cgetr(l+1);
+  p2[1] = evalexpo(0) | evalsigne(1);
+  p2[2] = (long)((beta-2)*C31);
+  if (!p2[2]) { p2[2] = (long)HIGHBIT; setexpo(p2,1); }
+  for (i=3; i<=l; i++) p2[i] = 0;
 
-  p3=cgetr(l+1); l-=2; l1=1; l2=3;
+  p3 = cgetr(l+1); l -= 2; l1 = 1; l2 = 3;
+  av = avma;
   for (i=1; i<=n; i++)
-  {
+  { /* execute p2 := (p2 + p1/p2)/2 */
     l0 = l1<<1;
-    if (l0<=l)
+    if (l0 <= l)
       { l2 += l1; l1=l0; }
     else
       { l2 += l+1-l1; l1=l+1; }
-    setlg(p3,l1+2); setlg(p1,l1+2); setlg(p2,l2);
-    divrrz(p1,p2,p3); addrrz(p2,p3,p2); setexpo(p2,expo(p2)-1);
+    setlg(p1, l1+2);
+    setlg(p3, l1+2);
+    setlg(p2, l2);
+    affrr(divrr(p1,p2), p3);
+    affrr(addrr(p2,p3), p2); setexpo(p2, expo(p2)-1);
+    avma = av;
+  }
+  affrr(p2,y); setexpo(y, expo(y)+ex);
+  avma = av0; return y;
+}
+
+/* O(p^e) */
+GEN
+padiczero(GEN p, long e)
+{
+  GEN y = cgetg(5,t_PADIC);
+  y[4] = zero;
+  y[3] = un;
+  copyifstack(p,y[2]);
+  y[1] = evalvalp(e) | evalprecp(0);
+  return y;
+}
+
+/* assume x unit, precp(x) = pp > 3 */
+static GEN
+sqrt_2adic(GEN x, long pp)
+{
+  GEN z = mod16(x)==1? gun: stoi(3);
+  long zp;
+  ulong av, lim;
+
+  if (pp == 4) return z;
+  zp = 3; /* number of correct bits in z (compared to sqrt(x)) */
+
+  av = avma; lim = stack_lim(av,2);
+  for(;;)
+  {
+    GEN mod;
+    zp = (zp<<1) - 1;
+    if (zp > pp) zp = pp;
+    mod = shifti(gun, zp);
+    z = addii(z, resmod2n(mulii(x, mpinvmod(z,mod)), zp));
+    z = shifti(z, -1); /* (z + x/z) / 2 */
+    if (pp == zp) return z;
+
+    if (zp < pp) zp--;
+
+    if (low_stack(lim,stack_lim(av,2)))
+    {
+      if (DEBUGMEM > 1) err(warnmem,"padic_sqrt");
+      z = gerepileuptoint(av,z);
+    }
+  }
+}
+
+/* x unit defined modulo modx = p^pp, p != 2, pp > 0 */
+static GEN
+sqrt_padic(GEN x, GEN modx, long pp, GEN p)
+{
+  GEN mod, z = mpsqrtmod(x, p);
+  long zp = 1;
+  ulong av, lim;
+
+  if (!z) err(sqrter5);
+  if (pp <= zp) return z; 
+
+  av = avma; lim = stack_lim(av,2);
+  mod = p;
+  for(;;)
+  { /* could use the hensel_lift_accel idea. Not really worth it */
+    GEN inv2;
+    zp <<= 1;
+    if (zp < pp) mod = sqri(mod); else { zp = pp; mod = modx; }
+    inv2 = shifti(mod, -1); /* = (mod + 1)/2 = 1/2 */
+    z = addii(z, resii(mulii(x, mpinvmod(z,mod)), mod));
+    z = mulii(z, inv2);
+    z = modii(z, mod); /* (z + x/z) / 2 */
+    if (pp <= zp) return z;
+
+    if (low_stack(lim,stack_lim(av,2)))
+    {
+      GEN *gptr[2]; gptr[0]=&z; gptr[1]=&mod;
+      if (DEBUGMEM>1) err(warnmem,"padic_sqrt");
+      gerepilemany(av,gptr,2);
+    }
   }
-  affrr(p2,y); setexpo(y,expo(y)+ex);
-  avma=av; return y;
 }
 
 static GEN
 padic_sqrt(GEN x)
 {
-  long av = avma, av2,lim,e,r,lpp,lp,pp;
-  GEN y;
+  long pp, e = valp(x);
+  GEN z,y,mod, p = (GEN)x[2];
+  ulong av;
 
-  e=valp(x); y=cgetg(5,t_PADIC); copyifstack(x[2],y[2]);
-  if (gcmp0(x))
-  {
-    y[4] = zero; e = (e+1)>>1;
-    y[3] = un;
-    y[1] = evalvalp(e) | evalprecp(precp(x));
-    return y;
-  }
+  if (gcmp0(x)) return padiczero(p, (e+1) >> 1);
   if (e & 1) err(sqrter6);
-  e>>=1; setvalp(y,e); y[3] = y[2];
+
+  y = cgetg(5,t_PADIC);
   pp = precp(x);
-  if (egalii(gdeux, (GEN)x[2]))
+  mod = (GEN)x[3];
+  x   = (GEN)x[4]; /* lift to int */
+  e >>= 1; av = avma;
+  if (egalii(gdeux, p))
   {
-    lp=3; y[4] = un; r = mod8((GEN)x[4]);
-    if ((r!=1 && pp>=2) && (r!=5 || pp!=2)) err(sqrter5);
-    if (pp <= lp) { setprecp(y,1); return y; }
-
-    x = dummycopy(x); setvalp(x,0); setvalp(y,0);
-    av2=avma; lim = stack_lim(av2,2);
-    y[3] = lstoi(8);
-    for(;;)
+    long r = mod8(x);
+    if (pp <= 3)
     {
-      lpp=lp; lp=(lp<<1)-1;
-      if (lp < pp) y[3] = lshifti((GEN)y[3], lpp-1);
-      else
-        { lp=pp; y[3] = x[3]; }
-      setprecp(y,lp); y = gdiv(gadd(y, gdiv(x,y)), gdeux);
-      if (lp < pp) lp--;
-      if (cmpii((GEN)y[4], (GEN)y[3]) >= 0)
-        y[4] = lsubii((GEN)y[4], (GEN)y[3]);
-
-      if (pp <= lp) break;
-      if (low_stack(lim,stack_lim(av2,2)))
-      {
-        if (DEBUGMEM > 1) err(warnmem,"padic_sqrt");
-        y = gerepileupto(av2,y);
+      switch(pp) {
+        case 1: break;
+        case 2: if ((r&3) == 1) break;
+        case 3: if (r == 1) break;
+        default: err(sqrter5);
       }
+      z = gun;
+      pp = 1;
     }
-    y = gcopy(y);
+    else
+    {
+      if (r != 1) err(sqrter5);
+      z = sqrt_2adic(x, pp);
+      z = gerepileuptoint(av, z);
+      pp--;
+    }
+    mod = shifti(gun, pp);
   }
   else /* p != 2 */
   {
-    lp=1; y[4] = (long)mpsqrtmod((GEN)x[4],(GEN)x[2]);
-    if (!y[4]) err(sqrter5);
-    if (pp <= lp) { setprecp(y,1); return y; }
-
-    x = dummycopy(x); setvalp(x,0); setvalp(y,0);
-    av2=avma; lim = stack_lim(av2,2);
-    for(;;)
-    {
-      lp<<=1;
-      if (lp < pp) y[3] = lsqri((GEN)y[3]);
-      else
-        { lp=pp; y[3] = x[3]; }
-      setprecp(y,lp); y = gdiv(gadd(y, gdiv(x,y)), gdeux);
-
-      if (pp <= lp) break;
-      if (low_stack(lim,stack_lim(av2,2)))
-      {
-        if (DEBUGMEM > 1) err(warnmem,"padic_sqrt");
-        y = gerepileupto(av2,y);
-      }
-    }
+    z = sqrt_padic(x, mod, pp, p);
+    z = gerepileuptoint(av, z);
+    mod = icopy(mod);
   }
-  setvalp(y,e); return gerepileupto(av,y);
+  y[1] = evalprecp(pp) | evalvalp(e);
+  copyifstack(p,y[2]);
+  y[3] = (long)mod;
+  y[4] = (long)z; return y;
 }
 
 GEN
@@ -980,13 +1024,8 @@ GEN padic_sqrtn(GEN x, GEN n, GEN *zetan)
   GEN *gptr[2];
   if (gcmp0(x))
   {
-    GEN y;
     long m = itos(n);
-    e=valp(x); y=cgetg(5,t_PADIC); copyifstack(x[2],y[2]);
-    y[4] = zero; e = (e+m-1)/m;
-    y[3] = un;
-    y[1] = evalvalp(e) | evalprecp(precp(x));
-    return y;
+    return padiczero(p, (valp(x)+m-1)/m);
   }
   /*First treat the ramified part using logarithms*/
   e=pvaluation(n,p,&q);
@@ -1230,57 +1269,65 @@ paexp(GEN x)
   return gerepileupto(av, y);
 }
 
+static GEN
+cxexp(GEN x, long prec)
+{
+  GEN r,p1,p2, y = cgetg(3,t_COMPLEX);
+  ulong av = avma, tetpil;
+  r=gexp((GEN)x[1],prec);
+  gsincos((GEN)x[2],&p2,&p1,prec);
+  tetpil = avma;
+  y[1] = lmul(r,p1);
+  y[2] = lmul(r,p2);
+  gerepilemanyvec(av,tetpil,y+1,2);
+  return y;
+}
+
+static GEN
+serexp(GEN x, long prec)
+{
+  ulong av;
+  long i,j,lx,ly,ex;
+  GEN p1,y,xd,yd;
+ 
+  if (gcmp0(x)) return gaddsg(1,x);
+  lx = lg(x); ex = valp(x);
+  if (ex < 0) err(negexper,"gexp");
+  if (ex)
+  {
+    ly = lx+ex; y = cgetg(ly,t_SER);
+    y[1] = evalsigne(1) | evalvalp(0) | evalvarn(varn(x));
+    /* zd[i] = coeff of X^i in z */
+    xd = x+2-ex; yd = y+2; ly -= 2;
+    yd[0] = un;
+    for (i=1; i<ex; i++) yd[i]=zero;
+    for (   ; i<ly; i++)
+    {
+      av = avma; p1 = gzero;
+      for (j=ex; j<=i; j++)
+        p1 = gadd(p1, gmulgs(gmul((GEN)xd[j],(GEN)yd[i-j]),j));
+      yd[i] = lpileupto(av, gdivgs(p1,i));
+    }
+    return y;
+  }
+  av = avma; y = cgetg(lx, t_SER);
+  y[1] = x[1]; y[2] = zero;
+  for (i=3; i <lx; i++) y[i] = x[i];
+  p1 = gexp((GEN)x[2],prec);
+  y = gmul(p1, serexp(normalize(y),prec));
+  return gerepileupto(av, y);
+}
+
 GEN
 gexp(GEN x, long prec)
 {
-  long av,tetpil,ex;
-  GEN r,y,p1,p2;
-
   switch(typ(x))
   {
-    case t_REAL:
-      return mpexp(x);
-
-    case t_COMPLEX:
-      y=cgetg(3,t_COMPLEX); av=avma;
-      r=gexp((GEN)x[1],prec);
-      gsincos((GEN)x[2],&p2,&p1,prec);
-      tetpil=avma;
-      y[1]=lmul(r,p1); y[2]=lmul(r,p2);
-      gerepilemanyvec(av,tetpil,y+1,2);
-      return y;
-
-    case t_PADIC:
-      return paexp(x);
-
-    case t_SER:
-      if (gcmp0(x)) return gaddsg(1,x);
-
-      ex=valp(x); if (ex<0) err(negexper,"gexp");
-      if (ex)
-      {
-	long i,j,ly=lg(x)+ex;
-
-	y=cgetg(ly,t_SER);
-	y[1] = evalsigne(1) | evalvalp(0) | evalvarn(varn(x));
-        y[2] = un;
-	for (i=3; i<ex+2; i++) y[i]=zero;
-	for (   ; i<ly; i++)
-	{
-	  av=avma; p1=gzero;
-	  for (j=ex; j<i-1; j++)
-	    p1=gadd(p1,gmulgs(gmul((GEN)x[j-ex+2],(GEN)y[i-j]),j));
-	  tetpil=avma; y[i]=lpile(av,tetpil,gdivgs(p1,i-2));
-	}
-	return y;
-      }
-      av=avma; p1=gcopy(x); p1[2]=zero;
-      p2=gexp(normalize(p1),prec);
-      p1=gexp((GEN)x[2],prec); tetpil=avma;
-      return gerepile(av,tetpil,gmul(p1,p2));
-
-    case t_INTMOD:
-      err(typeer,"gexp");
+    case t_REAL: return mpexp(x);
+    case t_COMPLEX: return cxexp(x,prec);
+    case t_PADIC: return paexp(x);
+    case t_SER: return serexp(x,prec);
+    case t_INTMOD: err(typeer,"gexp");
   }
   return transc(gexp,x,prec);
 }
@@ -1303,8 +1350,9 @@ gexpz(GEN x, GEN y)
 GEN
 mplog(GEN x)
 {
-  long l,l1,l2,m,m1,n,i,ex,s,sgn,ltop,av;
-  double alpha,beta,a,b;
+  ulong av, ltop;
+  long l,l1,l2,m,m1,n,i,ex,s,sgn;
+  double alpha,a,b;
   GEN y,p1,p2,p3,p4,p5,unr;
 
   if (typ(x)!=t_REAL) err(typeer,"mplog");
@@ -1314,32 +1362,33 @@ mplog(GEN x)
 
   l2 = l+1; p2=p1=cgetr(l2); affrr(x,p1);
   av=avma;
-  if (sgn > 0) p2 = divsr(1,p1); /* x<1 changer x en 1/x */
+  if (sgn > 0) p2 = divsr(1,p1); /* x<1 change x to 1/x */
   for (m1=1; expo(p2)>0; m1++) p2 = mpsqrt(p2);
   if (m1 > 1 || sgn > 0) { affrr(p2,p1); avma=av; }
 
   alpha = 1+p1[2]/C31; if (!alpha) alpha = 0.00000001;
   l -= 2; alpha = -log(alpha);
-  beta = (BITS_IN_LONG/2)*l*LOG2;
   a = alpha/LOG2;
   b = sqrt((BITS_IN_LONG/2)*l/3.0);
   if (a<=b)
   {
-    n=(long)(1+sqrt((BITS_IN_LONG/2)*3.0*l));
-    m=(long)(1+b-a);
+    n = (long)(1+sqrt((BITS_IN_LONG/2)*3.0*l));
+    m = (long)(1+b-a);
     l2 += m>>TWOPOTBITS_IN_LONG;
-    p4=cgetr(l2); affrr(p1,p4);
-    p1 = p4; av=avma;
-    for (i=1; i<=m; i++)  p1 = mpsqrt(p1);
-    affrr(p1,p4); avma=av;
+    p4 = cgetr(l2); affrr(p1,p4);
+    p1 = p4; av = avma;
+    for (i=1; i<=m; i++) p1 = mpsqrt(p1);
+    affrr(p1,p4); avma = av;
   }
   else
   {
-    n=(long)(1+beta/alpha);
-    m=0; p4 = p1;
+    double beta = (BITS_IN_LONG/2)*l*LOG2;
+    n = (long)(1+beta/alpha);
+    m = 0; p4 = p1;
   }
-  unr=realun(l2);
-  p2=cgetr(l2); p3=cgetr(l2); av=avma;
+  unr = realun(l2);
+  p2 = cgetr(l2);
+  p3 = cgetr(l2); av = avma;
 
   /* affrr needed here instead of setlg since prec may DECREASE */
   p1 = cgetr(l2); affrr(subrr(p4,unr), p1);
@@ -1347,103 +1396,105 @@ mplog(GEN x)
   p5 = addrr(p4,unr); setlg(p5,l2);
   affrr(divrr(p1,p5), p2);
   affrr(mulrr(p2,p2), p3);
-  affrr(divrs(unr,2*n+1), p4); setlg(p4,4); avma=av;
+  affrr(divrs(unr,2*n+1), p4); setlg(p4,4); avma = av;
 
-  s=0; ex=expo(p3); l1 = 4;
+  s = 0; ex = expo(p3); l1 = 4;
   for (i=n; i>=1; i--)
   {
-    setlg(p3,l1); p5 = mulrr(p4,p3);
+    setlg(p3, l1); p5 = mulrr(p4,p3);
     setlg(unr,l1); p1 = divrs(unr,2*i-1);
     s -= ex;
     l1 += s>>TWOPOTBITS_IN_LONG; if (l1>l2) l1=l2;
-    s %= BITS_IN_LONG;
-    setlg(p1,l1); setlg(p4,l1); setlg(p5,l1);
-    p1 = addrr(p1,p5); affrr(p1,p4); avma=av;
+    s &= (BITS_IN_LONG-1);
+    setlg(p1, l1);
+    setlg(p4, l1);
+    setlg(p5, l1); affrr(addrr(p1,p5), p4); avma=av;
   }
-  setlg(p4,l2); affrr(mulrr(p2,p4), y);
+  setlg(p4, l2); affrr(mulrr(p2,p4), y);
   setexpo(y, expo(y)+m+m1);
   if (sgn > 0) setsigne(y, -1);
-  avma=ltop; return y;
+  avma = ltop; return y;
 }
 
 GEN
 teich(GEN x)
 {
-  GEN aux,y,z,p1;
+  GEN p,q,y,z,aux,p1;
   long av,n,k;
 
   if (typ(x)!=t_PADIC) err(talker,"not a p-adic argument in teichmuller");
   if (!signe(x[4])) return gcopy(x);
-  y=cgetp(x); z=(GEN)x[4];
-  if (!cmpis((GEN)x[2],2))
+  p = (GEN)x[2];
+  q = (GEN)x[3];
+  z = (GEN)x[4];
+  y = cgetp(x); av = avma;
+  if (egalii(p, gdeux))
+    z = (mod4(z) & 2)? addsi(-1,q): gun;
+  else
   {
-    n = mod4(z);
-    if (n & 2)
-      addsiz(-1,(GEN)x[3],(GEN)y[4]);
-    else
-      affsi(1,(GEN)y[4]);
-    return y;
+    p1 = addsi(-1, p);
+    z = resii(z, p);
+    aux = divii(addsi(-1,q),p1); n = precp(x);
+    for (k=1; k<n; k<<=1)
+      z = modii(mulii(z,addsi(1,mulii(aux,addsi(-1,powmodulo(z,p1,q))))), q);
   }
-  av = avma; p1 = addsi(-1,(GEN)x[2]);
-  aux = divii(addsi(-1,(GEN)x[3]),p1); n = precp(x);
-  for (k=1; k<n; k<<=1)
-    z=modii(mulii(z,addsi(1,mulii(aux,addsi(-1,powmodulo(z,p1,(GEN)x[3]))))),
-            (GEN)x[3]);
-  affii(z,(GEN)y[4]); avma=av; return y;
+  affii(z, (GEN)y[4]); avma = av; return y;
 }
 
+/* Let x = 1 mod p and y := (x-1)/(x+1) = 0 (p). Then
+ * log(x) = log(1+y) - log(1-y) = 2 \sum_{k odd} y^k / k.
+ * palogaux(x) returns the last sum (not multiplied by 2) */
 static GEN
 palogaux(GEN x)
 {
-  long av1=avma,tetpil,k,e,pp;
-  GEN y,s,y2;
+  long k,e,pp;
+  GEN y,s,y2, p = (GEN)x[2];
 
-  if (egalii(gun,(GEN)x[4]))
+  if (egalii(gun, (GEN)x[4]))
   {
-    y=gaddgs(x,-1);
-    if (egalii(gdeux,(GEN)x[2]))
-    {
-      setvalp(y,valp(y)-1);
-      if (!gcmp1((GEN)y[3])) y[3]=lshifti((GEN)y[3],-1);
-    }
-    return gerepilecopy(av1,y);
+    long v = valp(x)+precp(x);
+    if (egalii(gdeux,p)) v--;
+    return padiczero(p, v);
   }
-  y=gdiv(gaddgs(x,-1),gaddgs(x,1)); e=valp(y); pp=e+precp(y);
-  if (egalii(gdeux,(GEN)x[2])) pp--;
+  y = gdiv(gaddgs(x,-1), gaddgs(x,1));
+  e = valp(y); pp = e+precp(y);
+  if (egalii(gdeux,p)) pp--;
   else
   {
-    long av=avma;
     GEN p1;
-
-    for (p1=stoi(e); cmpsi(pp,p1)>0; pp++)
-      p1 = mulii(p1,(GEN)x[2]);
-    avma=av; pp-=2;
+    for (p1=stoi(e); cmpsi(pp,p1)>0; pp++) p1 = mulii(p1, p);
+    pp -= 2;
   }
-  k=pp/e; if (!odd(k)) k--;
-  y2=gsqr(y); s=gdivgs(gun,k);
-  while (k>=3)
+  k = pp/e; if (!odd(k)) k--;
+  y2 = gsqr(y); s = gdivgs(gun,k);
+  while (k > 2)
   {
-    k-=2; s=gadd(gmul(y2,s),gdivgs(gun,k));
+    k -= 2; s = gadd(gmul(y2,s), gdivgs(gun,k));
   }
-  tetpil=avma; return gerepile(av1,tetpil,gmul(s,y));
+  return gmul(s,y);
 }
 
 GEN
 palog(GEN x)
 {
-  long av=avma,tetpil;
-  GEN p1,y;
+  ulong av = avma;
+  GEN y, p = (GEN)x[2];
 
   if (!signe(x[4])) err(talker,"zero argument in palog");
-  if (cmpis((GEN)x[2],2))
+  if (egalii(p, gdeux))
   {
-    y=cgetp(x); p1=gsubgs((GEN)x[2],1);
-    affii(powmodulo((GEN)x[4],p1,(GEN)x[3]),(GEN)y[4]);
-    y=gmulgs(palogaux(y),2); tetpil=avma;
-    return gerepile(av,tetpil,gdiv(y,p1));
+    y = gsqr(x); setvalp(y,0);
+    y = palogaux(y);
   }
-  y=gsqr(x); setvalp(y,0); tetpil=avma;
-  return gerepile(av,tetpil,palogaux(y));
+  else
+  { /* compute log(x^(p-1)) / (p-1) */
+    GEN mod = (GEN)x[3], p1 = subis(p,1);
+    y = cgetp(x);
+    y[4] = (long)powmodulo((GEN)x[4], p1, mod);
+    p1 = divii(subis(mod,1), p1); /* 1/(1-p) */
+    y = gmul(palogaux(y), mulis(p1, -2));
+  }
+  return gerepileupto(av,y);
 }
 
 GEN
