@@ -1,4 +1,4 @@
-/* $Id: gp.c,v 1.91 2001/10/17 13:32:46 karim Exp $
+/* $Id: gp.c,v 1.102 2002/01/10 23:59:19 karim Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -32,7 +32,7 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
 #include "gp.h"
 
 #ifdef READLINE
-  extern void init_readline();
+  extern void init_readline(void);
   long use_readline = 1;
   int readline_init = 1;
 BEGINEXTERN
@@ -50,7 +50,7 @@ BEGINEXTERN
 #  endif
   extern int isatty(int);
 ENDEXTERN
-static int history_is_dup(char *s);
+static void gp_add_history(char *s);
 #endif
 
 char*  _analyseur(void);
@@ -133,7 +133,7 @@ usage(char *s)
 
 /* must be called BEFORE pari_init() */
 static void
-gp_preinit()
+gp_preinit(void)
 {
   long i;
 
@@ -189,7 +189,7 @@ gp_preinit()
 static char*
 get_sep0(char *t, int colon)
 {
-  static char buf[GET_SEP_SIZE], *lim = buf + GET_SEP_SIZE-1;
+  static char buf[GET_SEP_SIZE], *lim = buf + (GET_SEP_SIZE-1);
   char *s = buf;
   int outer=1;
 
@@ -264,7 +264,7 @@ get_int(char *s, long dflt)
 
 /* tell TeXmacs GP will start outputing data */
 static void
-tm_start_output()
+tm_start_output(void)
 {
   if (!tm_is_waiting) { printf("%cverbatim:",DATA_BEGIN); fflush(stdout); }
   tm_is_waiting = 1;
@@ -272,7 +272,7 @@ tm_start_output()
 
 /* tell TeXmacs GP is done and is waiting for new data */
 static void
-tm_end_output()
+tm_end_output(void)
 {
   if (tm_is_waiting) { printf("%c", DATA_END); fflush(stdout); } 
   tm_is_waiting = 0;
@@ -302,10 +302,8 @@ print0(GEN *g, long flag)
 {
   int old=prettyp;
 
-  if (flag < NBFORMATS) added_newline=1;
-  else
-    { flag -= NBFORMATS; added_newline=0; }
-  prettyp=flag;
+  added_newline = (flag & f_NOEOL) == 0;
+  prettyp = flag & ~f_NOEOL;
 
   for( ; *g; g++)
     if (typ(*g)==t_STR)
@@ -346,7 +344,7 @@ gpwritebin(char *s, GEN x)
 }
 
 Buffer *
-new_buffer()
+new_buffer(void)
 {
   Buffer *b = (Buffer*) gpmalloc(sizeof(Buffer));
   b->len = paribufsize;
@@ -362,7 +360,7 @@ del_buffer(Buffer *b)
 }
 
 static void
-pop_buffer()
+pop_buffer(void)
 {
   Buffer *b = (Buffer*) pop_stack(&bufstack);
   del_buffer(b);
@@ -380,7 +378,7 @@ kill_all_buffers(Buffer *B)
 }
 
 static void
-jump_to_buffer()
+jump_to_buffer(void)
 {
   Buffer *b;
   while ( (b = current_buffer) )
@@ -745,7 +743,7 @@ sd_output(char *v, int flag)
   return sd_numeric(v,flag,"output",&prettyp, 0,3,msg);
 }
 
-extern void err_clean();
+extern void err_clean(void);
 
 void
 allocatemem0(unsigned long newsize)
@@ -967,12 +965,12 @@ default_type gp_default_list[] =
 };
 
 static void
-help_default()
+help_default(void)
 {
   default_type *dft;
 
   for (dft=gp_default_list; dft->fun; dft++)
-    ((void (*)(ANYARG)) dft->fun)("", d_ACKNOWLEDGE);
+    ((void (*)(char*,int)) dft->fun)("", d_ACKNOWLEDGE);
 }
 
 static GEN
@@ -985,7 +983,7 @@ setdefault(char *s,char *v, int flag)
     if (!strcmp(s,dft->name))
     {
       if (flag == d_EXISTS) return gun;
-      return ((GEN (*)(ANYARG)) dft->fun)(v,flag);
+      return ((GEN (*)(char*,int)) dft->fun)(v,flag);
     }
   if (flag == d_EXISTS) return gzero;
   err(talker,"unknown default: %s",s);
@@ -998,7 +996,7 @@ setdefault(char *s,char *v, int flag)
 /**                                                                **/
 /********************************************************************/
 static int
-has_ext_help()
+has_ext_help(void)
 {
   if (help_prg)
   {
@@ -1137,7 +1135,7 @@ print_user_member(entree *ep)
 }
 
 static void
-user_fun()
+user_fun(void)
 {
   entree *ep;
   int hash;
@@ -1152,7 +1150,7 @@ user_fun()
 }
 
 static void
-user_member()
+user_member(void)
 {
   entree *ep;
   int hash;
@@ -1179,7 +1177,7 @@ center(char *s)
 }
 
 static void
-community()
+community(void)
 {
   long len = strlen(GPMISCDIR) + 1024;
   char *s = gpmalloc(len);
@@ -1605,7 +1603,7 @@ print_hash_list(char *s)
 }
 
 static char *
-what_readline()
+what_readline(void)
 {
 #ifdef READLINE
   if (use_readline)
@@ -1616,7 +1614,7 @@ what_readline()
 }
 
 static void
-print_version()
+print_version(void)
 {
   char buf[64];
 
@@ -1627,10 +1625,10 @@ print_version()
 }
 
 static void
-gp_head()
+gp_head(void)
 {
   print_version(); pariputs("\n");
-  center("Copyright (C) 2000 The PARI Group");
+  center("Copyright (C) 2002 The PARI Group");
   print_text("\nPARI/GP is free software, covered by the GNU General Public \
 License, and comes WITHOUT ANY WARRANTY WHATSOEVER");
   pariputs("\n\
@@ -1650,7 +1648,7 @@ fix_buffer(Buffer *b, long newlbuf)
 }
 
 void
-gp_quit()
+gp_quit(void)
 {
   free_graph(); freeall();
   kill_all_buffers(NULL);
@@ -1703,7 +1701,7 @@ texmacs_output(GEN z, long n)
  * the output.  Fill the output buffer, wait until it is read.
  * Better than sleep(2): give possibility to print */
 static void
-prettyp_wait()
+prettyp_wait(void)
 {
   char *s = "                                                     \n";
   int i = 400;
@@ -1715,7 +1713,7 @@ prettyp_wait()
 
 /* initialise external prettyprinter (tex2mail) */
 static int
-prettyp_init()
+prettyp_init(void)
 {
   if (!prettyprinter_file)
     prettyprinter_file = try_pipe(prettyprinter, mf_OUT | mf_TEST);
@@ -1895,7 +1893,7 @@ escape0(char *tch)
         {
           long i, l = lg(x);
           err(warner,"setting %ld history entries", l-1);
-          for (i=1; i<l; i++) set_hist_entry((GEN)x[i]);
+          for (i=1; i<l; i++) (void)set_hist_entry((GEN)x[i]);
         }
       }
       break;
@@ -1986,7 +1984,7 @@ gprc_chk(char *s)
 
 /* Look for [._]gprc: $GPRC, then in $HOME, /, C:/ */
 static FILE *
-gprc_get()
+gprc_get(void)
 {
   FILE *f = NULL;
   char *str, *s, c;
@@ -2024,7 +2022,7 @@ static int get_line_from_file(char *prompt, Buffer *b, FILE *file);
 #define err_gprc(s,t,u) { fprintferr("\n"); err(talker2,s,t,u); }
 
 static char **
-gp_initrc()
+gp_initrc(void)
 {
   char **flist, *s,*s1,*s2;
   FILE *file = gprc_get();
@@ -2140,7 +2138,7 @@ do_time(long flag)
 }
 
 static void
-gp_handle_SIGINT()
+gp_handle_SIGINT(void)
 {
 #ifdef _WIN32
   if (++win32ctrlc >= 5) _exit(3);
@@ -2233,7 +2231,7 @@ do_prompt(char *p)
 }
 
 static void
-unblock_SIGINT()
+unblock_SIGINT(void)
 {
 #ifdef USE_SIGRELSE
   sigrelse(SIGINT);
@@ -2298,7 +2296,8 @@ gprl_input(Buffer *b, char **endp, char *prompt)
 
 #define ask_filtre(t) filtre("",NULL,t)
 
-static int				/* True if more than one line read */
+/* True if more than one line read */
+static int
 input_loop(Buffer *b, char *already_read, FILE *file, char *prompt)
 {
   const int TeXmacs = (under_texmacs && file == stdin);
@@ -2345,8 +2344,7 @@ input_loop(Buffer *b, char *already_read, FILE *file, char *prompt)
     if (!file) {
       free(already_read);
       already_read = gprl_input(b, &s, do_prompt(prompt_cont));
-      if (!history_is_dup(already_read))
-        add_history(already_read);	/* Makes a copy */
+      gp_add_history(already_read);	/* Makes a copy */
     } else
 #endif
       already_read = file_input(b,&s,file,TeXmacs);
@@ -2400,10 +2398,16 @@ get_line_from_user(char *prompt, Buffer *b)
 }
 #else	/* READLINE */
 static int
-history_is_dup(char *s)
+history_is_new(char *s)
 {
-  if (!history_length) return 0;
-  return !strcmp(s, history_get(history_length)->line);
+  return (*s && (!history_length ||
+                  strcmp(s, history_get(history_length)->line)));
+}
+
+static void
+gp_add_history(char *s)
+{
+  if (history_is_new(s)) add_history(s);
 }
 
 static int
@@ -2420,7 +2424,7 @@ get_line_from_user(char *prompt, Buffer *b)
     }
     /* Put the original read line into history */
     index = history_length;
-    if (!history_is_dup(buf)) add_history(buf);	/* Copies the entry */
+    gp_add_history(buf);	/* Copies the entry */
 
     added = input_loop(b,buf,NULL,prompt); /* free()s buf */
     unblock_SIGINT(); /* bug in readline 2.0: need to unblock ^C */
@@ -2434,7 +2438,7 @@ get_line_from_user(char *prompt, Buffer *b)
           HIST_ENTRY *e = remove_history(--i);
           free(e->line); free(e);
         }
-        if (!history_is_dup(s)) add_history(s);
+        gp_add_history(s);
       }
     
       /* update logfile */
@@ -2451,7 +2455,7 @@ get_line_from_user(char *prompt, Buffer *b)
 #endif
 
 static int
-is_interactive()
+is_interactive(void)
 {
 #if defined(UNIX) || defined(__EMX__)
   return (infile == stdin && !under_texmacs
@@ -2471,16 +2475,17 @@ read_line(char *promptbuf, Buffer *b)
     return get_line_from_file(DFT_PROMPT,b,infile);
 }
 
-static void
+static int
 chron(char *s)
 {
   if (*s)
-  {
-    char *old = s-1;
-    if (*s == '#') { pariputs(do_time(ti_LAST)); s++; }
-    if (*s) err(caracer1,s,old);
+  { /* if "#" or "##" timer metacommand. Otherwise let the parser get it */
+    if (*s == '#') s++;
+    if (*s) return 0;
+    pariputs(do_time(ti_LAST));
   }
   else { chrono = 1-chrono; sd_timer("",d_ACKNOWLEDGE); }
+  return 1;
 }
 
 static int
@@ -2489,7 +2494,7 @@ check_meta(char *buf)
   switch(*buf++)
   {
     case '?': aide(buf, h_REGULAR); break;
-    case '#': chron(buf); break;
+    case '#': return chron(buf);
     case '\\': escape(buf); break;
     case '\0': return 2;
     default: return 0;
@@ -2604,7 +2609,7 @@ extern0(char *s)
 }
 
 static int
-silent()
+silent(void)
 {
   if (gpsilent) return 1;
   { char c = _analyseur()[1]; return separe(c); }
@@ -2620,7 +2625,7 @@ default0(char *a, char *b, long flag)
 }
 
 GEN
-input0()
+input0(void)
 {
   Buffer *b = new_buffer();
   GEN x;
@@ -2760,6 +2765,7 @@ read_opt(long argc, char **argv)
 {
   char *b=NULL, *p=NULL, *s=NULL, **pre;
   int i=1, initrc=1;
+  long TOP;
 
   pari_outfile=stderr;
   while (i<argc)
@@ -2797,7 +2803,8 @@ read_opt(long argc, char **argv)
   /* override the values from gprc */
   testint(b, &paribufsize); if (paribufsize < 10) paribufsize = 10;
   testint(p, &primelimit);
-  testint(s, &top);
+  TOP = (long)top;
+  testint(s, &TOP); top = (ulong)TOP;
   if (under_emacs || under_texmacs) disable_color=1;
   pari_outfile=stdout; return pre;
 }
