@@ -1,4 +1,4 @@
-/* $Id: gp_rl.c,v 1.12 2001/10/16 16:33:43 karim Exp $
+/* $Id: gp_rl.c,v 1.14 2001/12/05 19:37:13 karim Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -23,7 +23,7 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
 #include "gp.h"
 
 #ifdef READLINE
-typedef char** (*CF)(char*, char* (*)()); /* completion function */
+typedef char** (*CF)(char*, char* (*)(void)); /* completion function */
 typedef char* (*GF)(char*, int); /* generator function */
 typedef int (*RLCI)(int, int); /* rl_complete and rl_insert functions */
 
@@ -47,8 +47,6 @@ extern char *filename_completion_function(char *text,int state);
 extern char *username_completion_function(char *text,int state);
 #endif
 char **pari_completion(char *text, int start, int end);
-extern int rl_completion_query_items;
-extern int rl_bind_key_in_map();
 ENDEXTERN
 
 void print_fun_list(char **matches, int nbli);
@@ -302,7 +300,7 @@ get_matches(int end, char *text, char* f(char*,int))
   rl_completion_append_character = ' ';
 #endif
   current_ep = NULL;
-  matches = COMPLETION_MATCHES(text, (char *(*)())f);
+  matches = COMPLETION_MATCHES(text, (char *(*)(void))f);
   if (matches && !matches[1]) /* a single match */
   {
     if (add_paren(end))
@@ -537,7 +535,7 @@ pari_completion(char *text, int START, int END)
     while (j <= END && isspace((int)rl_line_buffer[j])) j++;
     k = END;
     while (k > j && isspace((int)rl_line_buffer[k])) k--;
-    /* If we are in empty parens, insert arguments for the function: */
+    /* If we are in empty parens, output function help */
     if (do_args_complete && k == j
          && (rl_line_buffer[j] == ')' || !rl_line_buffer[j])
 	 && (iend - i < MAX_KEYWORD)
@@ -545,25 +543,6 @@ pari_completion(char *text, int START, int END)
 	      buf[iend - i] = 0, 1)
 	 && (ep = is_entry(buf)) && ep->help)
      {
-#if 0
-      char *s = ep->help;
-
-      while (is_keyword_char(*s)) s++;
-      if (*s++ == '(')
-      { /* Function, print arguments! */
-        char *endh = s;
-        while (*endh && *endh != ')' && *endh != '(') endh++;
-        if (*endh == ')')
-        { /* Well-formed help.  */
-          char *str = strncpy((char*) gpmalloc(endh-s + 1), s, endh-s);
-          char **ret = (char**)gpmalloc(sizeof(char*)*2);
-          str[endh-s] = 0;
-          ret[0] = str; ret[1] = NULL;
-          if (under_emacs) ret = matches_for_emacs("",ret);
-          return ret;
-        }
-      }
-#endif
       rl_print_aide(buf,h_RL);
       rl_attempted_completion_over = 1;
       return NULL;
@@ -620,7 +599,7 @@ rl_long_help(int count, int key)
 }
 
 void
-init_readline()
+init_readline(void)
 {
   /* Allow conditional parsing of the ~/.inputrc file. */
   rl_readline_name = "Pari-GP";
