@@ -1,4 +1,4 @@
-/* $Id: plotport.c,v 1.15 2001/03/26 13:52:07 karim Exp $
+/* $Id: plotport.c,v 1.17 2001/12/05 19:37:13 karim Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -26,7 +26,7 @@ extern void pop_val(entree *ep);
 extern void postdraw0(long *w, long *x, long *y, long lw);
 extern void postdraw00(long *w, long *x, long *y, long lw, long scale);
 extern void rectdraw0(long *w, long *x, long *y, long lw, long do_free);
-static void PARI_get_psplot();
+static void PARI_get_psplot(void);
 
 static long current_color[NUMRECT];
 PariRect **rectgraph = NULL;
@@ -186,7 +186,7 @@ plot(entree *ep, GEN a, GEN b, char *ch,GEN ysmlu,GEN ybigu, long prec)
 /**                                                                **/
 /********************************************************************/
 void
-init_graph()
+init_graph(void)
 {
   int n;
 
@@ -203,7 +203,7 @@ init_graph()
 }
 
 void
-free_graph()
+free_graph(void)
 {
   int i;
 
@@ -309,22 +309,6 @@ rectscale(long ne, GEN x1, GEN x2, GEN y1, GEN y2)
   rectscale0(ne, gtodouble(x1), gtodouble(x2), gtodouble(y1), gtodouble(y2));
 }
 
-/*
- * #ifdef LONG_IS_64BIT
- * #  define MAXDTOL 9223372036854775807.0
- * #else
- * #  define MAXDTOL 2147483647.0
- * #endif
- *
- *
- * long
- * DTOL(double t)
- * {
- *   if (fabs(t) > MAXDTOL) err(affer2);
- *   return (long)t;
- * }
- */
-
 static void
 rectmove0(long ne, double x, double y, long relative)
 {
@@ -1844,7 +1828,7 @@ postploth2(entree *ep, GEN a, GEN b, char *ch, long prec,
 }
 
 GEN
-plothsizes()
+plothsizes(void)
 {
   return plothsizes_flag(0);
 }
@@ -1895,7 +1879,7 @@ static void ps_string(FILE *psfile, int x, int y, char *c, long dir);
 #define JSCR 800  /* 1120 en haute resolution */
 
 static void
-PARI_get_psplot()
+PARI_get_psplot(void)
 {
   pari_psplot.height = JSCR - 60;
   pari_psplot.width  = ISCR - 40;
