@@ -1,4 +1,4 @@
-/* $Id: paridecl.h,v 1.124 2001/10/18 15:02:41 karim Exp $
+/* $Id: paridecl.h,v 1.137 2001/12/14 17:42:08 karim Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -126,6 +126,7 @@ GEN     matrixqz3(GEN x);
 GEN     signat(GEN a);
 GEN     smith(GEN x);
 GEN     smith2(GEN x);
+GEN     smithall(GEN x, GEN *ptU, GEN *ptV);
 GEN     smithclean(GEN z);
 GEN     sqred(GEN a);
 GEN     sqred1(GEN a);
@@ -319,6 +320,7 @@ GEN     mathnf0(GEN x,long flag);
 GEN     matsnf0(GEN x,long flag);
 long    nf_get_r1(GEN nf);
 long    nf_get_r2(GEN nf);
+void    nf_get_sign(GEN nf, long *r1, long *r2);
 long    nfgetprec(GEN x);
 GEN     nfinit0(GEN x,long flag, long prec);
 GEN     nfnewprec(GEN nf, long prec);
@@ -407,6 +409,7 @@ GEN     nfreducemodidele(GEN nf,GEN g,GEN idele,GEN structarch);
 GEN     reducemodinvertible(GEN x, GEN y);
 GEN     reducemodmatrix(GEN x, GEN y);
 GEN     reducemodHNF(GEN x, GEN y, GEN *Q);
+GEN     smithrel(GEN H, GEN *newU, GEN *newUi);
 GEN     zarchstar(GEN nf,GEN x,GEN arch,long nba);
 GEN     zideallog(GEN nf,GEN x,GEN bigideal);
 GEN     zidealstar(GEN nf, GEN x);
@@ -660,7 +663,7 @@ GEN     rayclassnolist(GEN bnf,GEN listes);
 GEN     rnfconductor(GEN bnf, GEN polrel, long flag);
 GEN     rnfkummer(GEN bnr, GEN subgroup, long all, long prec);
 GEN     rnfnormgroup(GEN bnr, GEN polrel);
-GEN     subgrouplist0(GEN bnr, long indexbound, long all);
+GEN     subgrouplist0(GEN bnr, GEN indexbound, long all);
 
 /* buch4.c */
 
@@ -840,6 +843,7 @@ GEN     matsize(GEN x);
 GEN     normalize(GEN x);
 GEN     normalizepol(GEN x);
 GEN     normalizepol_i(GEN x, long lx);
+long    polvaluation(GEN x, GEN *z);
 GEN     pureimag(GEN x);
 long    pvaluation(GEN x, GEN p, GEN *py);
 long    sizedigit(GEN x);
@@ -863,6 +867,7 @@ GEN     denom(GEN x);
 GEN     deriv(GEN x, long v);
 GEN     derivpol(GEN x);
 GEN     derivser(GEN x);
+GEN     divrem(GEN x, GEN y, long v);
 GEN     gand(GEN x, GEN y);
 GEN     gceil(GEN x);
 GEN     gcvtoi(GEN x, long *e);
@@ -924,6 +929,7 @@ GEN     mulmat_real(GEN x, GEN y);
 GEN     numer(GEN x);
 long    padicprec(GEN x, GEN p);
 GEN     polcoeff0(GEN x,long n,long v);
+GEN     polcoeff_i(GEN x, long n, long v);
 long    poldegree(GEN x,long v);
 GEN     poleval(GEN x, GEN y);
 GEN     pollead(GEN x,long v);
@@ -1041,6 +1047,7 @@ GEN     mulss(long x, long y);
 GEN     mului(ulong x, GEN y);
 GEN     muluu(ulong x, ulong y);
 int     ratlift(GEN x, GEN m, GEN *a, GEN *b, GEN amax, GEN bmax);
+GEN     resmod2n(GEN x, long n);
 GEN     resss(long x, long y);
 double  rtodbl(GEN x);
 GEN     shifti(GEN x, long n);
@@ -1107,9 +1114,11 @@ GEN     centermod(GEN x, GEN p);
 GEN     content(GEN x);
 GEN     discsr(GEN x);
 GEN     divide_conquer_prod(GEN x, GEN (*mul)(GEN,GEN));
+GEN     leftright_pow(GEN,GEN,void*,GEN (*sqr)(void*,GEN),GEN (*mul)(void*,GEN,GEN));
 GEN     factor(GEN x);
 GEN     factor0(GEN x,long flag);
 GEN     factorback(GEN fa,GEN nf);
+GEN     factorback0(GEN fa,GEN e, GEN nf);
 GEN     factpol(GEN x, long hint);
 GEN     gbezout(GEN x, GEN y, GEN *u, GEN *v);
 GEN     gcd0(GEN x, GEN y,long flag);
@@ -1133,6 +1142,7 @@ GEN     polinflate(GEN x0, long d);
 GEN     polresultant0(GEN x, GEN y,long v,long flag);
 GEN     polsym(GEN x, long n);
 GEN     primitive_part(GEN x, GEN *c);
+GEN     primpart(GEN x);
 GEN     quadgen(GEN x);
 GEN     quadpoly(GEN x);
 GEN     quadpoly0(GEN x, long v);
@@ -1212,6 +1222,7 @@ GEN     quickmul(GEN a, GEN b, long na, long nb);
 GEN     quicksqr(GEN a, long na);
 GEN     small_to_mat(GEN z);
 GEN     small_to_pol(GEN z, long v);
+GEN     small_to_vec(GEN z);
 GEN     pol_to_small(GEN x);
 GEN     rescale_pol(GEN P, GEN h);
 GEN     stopoly(long m, long p, long v);
@@ -1241,8 +1252,8 @@ GEN     conjugates(GEN pol);
 
 /* subgroup.c */
 
-void    forsubgroup(entree *oep, GEN cyc, long bound, char *och);
-GEN     subgrouplist(GEN cyc, long bound);
+void    forsubgroup(entree *oep, GEN cyc, GEN bound, char *och);
+GEN     subgrouplist(GEN cyc, GEN bound);
 
 /* stark.c */
 
@@ -1282,6 +1293,7 @@ GEN     sumpos0(entree *ep, GEN a, char *ch, long flag,long prec);
 GEN     sumpos2(entree *ep, GEN a, char *ch, long prec);
 GEN     suminf(entree *ep, GEN a, char *ch, long prec);
 GEN     vecteur(GEN nmax, entree *ep, char *ch);
+GEN     vecteursmall(GEN nmax, entree *ep, char *ch);
 GEN     vvecteur(GEN nmax, entree *ep, char *ch);
 GEN     zbrent(entree *ep, GEN a, GEN b, char *ch, long prec);
 
@@ -1319,6 +1331,7 @@ GEN     mpexp1(GEN x);
 GEN     mplog(GEN x);
 GEN     mppi(long prec);
 GEN     mpsqrt(GEN x);
+GEN     padiczero(GEN p, long e);
 GEN     palog(GEN x);
 GEN     powgi(GEN x, GEN n);
 GEN     teich(GEN x);
