@@ -1,4 +1,4 @@
-/* $Id: compat.c,v 1.16 2001/09/18 11:04:50 karim Exp $
+/* $Id: compat.c,v 1.17 2001/12/05 19:37:14 karim Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -46,7 +46,7 @@ static GEN
 _factpol(GEN x, long unused, long hint) { return factpol(x,hint); }
 
 static void
-suppressed(ANYARG) {err(talker,"this function has been suppressed");}
+suppressed(void) {err(talker,"this function has been suppressed");}
 
 entree oldfonctions[]={
 {"O",50,NULL,7,NULL},
