@@ -1,4 +1,4 @@
-/* $Id: es.c,v 1.66 2001/10/18 12:25:00 karim Exp $
+/* $Id: es.c,v 1.70 2001/12/05 19:37:14 karim Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -30,10 +30,10 @@ static char format;
 static long decimals, chmp, initial;
 
 /* output a space or do nothing depending on original caller */
-static void (*sp)();
+static void (*sp)(void);
 
 void
-hit_return()
+hit_return(void)
 {
   int c;
   if (under_texmacs || under_emacs) return;
@@ -342,7 +342,7 @@ static int col_index, lin_index, max_width, max_lin;
 #endif
 
 static int
-term_width_intern()
+term_width_intern(void)
 {
 #ifdef HAS_TIOCGWINSZ
   {
@@ -367,7 +367,7 @@ term_width_intern()
 }
 
 static int
-term_height_intern()
+term_height_intern(void)
 {
 #ifdef HAS_TIOCGWINSZ
   {
@@ -395,14 +395,14 @@ term_height_intern()
 #define DFT_TERM_HEIGHT 20
 
 int
-term_width()
+term_width(void)
 {
   int n = term_width_intern();
   return (n>1)? n: DFT_TERM_WIDTH;
 }
 
 int
-term_height()
+term_height(void)
 {
   int n = term_height_intern();
   return (n>1)? n: DFT_TERM_HEIGHT;
@@ -649,8 +649,8 @@ GENtostr(GEN x) { return GENtostr0(x,outbrute); }
 #define sp_plus_sp() sp(), pariputc('+'), sp()
 #define comma_sp() pariputc(','), sp()
 
-static void wr_space() {pariputc(' ');}
-static void no_space() {}
+static void wr_space(void) {pariputc(' ');}
+static void no_space(void) {}
 
 static void
 blancs(long nb) { while (nb-- > 0) pariputc(' '); }
@@ -740,13 +740,13 @@ static void
 wr_vecsmall(GEN g)
 {
   long i,l;
-  pariputc('['); l = lg(g);
+  pariputs("Vecsmall(["); l = lg(g);
   for (i=1; i<l; i++)
   {
     pariputsf("%ld", g[i]);
     if (i<l-1) comma_sp();
   }
-  pariputc(']');
+  pariputs("])");
 }
 /********************************************************************/
 /**                                                                **/
@@ -931,12 +931,12 @@ voir2(GEN x, long nb, long bl)
   sorstring(VOIR_STRING2,x[0]);
   if (! is_recursive_t(tx)) /* t_SMALL, t_INT, t_REAL, t_STR, t_VECSMALL */
   {
-    if (tx == t_STR)
-	pariputs("chars:");
+    if (tx == t_STR) 
+      pariputs("chars:");
     else if (tx == t_INT)
-	pariputsf("(%c,lgef=%ld):", vsigne(x), lgefint(x));
+      pariputsf("(%c,lgef=%ld):", vsigne(x), lgefint(x));
     else if (tx == t_REAL)
-	pariputsf("(%c,expo=%ld):", vsigne(x), expo(x));
+      pariputsf("(%c,expo=%ld):", vsigne(x), expo(x));
     if (nb<0) nb = (tx==t_INT)? lgefint(x): lx;
     if (tx == t_VECSMALL) nb = lx;
     for (i=1; i < nb; i++) sorstring(VOIR_STRING2,x[i]);
@@ -1448,6 +1448,7 @@ bruti(GEN g, long nosign)
     case t_PADIC:
     {
       GEN p = (GEN)g[2];
+      ulong av = avma;
       i = valp(g); l = precp(g)+i;
       g = (GEN)g[4]; v = GENtostr(p);
       for (; i<l; i++)
@@ -1462,6 +1463,7 @@ bruti(GEN g, long nosign)
 	  if (i) padic_nome(v,i);
           sp_plus_sp();
 	}
+        if ((i & 0xfff) == 0) g = gerepileuptoint(av,g);
       }
       pariputs("O("); padic_nome(v,i); pariputc(')');
       free(v); break;
@@ -1912,7 +1914,7 @@ void
 bruteall(GEN g, char f, long d, long flag)
 {
   long av = avma;
-  void (*oldsp)() = sp;
+  void (*oldsp)(void) = sp;
 
   sp = flag? &wr_space: &no_space;
   format = f; decimals = d;
@@ -2129,7 +2131,7 @@ pari_unlink(char *s)
  * Check for leaked file handlers (temporary files)
  */
 int
-popinfile()
+popinfile(void)
 {
   pariFILE *f;
 
@@ -2840,7 +2842,7 @@ env_ok(char *s)
 }
 
 static char*
-pari_tmp_dir()
+pari_tmp_dir(void)
 {
   char *s;
 #ifdef WINCE
