@@ -1,4 +1,4 @@
-/* $Id: helpmsg.c,v 1.32 2001/10/15 19:01:33 karim Exp $
+/* $Id: helpmsg.c,v 1.34 2001/11/09 00:48:20 karim Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -113,7 +113,7 @@ char *helpmessages_basic[]={
   "dirmul(x,y): multiplication of the Dirichlet series x by the Dirichlet series y",
   "dirzetak(nf,b): Dirichlet series of the Dedekind zeta function of the number field nf up to the bound b-1",
   "divisors(x): gives a vector formed by the divisors of x in increasing order",
-  "divrem(x,y): euclidean division of x by y giving as a 2-dimensional column vector the quotient and the remainder",
+  "divrem(x,y,{v}): euclidean division of x by y giving as a 2-dimensional column vector the quotient and the remainder, with respect to v (to main variable if v is omitted)",
   "eint1(x,{n}): exponential integral E1(x). If n is present, computes the vector of the first n values of the exponential integral E1(n.x) (x > 0)",
   "elladd(e,z1,z2): sum of the points z1 and z2 on elliptic curve e",
   "ellak(e,n): computes the n-th Fourier coefficient of the L-function of the elliptic curve e",
@@ -462,6 +462,7 @@ char *helpmessages_basic[]={
   "vecmin(x): minimum of the elements of the vector/matrix x",
   "vecsort(x,{k},{flag=0}): sorts the vector of vectors (or matrix) x, according to the value of its k-th component if k is not omitted. Binary digits of flag (if present) mean: 1: indirect sorting, return the permutation instead of the permuted vector, 2: sort using ascending lexicographic order",
   "vector(n,{X},{expr=0}): row vector with n components of expression expr (X ranges from 1 to n). By default, fill with 0s",
+  "vectorsmall(n,{X},{expr=0}): VECSMALL with n components of expression expr (X ranges from 1 to n) which must be small integers. By default, fill with 0s",
   "vectorv(n,{X},{expr=0}): column vector with n components of expression expr (X ranges from 1 to n). By default, fill with 0s",
   "weber(x,{flag=0}): One of Weber's f function of x. flag is optional, and can be 0: default, function f(x)=exp(-i*Pi/24)*eta((x+1)/2)/eta(x) such that (j=(f^24-16)^3/f^24), 1: function f1(x)=eta(x/2)/eta(x) such that (j=(f1^24+16)^3/f2^24), 2: function f2(x)=sqrt(2)*eta(2*x)/eta(x) such that (j=(f2^24+16)^3/f2^24)",
   "while(a,seq):  while a is nonzero evaluate the expression sequence seq. Otherwise 0",
