@@ -1,4 +1,4 @@
-/* $Id: init.c,v 1.107 2001/10/17 22:20:12 karim Exp $
+/* $Id: init.c,v 1.119 2002/01/09 12:04:52 karim Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -170,7 +170,7 @@ gprealloc(void *pointer, size_t size)
 }
 
 static void
-pari_handle_SIGINT()
+pari_handle_SIGINT(void)
 {
 #ifdef _WIN32
   if (++win32ctrlc >= 5) _exit(3);
@@ -424,7 +424,7 @@ reset_traps(int warn)
 }
 
 static void
-init_universal_constants()
+init_universal_constants(void)
 {
   /* 2 (gnil) + 2 (gzero) + 3 (gun) + 3 (gdeux) + 3 (half) + 3 (gi) */
   GEN p = universal_constants = (long*) gpmalloc(16*sizeof(long));
@@ -992,7 +992,7 @@ err_leave_default(long n)
 
 /* untrapped error: remove all traps depending from a longjmp */
 void
-err_clean()
+err_clean(void)
 {
   stack *s = err_catch_stack, *lasts = NULL;
   cell *c;
@@ -1135,14 +1135,14 @@ err(long numerr, ...)
       case operi: case operf:
       {
         char *f, *op = va_arg(ap, char*);
-        long   x = va_arg(ap, long);
-        long   y = va_arg(ap, long);
+        GEN x = va_arg(ap, GEN);
+        GEN y = va_arg(ap, GEN);
              if (*op == '+') f = "addition";
         else if (*op == '*') f = "multiplication";
-        else if (*op == '/' || *op == '%') f = "division";
+        else if (*op == '/' || *op == '%' || *op == '\\') f = "division";
         else if (*op == 'g') { op = ","; f = "gcd"; }
         else { op = "-->"; f = "assignment"; }
-        pariputsf(" %s %s %s %s.",f,type_name(x),op,type_name(y));
+        pariputsf(" %s %s %s %s.",f,type_name(typ(x)),op,type_name(typ(y)));
         break;
       }
 
@@ -1398,7 +1398,7 @@ shiftaddress(GEN x, long dec)
   }
 }
 
-/* return a clone of x structured as a gcopy. To free it, free(base). */
+/* return a clone of x structured as a gcopy */
 GENbin*
 copy_bin(GEN x)
 {
@@ -1438,7 +1438,7 @@ bin_copy(GENbin *p)
  *   stackdummy(z,l); z += l; We lose l words but save a costly gerepile.
  */
 void
-stackdummy(GEN z, long l) { z[0] = evaltyp(t_STR) | evallg(l); }
+stackdummy(GEN z, long l) { z[0] = evaltyp(t_VECSMALL) | evallg(l); }
 
 /* gerepileupto(av, forcecopy(x)) */
 GEN
@@ -1677,10 +1677,10 @@ checkmemory(GEN z)
 #endif
 
 void
-fill_stack()
+fill_stack(void)
 {
   GEN x = ((GEN)bot);
-  while (x < (GEN)avma) *x++ = 0xfefefefe;
+  while (x < (GEN)avma) *x++ = 0xfefefefeUL;
 }
 
 /*******************************************************************/
@@ -1781,7 +1781,7 @@ fill_stack()
 
 #define is_valid_timer(t) ((t) < MAX_TIMER || (t) >= MIN_TIMER)
 long
-gptimer() {return timer_proto(0);}
+gptimer(void) {return timer_proto(0);}
 long
 timer(void)   {return timer_proto(1);}
 long
@@ -1913,13 +1913,11 @@ geni(void) { return gi; }
  *     The unquoted components can be of any pari type (converted according to
  *     the current output format)
  *  s* any number of strings (see s)
- *  s*p idem, setting prettyp=1
- *  s*t idem, in TeX format.
  *  D  Has a default value. Format is "Dvalue,type," (the ending comma is
  *     mandatory). Ex: D0,L, (arg is long, 0 by default).
  *     Special syntax:
  *       if type = G, &, I or V:  D[G&IV] all send NULL.
- *       if type = v: Dn sends -1.
+ *       if type = n: Dn sends -1.
  *
  *     The user-given args are read first, then completed by the defaults
  *
@@ -2047,7 +2045,7 @@ entree functions_basic[]={
 {"dirmul",2,(void*)dirmul,4,"GG"},
 {"dirzetak",2,(void*)dirzetak,6,"GG"},
 {"divisors",18,(void*)divisors,4,"G"},
-{"divrem",2,(void*)gdiventres,1,"GG"},
+{"divrem",2,(void*)divrem,1,"GGDn"},
 {"eint1",99,(void*)veceint1,3,"GDGp"},
 {"elladd",3,(void*)addell,5,"GGG"},
 {"ellak",2,(void*)akell,5,"GG"},
@@ -2084,7 +2082,7 @@ entree functions_basic[]={
 {"eval",18,(void*)geval,7,"G"},
 {"exp",1,(void*)gexp,3,"Gp"},
 {"factor",99,(void*)factor0,4,"GD-1,L,"},
-{"factorback",99,(void*)factorback,4,"GDG"},
+{"factorback",99,(void*)factorback0,4,"GDGDG"},
 {"factorcantor",2,(void*)factcantor,4,"GG"},
 {"factorff",3,(void*)factmod9,4,"GGG"},
 {"factorial",99,(void*)mpfactr,4,"Lp"},
@@ -2099,7 +2097,7 @@ entree functions_basic[]={
 {"fordiv",84,(void*)fordiv,11,"vGVI"},
 {"forprime",83,(void*)forprime,11,"vV=GGI"},
 {"forstep",86,(void*)forstep,11,"vV=GGGI"},
-{"forsubgroup",99,(void*)forsubgroup,11,"vV=GD0,L,I"},
+{"forsubgroup",99,(void*)forsubgroup,11,"vV=GDGI"},
 {"forvec",87,(void*)forvec,11,"vV=GID0,L,"},
 {"frac",18,(void*)gfrac,2,"G"},
 {"galoisfixedfield",99,(void*)galoisfixedfield,6,"GGD0,L,Dn"},
@@ -2371,7 +2369,7 @@ entree functions_basic[]={
 {"sqrt",1,(void*)gsqrt,3,"Gp"},
 {"sqrtint",18,(void*)racine,4,"G"},
 {"sqrtn",99,(void*)gsqrtn,3,"GGD&p"},
-{"subgrouplist",99,(void*)subgrouplist0,6,"GD0,L,D0,L,p"},
+{"subgrouplist",99,(void*)subgrouplist0,6,"GDGD0,L,p"},
 {"subst",99,(void*)gsubst0,7,"GGG"},
 {"sum",48,(void*)somme,9,"V=GGEDG"},
 {"sumalt",99,(void*)sumalt0,9,"V=GED0,L,p"},
@@ -2396,6 +2394,7 @@ entree functions_basic[]={
 {"vecmin",18,(void*)vecmin,1,"G"},
 {"vecsort",99,(void*)vecsort0,8,"GDGD0,L,"},
 {"vector",28,(void*)vecteur,8,"GDVDI"},
+{"vectorsmall",28,(void*)vecteursmall,8,"GDVDI"},
 {"vectorv",28,(void*)vvecteur,8,"GDVDI"},
 {"weber",99,(void*)weber0,3,"GD0,L,p"},
 {"while",81,NULL,11,NULL},
