@@ -1,4 +1,4 @@
-/* $Id: sumiter.c,v 1.25 2001/10/12 14:02:07 karim Exp $
+/* $Id: sumiter.c,v 1.27 2001/11/16 14:25:14 karim Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -656,6 +656,32 @@ vecteur(GEN nmax, entree *ep, char *ch)
   pop_val(ep); return y;
 }
 
+GEN
+vecteursmall(GEN nmax, entree *ep, char *ch)
+{
+  GEN y,p1;
+  long i,m;
+  long c[]={evaltyp(t_INT)|m_evallg(3), evalsigne(1)|evallgefint(3), 0};
+
+  if (typ(nmax)!=t_INT || signe(nmax) < 0)
+    err(talker,"bad number of components in vector");
+  m=itos(nmax); y=cgetg(m+1,t_VECSMALL);
+  if (!ep || !ch)
+  {
+    for (i=1; i<=m; i++) y[i] = 0;
+    return y;
+  }
+  push_val(ep, c);
+  for (i=1; i<=m; i++)
+  {
+    c[2]=i; p1 = lisseq(ch);
+    if (did_break()) err(breaker,"vector");
+    y[i] = itos(p1);
+  }
+  pop_val(ep); return y;
+}
+
+
 GEN
 vvecteur(GEN nmax, entree *ep, char *ch)
 {
@@ -672,7 +698,7 @@ matrice(GEN nlig, GEN ncol,entree *ep1, entree *ep2, char *ch)
   long c2[]={evaltyp(t_INT)|m_evallg(3), evalsigne(1)|evallgefint(3), 1};
 
   s=signe(ncol);
-  if (ep1 == ep2) err(talker, "identical index variables in matrix");
+  if (ep1 == ep2 && ep1) err(talker, "identical index variables in matrix");
   if (typ(ncol)!=t_INT || s<0) err(talker,"bad number of columns in matrix");
   if (!s) return cgetg(1,t_MAT);
 
