@@ -1,4 +1,4 @@
-/* $Id: kummer.c,v 1.17 2001/10/01 12:11:33 karim Exp $
+/* $Id: kummer.c,v 1.26 2002/01/09 14:05:07 karim Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -26,43 +26,29 @@ extern GEN vecinv(GEN x);
 extern GEN T2_from_embed_norm(GEN x, long r1);
 extern GEN pol_to_vec(GEN x, long N);
 extern GEN famat_to_nf(GEN nf, GEN f);
+extern GEN vconcat(GEN A, GEN B);
+extern long int_elt_val(GEN nf, GEN x, GEN p, GEN b, GEN *newx, long v);
+    
 
 static long rc,ell,degK,degKz,m,d,vnf,dv;
-static GEN matexpoteta1,nf,raycyc,polnf;
-static GEN bnfz,nfz,U,uu,gell,cyc,gencyc,vecalpha,R,g;
-static GEN listmod,listprSp,listbid,listunif,listellrank;
-static GEN listbidsup,listellranksup,vecw;
+static GEN invexpoteta1,nfz,gell;
 
 /* row vector B x matrix T : c_j=prod_i (b_i^T_ij) */
 static GEN
 groupproduct(GEN B, GEN T)
 {
-  long lB,lT,i,j;
-  GEN c,p1;
-
-  lB=lg(B)-1;
-  lT=lg(T)-1;
-  c=cgetg(lT+1,t_VEC);
-  for (j=1; j<=lT; j++)
+  long i,j, lB = lg(B), lT = lg(T);
+  GEN c = cgetg(lT,t_VEC);
+ 
+  for (j=1; j<lT; j++)
   {
-    p1=gun;
-    for (i=1; i<=lB; i++) p1=gmul(p1,gpui((GEN)B[i],gcoeff(T,i,j),0));
-    c[j]=(long)p1;
+    GEN p1 = gun;
+    for (i=1; i<lB; i++) p1 = gmul(p1, powgi((GEN)B[i],gcoeff(T,i,j)));
+    c[j] = (long)p1;
   }
   return c;
 }
 
-static GEN
-grouppows(GEN B, long ex)
-{
-  long lB = lg(B),j;
-  GEN c;
-
-  c = cgetg(lB,t_VEC);
-  for (j=1; j<lB; j++) c[j] = lpowgs((GEN)B[j], ex);
-  return c;
-}
-
 static int
 ok_x(GEN X, GEN arch, GEN vecmunit2, GEN msign)
 {
@@ -120,10 +106,10 @@ get_listx(GEN arch,GEN msign,GEN munit,GEN vecmunit2,long ell,long lSp,long nbvu
   }
 }
 
+/* given x in nf -- possibly not integral -- compute an algebraic integer
+ * of the form x * y^gell */
 static GEN
 reducealpha(GEN nf,GEN x,GEN gell)
-/* etant donne un nombre algebrique x du corps nf -- non necessairement
-   entier -- calcule un entier algebrique de la forme x*g^gell */
 {
   long tx=typ(x),i;
   GEN den,fa,fac,ep,p1,y;
@@ -141,7 +127,7 @@ reducealpha(GEN nf,GEN x,GEN gell)
 }
 
 long
-ellrank(GEN cyc, long ell)
+prank(GEN cyc, long ell)
 {
   long i;
   for (i=1; i<lg(cyc); i++)
@@ -171,7 +157,7 @@ rnfkummersimple(GEN bnr, GEN subgroup, long all)
   long nbgenclK,lSml2,lSl,lSp,rc,nbvunit;
   long lastbid,nbcol,nblig,k,llistx,e,vp,vd;
 
-  GEN bnf,nf,classgroup,cyclic,bid,ideal,arch,cycgen,gell,p1,p2,p3;
+  GEN polnf,bnf,nf,classgroup,cyclic,bid,ideal,arch,cycgen,gell,p1,p2,p3;
   GEN cyclicK,genK,listgamma,listalpha,fa,prm,expom,Sm,Sml1,Sml2,Sl,ESml2;
   GEN p,factell,Sp,vecbeta,matexpo,vunit,id,pr,vsml2,vecalpha0;
   GEN cycpro,munit,vecmunit2,msign,archartif,listx,listal,listg,listgamma0;
@@ -195,7 +181,7 @@ rnfkummersimple(GEN bnr, GEN subgroup, long all)
   gell = all? stoi(all): det(subgroup);
   ell = itos(gell);
 
-  cyclicK= gmael3(bnf,8,1,2); rc = ellrank(cyclicK,ell);
+  cyclicK= gmael3(bnf,8,1,2); rc = prank(cyclicK,ell);
   genK   = gmael3(bnf,8,1,3); nbgenclK = lg(genK)-1;
   listgamma0=cgetg(nbgenclK+1,t_VEC);
   listgamma=cgetg(nbgenclK+1,t_VEC);
@@ -300,7 +286,7 @@ rnfkummersimple(GEN bnr, GEN subgroup, long all)
     if (! gdivise((GEN)cycpro[l],gell)) err(talker,"bug5 in kummer");
   }
   bid = zidealstarinitgen(nf,id);
-  lastbid = ellrank(gmael(bid,2,2), ell);
+  lastbid = prank(gmael(bid,2,2), ell);
   nbcol=nbvunit+lSp; nblig=lastbid+rc;
   munit=cgetg(nbcol+1,t_MAT); vecmunit2=cgetg(lSml2+1,t_VEC);
   msign=cgetg(nbcol+1,t_MAT);
@@ -361,131 +347,111 @@ rnfkummersimple(GEN bnr, GEN subgroup, long all)
 }
 
 static GEN
-tauofideal(GEN id)
+tauofelt(GEN nf, GEN x, GEN U)
 {
-  long j;
-  GEN p1,p2;
-
-  p1=gsubst(gmul((GEN)nfz[7],id),vnf,U);
-  p2=cgetg(lg(p1),t_MAT);
-  for (j=1; j<lg(p1); j++) p2[j]=(long)algtobasis(nfz,(GEN)p1[j]);
-  return p2;
+  return algtobasis(nfz, gsubst(gmul((GEN)nf[7],x),vnf,U));
+}
+static GEN
+tauofalg(GEN x, GEN U)
+{
+  return gsubst(lift(x),vnf,U);
 }
 
 static GEN
-tauofprimeideal(GEN pr)
+lifttoKz(GEN nfz, GEN nf, GEN id, GEN A1)
 {
-  GEN p1 = dummycopy(pr);
-
-  p1[2] = (long)algtobasis(nfz, gsubst(gmul((GEN)nfz[7],(GEN)pr[2]),vnf,U));
-  return gcoeff(idealfactor(nfz,prime_to_ideal(nfz,p1)),1,1);
+  GEN I = ideal_two_elt(nf,id);
+  I[2] = (long)tauofelt(nf,(GEN)I[2],A1);
+  return prime_to_ideal(nfz,I);
 }
 
-static long
-isprimeidealconj(GEN pr1, GEN pr2)
-{
-  GEN pr=pr1;
 
-  do
+static GEN
+tauofideal(GEN id, GEN U)
+{
+  GEN I = ideal_two_elt(nfz,id);
+  I[2] = (long)tauofelt(nfz,(GEN)I[2],U);
+  return prime_to_ideal(nfz,I);
+}
+
+static int
+isprimeidealconj(GEN pr1, GEN pr2, GEN U)
+{
+  GEN p = (GEN)pr1[1];
+  GEN x = (GEN)pr1[2];
+  GEN b1= (GEN)pr1[5];
+  GEN b2= (GEN)pr2[5];
+  if (!egalii(p, (GEN)pr2[1])
+   || !egalii((GEN)pr1[3], (GEN)pr2[3])
+   || !egalii((GEN)pr1[4], (GEN)pr2[4])) return 0;
+  if (gegal(x,(GEN)pr2[2])) return 1;
+  for(;;)
   {
-    if (gegal(pr,pr2)) return 1;
-    pr = tauofprimeideal(pr);
+    if (int_elt_val(nfz,x,p,b2,NULL,0)) return 1;
+    x = FpV_red(tauofelt(nfz,x,U), p);
+    if (int_elt_val(nfz,x,p,b1,NULL,0)) return 0;
   }
-  while (!gegal(pr,pr1));
+}
+
+static int
+isconjinprimelist(GEN S, GEN pr2, GEN U)
+{
+  long lS=lg(S),i;
+
+  for (i=1; i<lS; i++)
+    if (isprimeidealconj((GEN)S[i],pr2,U)) return 1;
   return 0;
 }
 
-static long
-isconjinprimelist(GEN listpr, GEN pr2)
+static GEN
+downtoK(GEN polnf, GEN x)
 {
-  long ll=lg(listpr)-1,i;
-
-  for (i=1; i<=ll; i++)
-    if (isprimeidealconj((GEN)listpr[i],pr2)) return 1;
-  return 0;
-}
-
-static void
-computematexpoteta1(GEN A1, GEN R)
-{
-  long j;
-  GEN Aj = polun[vnf];
-  matexpoteta1 = cgetg(degK+1,t_MAT);
-  for (j=1; j<=degK; j++)
-  {
-    matexpoteta1[j] = (long)pol_to_vec(Aj, degKz);
-    if (j<degK) Aj = gmod(gmul(Aj,A1), R);
-  }
+  GEN y = gmul(invexpoteta1, pol_to_vec(lift(x), degKz));
+  return gmodulcp(gtopolyrev(y,vnf),polnf);
 }
 
 static GEN
-downtoK(GEN x)
+tracetoK(GEN polnf, GEN x, GEN U)
 {
+  GEN p1 = x;
   long i;
-  GEN p2,p3;
-
-  p2 = inverseimage(matexpoteta1, pol_to_vec(lift(x), degKz));
-  if (lg(p2)==1) err(talker,"not an element of K in downtoK");
-  p3 = (GEN)p2[degK];
-  for (i=degK-1; i; i--) p3 = gadd((GEN)p2[i],gmul(polx[vnf],p3));
-  return gmodulcp(p3,polnf);
+  for (i=1; i<m; i++) p1 = gadd(x, tauofalg(p1,U));
+  return downtoK(polnf,p1);
 }
 
 static GEN
-tracetoK(GEN x)
+normtoK(GEN polnf, GEN x, GEN U)
 {
+  GEN p1 = x;
   long i;
-  GEN p1,p2;
-
-  p1=x; p2=x;
-  for (i=1; i<=m-1; i++)
-  {
-    p1=gsubst(lift(p1),vnf,U);
-    p2=gadd(p2,p1);
-  }
-  return downtoK(p2);
+  for (i=1; i<m; i++) p1 = gmul(x, tauofalg(p1,U));
+  return downtoK(polnf,p1);
 }
 
 static GEN
-normtoK(GEN x)
-{
-  long i;
-  GEN p1,p2;
-
-  p1=x; p2=x;
-  for (i=1; i<=m-1; i++)
-  {
-    p1=gsubst(lift(p1),vnf,U);
-    p2=gmul(p2,p1);
-  }
-  return downtoK(p2);
-}
-
-static GEN
-computepolrel(void)
+computepolrel(GEN polnf, GEN be, GEN U)
 {
+  GEN p1 = gun, p2 = be;
   long i,j;
-  GEN p1,p2;
 
-  p1=gun; p2=gmodulcp(polx[vnf],R);
-  for (i=0; i<=m-1; i++)
+  for (i=0; i<m; i++)
   {
-    p1=gmul(p1,gsub(polx[0],p2));
-    if (i<m-1) p2=gsubst(lift(p2),vnf,U);
+    p1 = gmul(p1,gsub(polx[0],p2));
+    if (i<m-1) p2 = tauofalg(p2,U);
   }
-  for (j=2; j<=m+2; j++) p1[j]=(long)downtoK((GEN)p1[j]);
+  for (j=2; j<=m+2; j++) p1[j]=(long)downtoK(polnf,(GEN)p1[j]);
   return p1;
 }
 
 /* alg. 5.2.15. with remark */
 static GEN
-isprincipalell(GEN id)
+isprincipalell(GEN bnfz, GEN id, GEN vecalpha, GEN uu)
 {
   long i, l = lg(vecalpha);
   GEN y,logdisc,be;
 
   y = isprincipalgenforce(bnfz,id);
-  logdisc = (GEN)y[1];
+  logdisc = gmod((GEN)y[1], gell);
   be = basistoalg(bnfz,(GEN)y[2]);
   for (i=rc+1; i<l; i++)
     be = gmul(be, powgi((GEN)vecalpha[i], modii(mulii((GEN)logdisc[i],(GEN)uu[i]),gell)));
@@ -497,14 +463,14 @@ isprincipalell(GEN id)
 
 /* alg. 5.3.11. */
 static GEN
-isvirtualunit(GEN v)
+isvirtualunit(GEN bnfz, GEN v, GEN vecalpha, GEN cyc)
 {
   long llist,i,j,ex, l = lg(vecalpha);
   GEN p1,listex,listpr,q,ga,eps,vecy,logdisc;
 
   p1=idealfactor(nfz,v);
   listpr = (GEN)p1[1]; llist = lg(listpr)-1;
-  listex = (GEN)p1[2]; q = idmat(degKz);
+  listex = (GEN)p1[2]; q = gun;
   for (i=1; i<=llist; i++)
   {
     ex = itos((GEN)listex[i]);
@@ -532,22 +498,7 @@ isvirtualunit(GEN v)
 }
 
 static GEN
-lifttokz(GEN id, GEN A1)
-{
-  long i,j;
-  GEN p1,p2,p3;
-
-  p1=gsubst(gmul((GEN)nf[7],id),vnf,A1);
-  p2=gmodulcp((GEN)nfz[7],R);
-  p3=cgetg(degK*degKz+1,t_MAT);
-  for (i=1; i<=degK; i++)
-    for (j=1; j<=degKz; j++)
-      p3[(i-1)*degKz+j]=(long)algtobasis(nfz,gmul((GEN)p1[i],(GEN)p2[j]));
-  return hnfmod(p3,detint(p3));
-}
-
-static GEN
-steinitzaux(GEN id, GEN polrel)
+steinitzaux(GEN nf, GEN id, GEN polrel)
 {
   long i,j;
   GEN p1,p2,p3,vecid,matid,pseudomat,pid;
@@ -572,27 +523,28 @@ steinitzaux(GEN id, GEN polrel)
 }
 
 static GEN
-normrelz(GEN id, GEN polrel, GEN steinitzZk)
+normrelz(GEN nf, GEN id, GEN polrel, GEN steinitzZk)
 {
-  GEN p1 = steinitzaux(idealhermite(nfz, id), polrel);
+  GEN p1 = steinitzaux(nf,idealhermite(nfz, id), polrel);
   return idealdiv(nf,p1,steinitzZk);
 }
 
 static GEN
-invimsubgroup(GEN bnrz, GEN bnr, GEN subgroup)
+invimsubgroup(GEN bnrz, GEN bnr, GEN subgroup, GEN tau)
 {
   long l,j;
   GEN g,Plog,raycycz,rayclgpz,genraycycz,U,polrel,steinitzZk;
+  GEN nf = checknf(bnr);
 
-  polrel = computepolrel();
-  steinitzZk = steinitzaux(idmat(degKz), polrel); 
+  polrel = computepolrel((GEN)nf[1], polx[vnf], tau);
+  steinitzZk = steinitzaux(nf,idmat(degKz), polrel); 
   rayclgpz = (GEN)bnrz[5];
   raycycz   = (GEN)rayclgpz[2]; l=lg(raycycz);
   genraycycz= (GEN)rayclgpz[3];
   Plog = cgetg(l,t_MAT);
   for (j=1; j<l; j++)
   {
-    g = normrelz((GEN)genraycycz[j],polrel,steinitzZk);
+    g = normrelz(nf,(GEN)genraycycz[j],polrel,steinitzZk);
     Plog[j] = (long)isprincipalray(bnr, g);
   }
   U = (GEN)hnfall(concatsp(Plog, subgroup))[2];
@@ -600,32 +552,6 @@ invimsubgroup(GEN bnrz, GEN bnr, GEN subgroup)
   return hnfmod(concatsp(U, diagonal(raycycz)), (GEN)raycycz[1]);
 }
 
-static GEN
-ideallogaux(long i, GEN al)
-{
-  long llogli,valal;
-  GEN p1;
-
-  valal = element_val(nfz,al,(GEN)listprSp[i]);
-  al = gmul(al,gpuigs((GEN)listunif[i],valal));
-  p1 = zideallog(nfz,al,(GEN)listbid[i]);
-  llogli = listellrank[i];
-  setlg(p1,llogli+1); return p1;
-}
-
-static GEN
-ideallogauxsup(long i, GEN al)
-{
-  long llogli,valal;
-  GEN p1;
-
-  valal = element_val(nfz,al,(GEN)listprSp[i]);
-  al = gmul(al,gpuigs((GEN)listunif[i],valal));
-  p1 = zideallog(nfz,al,(GEN)listbidsup[i]);
-  llogli = listellranksup[i];
-  setlg(p1,llogli+1); return p1;
-}
-
 static GEN
 vectau(GEN list)
 {
@@ -713,7 +639,7 @@ mulpoltau(GEN poltau, GEN list)
 
 /* th. 5.3.5. and prop. 5.3.9. */
 static GEN
-computepolrelbeta(GEN be)
+computepolrelbeta(GEN polnf, GEN be, long g, GEN U)
 {
   long i,a,b,j;
   GEN e,u,u1,u2,u3,p1,p2,p3,p4,zet,be1,be2,listr,s,veczi,vecci,powtaubet;
@@ -721,22 +647,22 @@ computepolrelbeta(GEN be)
   switch (ell)
   {
     case 2: err(talker,"you should not be here in rnfkummer !!"); break;
-    case 3: e=normtoK(be); u=tracetoK(be);
+    case 3: e=normtoK(polnf,be,U); u=tracetoK(polnf,be,U);
       return gsub(gmul(polx[0],gsub(gsqr(polx[0]),gmulsg(3,e))),gmul(e,u));
-    case 5: e=normtoK(be);
+    case 5: e=normtoK(polnf,be,U);
       if (d==2)
       {
-	u=tracetoK(gpuigs(be,3));
+	u=tracetoK(polnf,gpuigs(be,3),U);
 	p1=gadd(gmulsg(5,gsqr(e)), gmul(gsqr(polx[0]), gsub(gsqr(polx[0]),gmulsg(5,e))));
 	return gsub(gmul(polx[0],p1),gmul(e,u));
       }
       else
       {
-	be1=gsubst(lift(be),vnf,U);
-        be2=gsubst(lift(be1),vnf,U);
-	u1=tracetoK(gmul(be,be1));
-        u2=tracetoK(gmul(gmul(be,be2),gsqr(be1)));
-	u3=tracetoK(gmul(gmul(gsqr(be),gpuigs(be1,3)),be2));
+	be1=tauofalg(be,U);
+        be2=tauofalg(be1,U);
+	u1=tracetoK(polnf,gmul(be,be1),U);
+        u2=tracetoK(polnf,gmul(gmul(be,be2),gsqr(be1)),U);
+	u3=tracetoK(polnf,gmul(gmul(gsqr(be),gpuigs(be1,3)),be2),U);
 	p1=gsub(gsqr(polx[0]),gmulsg(10,e));
 	p1=gsub(gmul(polx[0],p1),gmulsg(5,gmul(e,u1)));
 	p1=gadd(gmul(polx[0],p1),gmul(gmulsg(5,e),gsub(e,u2)));
@@ -749,7 +675,7 @@ computepolrelbeta(GEN be)
       zet=gmodulcp(polx[vnf], cyclo(ell,vnf));
       listr=cgetg(m+1,t_VEC);
       listr[1]=un;
-      for (i=2; i<=m; i++) listr[i]=(long)modii(mulii((GEN)listr[i-1],g),gell);
+      for (i=2; i<=m; i++) listr[i]=(long)modii(mulis((GEN)listr[i-1],g),gell);
       veczi=cgetg(m+1,t_VEC);
       for (b=0; b<m; b++)
       {
@@ -768,14 +694,14 @@ computepolrelbeta(GEN be)
       }
       powtaubet=cgetg(m+1,t_VEC);
       powtaubet[1]=(long)be;
-      for (i=2; i<=m; i++) powtaubet[i]=(long)gsubst(lift((GEN)powtaubet[i-1]),vnf,U);
+      for (i=2; i<=m; i++) powtaubet[i]=(long)tauofalg((GEN)powtaubet[i-1],U);
       err(impl,"difficult Kummer for ell>=7"); return gzero;
   }
   return NULL; /* not reached */
 }
 
 static GEN
-fix_be(GEN be, GEN u)
+fix_be(GEN bnfz, GEN be, GEN u)
 {
   GEN e,g, nf = checknf(bnfz), fu = gmael(bnfz,8,5);
   long i,lu;
@@ -809,7 +735,7 @@ logarch2arch(GEN x, long r1, long prec)
 
 /* multiply be by ell-th powers of units as to find small L2-norm for new be */
 static GEN
-reducebetanaive(GEN be, GEN b)
+reducebetanaive(GEN bnfz, GEN be, GEN b)
 {
   long i,k,n,ru,r1, prec = nfgetprec(bnfz);
   GEN z,p1,p2,nmax,c, nf = checknf(bnfz);
@@ -843,16 +769,16 @@ reducebetanaive(GEN be, GEN b)
     b = B; c[besti] = laddis((GEN)c[besti], bestk);
   }
   if (DEBUGLEVEL) fprintferr("unit exponents = %Z\n",c);
-  return fix_be(be,c);
+  return fix_be(bnfz,be,c);
 }
 
 static GEN
-reducebeta(GEN be)
+reducebeta(GEN bnfz, GEN be)
 {
   long j,ru, prec = nfgetprec(bnfz);
   GEN emb,z,u,matunit, nf = checknf(bnfz);
 
-  if (gcmp1(gnorm(be))) return reducebetanaive(be,NULL);
+  if (gcmp1(gnorm(be))) return reducebetanaive(bnfz,be,NULL);
   matunit = gmulgs(greal((GEN)bnfz[3]), ell); /* log. embeddings of fu^ell */
   for (;;)
   {
@@ -864,22 +790,23 @@ reducebeta(GEN be)
   }
   z = concatsp(matunit, z);
   u = lllintern(z, 1, prec);
-  if (!u) return reducebetanaive(be,emb); /* shouldn't occur */
+  if (!u) return reducebetanaive(bnfz,be,emb); /* shouldn't occur */
   ru = lg(u);
   for (j=1; j<ru; j++)
     if (smodis(gcoeff(u,ru-1,j), ell)) break; /* prime to ell */
   u = (GEN)u[j]; /* coords on (fu^ell, be) of a small generator */
   ru--; setlg(u, ru);
   be = powgi(be, (GEN)u[ru]);
-  return reducebetanaive(fix_be(be, u), NULL);
+  return reducebetanaive(bnfz,fix_be(bnfz,be,u), NULL);
 }
 
 /* cf. algo 5.3.18 */
 static GEN
-testx(GEN bnf, GEN X, GEN module, GEN subgroup, GEN vecMsup)
+testx(GEN bnfz, GEN bnf, GEN X, GEN module, GEN subgroup, GEN vecMsup,
+      GEN vecWB, long g, GEN U)
 {
-  long i,v,l,lX;
-  GEN be,polrelbe,p1;
+  long i,l,lX;
+  GEN be,polrelbe,p1,nf;
 
   if (gcmp0(X)) return NULL;
   lX = lg(X);
@@ -890,32 +817,56 @@ testx(GEN bnf, GEN X, GEN module, GEN subgroup, GEN vecMsup)
     if (gcmp0(FpV_red(gmul((GEN)vecMsup[i],X), gell))) return NULL;
   be = gun;
   for (i=1; i<lX; i++)
-    be = gmul(be, powgi((GEN)vecw[i], (GEN)X[i]));
+    be = gmul(be, powgi((GEN)vecWB[i], (GEN)X[i]));
   if (DEBUGLEVEL>1) fprintferr("reducing beta = %Z\n",be);
-  be = reducebeta(be);
+  be = reducebeta(bnfz, be);
   if (DEBUGLEVEL>1) fprintferr("beta reduced = %Z\n",be);
-  polrelbe = computepolrelbeta(be);
-  v = varn(polrelbe);
-  p1 = unifpol((GEN)bnf[7],polrelbe,0);
+  nf = (GEN)bnf[7];
+  polrelbe = computepolrelbeta((GEN)nf[1],be,g,U);
+ 
+  p1 = unifpol(nf,polrelbe,0);
+  l = lg(p1); /* lift to Q rational coeffs */
+  for (i=2; i<l; i++)
+    if (isnfscalar((GEN)p1[i])) polrelbe[i] = mael(p1,i,1);
+   
   p1 = denom(gtovec(p1));
-  polrelbe = gsubst(polrelbe,v, gdiv(polx[v],p1));
-  polrelbe = gmul(polrelbe, gpowgs(p1, degpol(polrelbe)));
+  polrelbe = rescale_pol(polrelbe,p1);
   if (DEBUGLEVEL>1) fprintferr("polrelbe = %Z\n",polrelbe);
   p1 = rnfconductor(bnf,polrelbe,0);
   if (!gegal((GEN)p1[1],module) || !gegal((GEN)p1[3],subgroup)) return NULL;
   return polrelbe;
 }
 
+static GEN
+logall(long dv, GEN mod, GEN vecWA, long mginv, GEN pr, GEN unif)
+{
+  GEN p1,al,M,bid = zidealstarinitgen(nfz,mod);
+  long ellrank = prank(gmael(bid,2,2), ell);
+  long val,i, l = lg(vecWA);
+
+  M = cgetg(l,t_MAT);
+  for (i=1; i<l; i++)
+  {
+    al = (GEN)vecWA[i]; val = element_val(nfz,al,pr);
+    if (val) al = gmul(al, gpuigs(unif,val));
+    p1 = zideallog(nfz, al, bid); setlg(p1,ellrank+1);
+    if (i <= dv) p1 = gmulsg(mginv, p1);
+    M[i] = (long)p1;
+  }
+  return M;
+}
+
 GEN
 rnfkummer(GEN bnr, GEN subgroup, long all, long prec)
 {
-  long i,j,l,av=avma,e,vp,vd,dK,lSl,lSp,lSl2,lSml2,dc,ru,rv,nbcol;
-  GEN p1,p2,p3,p4,wk,pr;
-  GEN bnf,rayclgp,bid,ideal,cycgen,vselmer;
+  long i,j,l,av=avma,e,vp,vd,dK,lSl,lSp,lSl2,lSml2,dc,ru,rv,nbcol,g,mginv;
+  GEN p1,p2,p3,p4,wk,pr,U,R;
+  GEN polnf,nf,bnf,bnfz,rayclgp,bid,ideal,cycgen,vselmer;
   GEN kk,clgp,fununits,torsunit,vecB,vecC,Tc,Tv,P;
-  GEN Q,Qt,idealz,gothf,factgothf,listpr,listex,factell,p,vecnul;
-  GEN M,al,K,Msup,X,finalresult,y,module,A1,A2,vecMsup;
-  GEN listmodsup,vecalphap,vecbetap,mginv,matP,ESml2,Sp,Sm,Sml1,Sml2,Sl;
+  GEN Q,idealz,gothf,factgothf,factell,p;
+  GEN listprSp,listpr,listex,vecw,vecWA,vecWB;
+  GEN M,K,X,finalresult,y,module,A1,A2,vecMsup;
+  GEN uu,gencyc,cyc,vecalpha,vecalphap,vecbetap,matP,ESml2,Sp,Sm,Sml1,Sml2,Sl;
 
   checkbnrgen(bnr);
   wk = gmael4(bnr,1,8,4,1);
@@ -924,6 +875,7 @@ rnfkummer(GEN bnr, GEN subgroup, long all, long prec)
   {
     if (!gcmp0(subgroup)) gell = det(subgroup);
     else gell = det(diagonal(gmael(bnr,5,2)));
+    if (typ(gell) != t_INT) err(arither1);
   }
   if (gcmp1(gell)) { avma = av; return polx[varn(gmael3(bnr,1,7,1))]; }
   if (!isprime(gell)) err(impl,"kummer for composite relative degree");
@@ -932,7 +884,7 @@ rnfkummer(GEN bnr, GEN subgroup, long all, long prec)
   if (all && gcmp0(subgroup))
     err(talker,"kummer when zeta not in K requires a specific subgroup");
   bnf = (GEN)bnr[1];
-  nf = (GEN)bnf[7];
+  nf  = (GEN)bnf[7];
   polnf = (GEN)nf[1]; vnf = varn(polnf); degK = degpol(polnf);
   if (!vnf) err(talker,"main variable in kummer must not be x");
       /* step 7 */
@@ -941,10 +893,10 @@ rnfkummer(GEN bnr, GEN subgroup, long all, long prec)
   bnr = (GEN)p1[2]; 
   subgroup = (GEN)p1[3];
   rayclgp = (GEN)bnr[5];
-  raycyc = (GEN)rayclgp[2];
+  
   bid = (GEN)bnr[2]; module=(GEN)bid[1];
   ideal = (GEN)module[1];
-  if (gcmp0(subgroup)) subgroup = diagonal(raycyc);
+  if (gcmp0(subgroup)) subgroup = diagonal((GEN)rayclgp[2]);
   ell = itos(gell);
       /* step 1 of alg 5.3.5. */
   if (DEBUGLEVEL>2) fprintferr("Step 1\n");
@@ -954,43 +906,36 @@ rnfkummer(GEN bnr, GEN subgroup, long all, long prec)
   A1= (GEN)p1[2];
   A2= (GEN)p1[3];
   kk= (GEN)p1[4];
-  if (signe(leadingcoeff(R)) < 0)
-  {
-    R = gneg_i(R);
-    A1 = gmodulcp(lift(A1),R);
-    A2 = gmodulcp(lift(A2),R);
-  }
       /* step 2 */
   if (DEBUGLEVEL>2) fprintferr("Step 2\n");
   degKz = degpol(R);
   m = degKz/degK;
   d = (ell-1)/m;
-  g = lift(gpuigs(gener(gell),d)); /* g has order m in all (Z/ell^k)^* */
-  if (gcmp1(powmodulo(g, stoi(m), stoi(ell*ell)))) g = addsi(ell,g);
+  g = powuumod(itos(lift(gener(gell))), d, ell);
+  if (powuumod(g, m, ell*ell) == 1) g += ell; /* ord(g)=m in all (Z/ell^k)^* */
       /* step reduction of R */
-  if (degKz<=20)
+  if (degKz <= 20)
   {
     GEN A3,A3rev;
 
     if (DEBUGLEVEL>2) fprintferr("Step reduction\n");
     p1 = polredabs2(R,prec);
-    if (DEBUGLEVEL>2) fprintferr("polredabs = %Z",p1[1]);
-    R = (GEN)p1[1];
+    R = (GEN)p1[1]; if (DEBUGLEVEL>2) fprintferr("polredabs = %Z",R);
     A3= (GEN)p1[2];
     A1 = poleval(lift(A1), A3);
     A2 = poleval(lift(A2), A3);
     A3rev= polymodrecip(A3);
-    U = gadd(powgi(A2,g), gmul(kk,A1));
+    U = gadd(gpowgs(A2,g), gmul(kk,A1));
     U = poleval(lift(A3rev), U);
   }
-  else U = gadd(powgi(A2,g), gmul(kk,A1));
+  else U = gadd(gpowgs(A2,g), gmul(kk,A1));
       /* step 3 */
       /* one could factor disc(R) using th. 2.1.6. */
   if (DEBUGLEVEL>2) fprintferr("Step 3\n");
   bnfz = bnfinit0(R,1,NULL,prec);
   nfz = (GEN)bnfz[7];
   clgp = gmael(bnfz,8,1);
-  cyc   = (GEN)clgp[2]; rc = ellrank(cyc,ell);
+  cyc   = (GEN)clgp[2]; rc = prank(cyc,ell);
   gencyc= (GEN)clgp[3]; l = lg(cyc);
   vecalpha=cgetg(l,t_VEC);
   cycgen = check_and_build_cycgen(bnfz);
@@ -1015,7 +960,7 @@ rnfkummer(GEN bnr, GEN subgroup, long all, long prec)
   Tc=cgetg(rc+1,t_MAT);
   for (j=1; j<=rc; j++)
   {
-    p1 = isprincipalell(tauofideal((GEN)gencyc[j]));
+    p1 = isprincipalell(bnfz, tauofideal((GEN)gencyc[j],U), vecalpha,uu);
     Tc[j]  = p1[1];
     vecB[j]= p1[2];
   }
@@ -1027,8 +972,8 @@ rnfkummer(GEN bnr, GEN subgroup, long all, long prec)
   p3=vecB;
   for (j=1; j<=m-1; j++)
   {
-    p3 = gsubst(lift(p3),vnf,U);
-    p4 = groupproduct(grouppows(p3,(j*d)%ell),(GEN)p1[m-j]);
+    p3 = tauofalg(p3,U);
+    p4 = groupproduct(p3, gmod(gmulsg((j*d)%ell,(GEN)p1[m-j]), gell));
     for (i=1; i<=rc; i++) p2[i] = lmul((GEN)p2[i],(GEN)p4[i]);
   }
   vecC=p2;
@@ -1037,10 +982,10 @@ rnfkummer(GEN bnr, GEN subgroup, long all, long prec)
   Tv = cgetg(rv+1,t_MAT);
   for (j=1; j<=rv; j++)
   {
-    Tv[j] = isvirtualunit(gsubst(lift((GEN)vselmer[j]),vnf,U))[1];
+    Tv[j] = isvirtualunit(bnfz, tauofalg((GEN)vselmer[j],U), vecalpha,cyc)[1];
     if (DEBUGLEVEL>2) fprintferr("   %ld\n",j);
   }
-  P = FpM_ker(gsub(Tv, g), gell);
+  P = FpM_ker(gsubgs(Tv, g), gell);
   dv= lg(P)-1; vecw = cgetg(dv+1,t_VEC);
   for (j=1; j<=dv; j++)
   {
@@ -1050,19 +995,27 @@ rnfkummer(GEN bnr, GEN subgroup, long all, long prec)
   }
       /* step 6 */
   if (DEBUGLEVEL>2) fprintferr("Step 6\n");
-  Q = FpM_ker(gsub(gtrans(Tc), g), gell);
-  Qt = gtrans(Q); dc = lg(Q)-1;
+  Q = FpM_ker(gsubgs(gtrans(Tc), g), gell);
+  dc = lg(Q)-1;
       /* step 7 done above */
       /* step 8 */
   if (DEBUGLEVEL>2) fprintferr("Step 7 and 8\n");
-  idealz=lifttokz(ideal, A1);
-  computematexpoteta1(lift(A1), R);
+  idealz = lifttoKz(nfz, nf, ideal, A1);
+  A1 = lift_intern(A1);
+  p1 = polun[vnf];
+  p2 = cgetg(degK+1,t_MAT);
+  for (j=1; j<=degK; j++)
+  {
+    p2[j] = (long)pol_to_vec(p1, degKz);
+    if (j<degK) p1 = gmod(gmul(p1,A1), R);
+  }
+  invexpoteta1 = invmat(p2); /* left inverse */
+ 
   if (!divise(idealnorm(nf,ideal),gell)) gothf = idealz;
   else
   {
-    GEN bnrz, subgroupz;
-    bnrz = buchrayinitgen(bnfz,idealz);
-    subgroupz = invimsubgroup(bnrz,bnr,subgroup);
+    GEN bnrz = buchrayinitgen(bnfz,idealz);
+    GEN subgroupz = invimsubgroup(bnrz,bnr,subgroup,U);
     gothf = conductor(bnrz,subgroupz,0);
   }
       /* step 9 */
@@ -1084,7 +1037,7 @@ rnfkummer(GEN bnr, GEN subgroup, long all, long prec)
     if (!egalii(p,gell))
     {
       if (vp != 1) { avma = av; return gzero; }
-      if (!isconjinprimelist(Sm,pr)) _append(Sm,pr);
+      if (!isconjinprimelist(Sm,pr,U)) _append(Sm,pr);
     }
     else
     {
@@ -1092,12 +1045,12 @@ rnfkummer(GEN bnr, GEN subgroup, long all, long prec)
       if (vd > 0) { avma = av; return gzero; }
       if (vd==0)
       {
-	if (!isconjinprimelist(Sml1,pr)) _append(Sml1, pr);
+	if (!isconjinprimelist(Sml1,pr,U)) _append(Sml1, pr);
       }
       else
       {
 	if (vp==1) { avma = av; return gzero; }
-        if (!isconjinprimelist(Sml2,pr))
+        if (!isconjinprimelist(Sml2,pr,U))
         {
           _append(Sml2, pr);
           _append(ESml2,(GEN)vp);
@@ -1110,7 +1063,7 @@ rnfkummer(GEN bnr, GEN subgroup, long all, long prec)
   {
     pr = (GEN)factell[i];
     if (!idealval(nfz,gothf,pr))
-      if (!isconjinprimelist(Sl,pr)) _append(Sl, pr);
+      if (!isconjinprimelist(Sl,pr,U)) _append(Sl, pr);
   }
   lSml2=lg(Sml2)-1; lSl=lg(Sl)-1; lSl2=lSml2+lSl;
   Sp=concatsp(Sm,Sml1); lSp=lg(Sp)-1;
@@ -1121,89 +1074,62 @@ rnfkummer(GEN bnr, GEN subgroup, long all, long prec)
   matP=cgetg(lSp+1,t_MAT);
   for (j=1; j<=lSp; j++)
   {
-    p1=isprincipalell((GEN)Sp[j]);
-    matP[j]=p1[1];
-    p2=gun;
+    p1 = isprincipalell(bnfz, (GEN)Sp[j], vecalpha,uu);
+    matP[j] = p1[1];
+    p2 = gun;
     for (i=1; i<=rc; i++)
       p2 = gmul(p2, powgi((GEN)vecC[i],gmael(p1,1,i)));
     p3=gdiv((GEN)p1[2],p2); vecbetap[j]=(long)p3;
     p2=gun;
     for (i=0; i<m; i++)
     {
-      p2 = gmul(p2, powgi(p3, powmodulo(g,stoi(m-1-i),gell)));
-      if (i<m-1) p3=gsubst(lift(p3),vnf,U);
+      p2 = gmul(p2, gpowgs(p3, powuumod(g,m-1-i,ell)));
+      if (i<m-1) p3 = tauofalg(p3,U);
     }
     vecalphap[j]=(long)p2;
   }
       /* step 13 */
   if (DEBUGLEVEL>2) fprintferr("Step 13\n");
-  nbcol=lSp+dv;
-  vecw=concatsp(vecw,vecbetap);
-  listmod=cgetg(lSl2+1,t_VEC);
-  listunif=cgetg(lSl2+1,t_VEC);
+  nbcol= lSp+dv;
+  vecWB = concatsp(vecw, vecbetap);
+  vecWA = concatsp(vecw, vecalphap);
   listprSp = concatsp(Sml2, Sl);
-  for (j=1; j<=lSl2; j++)
-  {
-    GEN pr = (GEN)listprSp[j];
-    long e = itos((GEN)pr[3]), z = ell * (e / (ell-1));
-    if (j <= lSml2) z += 1 - ESml2[j];
-    listmod[j]=(long)idealpows(nfz,pr, z);
-    listunif[j]=(long)basistoalg(nfz,gdiv((GEN)pr[5],(GEN)pr[1]));
-  }
-      /* step 14 and step 15 */
-  if (DEBUGLEVEL>2) fprintferr("Step 14 and 15\n");
-  listbid=cgetg(lSl2+1,t_VEC);
-  listellrank = cgetg(lSl2+1,t_VECSMALL);
+      /* step 14, 15, and 17 */
+  if (DEBUGLEVEL>2) fprintferr("Step 14, 15 and 17\n");
+  mginv = (m * u_invmod(g,ell)) % ell;
+  vecMsup = cgetg(lSml2+1,t_VEC);
+  M = NULL;
   for (i=1; i<=lSl2; i++)
   {
-    listbid[i]=(long)zidealstarinitgen(nfz,(GEN)listmod[i]);
-    listellrank[i] = ellrank(gmael3(listbid,i,2,2), ell);
+    GEN pr = (GEN)listprSp[i];
+    long e = itos((GEN)pr[3]), z = ell * (e / (ell-1));
+    GEN mod, unif = basistoalg(nfz,gdiv((GEN)pr[5],(GEN)pr[1]));
+
+    if (i <= lSml2) z += 1 - ESml2[i];
+    mod  = idealpows(nfz,pr, z);
+
+    M = vconcat(M, logall(dv,mod,vecWA,mginv,pr,unif));
+    if (i <= lSml2)
+    {
+      GEN modsup = idealmul(nfz,pr,mod);
+      vecMsup[i] = (long)logall(dv,modsup,vecWA,mginv,pr,unif);
+    }
   }
-  mginv = modii(mulsi(m, mpinvmod(g,gell)), gell);
-  vecnul=cgetg(dc+1,t_COL); for (i=1; i<=dc; i++) vecnul[i]=zero;
-  M=cgetg(nbcol+1,t_MAT);
-  for (j=1; j<=dv; j++)
+  if (dc)
   {
-    p1=cgetg(1,t_COL);
-    al=(GEN)vecw[j];
-    for (i=1; i<=lSl2; i++) p1 = concatsp(p1,ideallogaux(i,al));
-    p1=gmul(mginv,p1);
-    M[j]=(long)concatsp(p1,vecnul);
-  }
-  for (   ; j<=nbcol; j++)
-  {
-    p1=cgetg(1,t_COL);
-    al=(GEN)vecalphap[j-dv];
-    for (i=1; i<=lSl2; i++) p1 = concatsp(p1,ideallogaux(i,al));
-    M[j]=(long)concatsp(p1,gmul(Qt,(GEN)matP[j-dv]));
+    GEN QtP = gmul(gtrans_i(Q),matP);
+    M = vconcat(M, concatsp(zeromat(dc,dv), QtP));
   }
+  if (!M) M = zeromat(1,nbcol);
       /* step 16 */
   if (DEBUGLEVEL>2) fprintferr("Step 16\n");
   K = FpM_ker(M, gell);
   dK= lg(K)-1; if (!dK) { avma=av; return gzero; }
-      /* step 17 */
-  if (DEBUGLEVEL>2) fprintferr("Step 17\n");
-  listmodsup=cgetg(lSml2+1,t_VEC);
-  listbidsup=cgetg(lSml2+1,t_VEC);
-  listellranksup=cgetg(lSml2+1,t_VECSMALL);
-  for (i=1; i<=lSml2; i++)
-  {
-    listmodsup[i]=(long)idealmul(nfz,(GEN)listprSp[i],(GEN)listmod[i]);
-    listbidsup[i]=(long)zidealstarinitgen(nfz,(GEN)listmodsup[i]);
-    listellranksup[i] = ellrank(gmael3(listbidsup,i,2,2), ell);
-  }
-  vecMsup=cgetg(lSml2+1,t_VEC);
-  for (i=1; i<=lSml2; i++)
-  {
-    Msup=cgetg(nbcol+1,t_MAT); vecMsup[i]=(long)Msup;
-    for (j=1; j<=dv; j++) Msup[j]=lmul(mginv,ideallogauxsup(i,(GEN)vecw[j]));
-    for (   ; j<=nbcol; j++) Msup[j]=(long)ideallogauxsup(i,(GEN)vecalphap[j-dv]);
-  }
       /* step 18 */
   if (DEBUGLEVEL>2) fprintferr("Step 18\n");
+  y = cgetg(dK,t_VECSMALL);
   do
   {
-    y = cgetg(dK,t_VECSMALL);
     for (i=1; i<dK; i++) y[i] = 0;
 	/* step 19 */
     for(;;)
@@ -1211,7 +1137,7 @@ rnfkummer(GEN bnr, GEN subgroup, long all, long prec)
       X = (GEN)K[dK];
       for (j=1; j<dK; j++) X = gadd(X, gmulsg(y[j],(GEN)K[j]));
       X = FpV_red(X, gell);
-      finalresult = testx(bnf,X,module,subgroup,vecMsup);
+      finalresult = testx(bnfz,bnf,X,module,subgroup,vecMsup,vecWB,g,U);
       if (finalresult) return gerepilecopy(av, finalresult);
         /* step 20,21,22 */
       i = dK;
