@@ -1,4 +1,4 @@
-/* $Id: subfield.c,v 1.32 2001/10/16 17:23:04 karim Exp $
+/* $Id: subfield.c,v 1.34 2001/11/23 22:17:42 karim Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -496,6 +496,36 @@ chinese_retrieve_pol(GEN DATA, GEN listdelta)
   return FpX_res(FpX_red(S, p), FpX_red((GEN)DATA[1],p), p);
 }
 
+/* return P(X + c) using destructive Horner */
+static GEN
+TR_pol(GEN P, GEN c)
+{
+  GEN Q = dummycopy(P), *R;
+  long i,k,n;
+  
+  if (!signe(P)) return Q;
+  R = (GEN*)(Q+2); n = degpol(P);
+  if (gcmp1(c))
+  {
+    for (i=1; i<=n; i++)
+      for (k=n-i; k<n; k++)
+        R[k] = gadd(R[k], R[k+1]);
+  }
+  else if (gcmp_1(c))
+  {
+    for (i=1; i<=n; i++)
+      for (k=n-i; k<n; k++)
+        R[k] = gsub(R[k], R[k+1]);
+  }
+  else
+  {
+    for (i=1; i<=n; i++)
+      for (k=n-i; k<n; k++)
+        R[k] = gadd(R[k], gmul(c, R[k+1]));
+  }
+  return Q;
+}
+
 /* g in Z[X] potentially defines a subfield of Q[X]/f. It is a subfield iff A
  * (cf cand_for_subfields) was a block system; then there
  * exists h in Q[X] such that f | g o h. listdelta determines h s.t f | g o h
@@ -559,7 +589,7 @@ embedding_of_potential_subfields(GEN g,GEN DATA,GEN listdelta)
     w0 = w1; w0_Q = w1_Q; p = q; q = q2;
   }
   TR = (GEN)DATA[14];
-  if (!gcmp0(TR)) w1_Q = poleval(w1_Q, gadd(polx[0], TR));
+  if (!gcmp0(TR)) w1_Q = TR_pol(w1_Q, TR);
   return gdiv(w1_Q,ind);
 }
 
@@ -732,8 +762,8 @@ struct poldata
 static GEN
 compute_data(GEN DATA, struct poldata PD, long d, GEN ff, GEN T,GEN p)
 {
-  long i,j,l,e,N;
-  GEN den,roo,pe,p1,p2,fk,fhk,MM,maxroot,pol,interp,bezoutC;
+  long i,j,l,e,N, lff = lg(ff);
+  GEN ffL,den,roo,pe,p1,p2,fk,fhk,MM,maxroot,pol,interp,bezoutC;    
 
   if (DEBUGLEVEL>1) { fprintferr("Entering compute_data()\n\n"); flusherr(); }
   pol = PD.pol; N = degpol(pol);
@@ -743,8 +773,9 @@ compute_data(GEN DATA, struct poldata PD, long d, GEN ff, GEN T,GEN p)
   {
     GEN Xm1 = gsub(polx[varn(pol)], gun);
     GEN TR = addis((GEN)DATA[14],1);
+    GEN mTR = negi(TR), mun = negi(gun);
     DATA[14] = (long)TR;
-    pol = poleval(pol, gsub(polx[varn(pol)], TR));
+    pol = TR_pol(pol, mTR);
     p1 = dummycopy(roo); l = lg(p1);
     for (i=1; i<l; i++) p1[i] = ladd(TR, (GEN)p1[i]);
     roo = p1;
@@ -758,29 +789,31 @@ compute_data(GEN DATA, struct poldata PD, long d, GEN ff, GEN T,GEN p)
     {
       if (degpol(interp[i]) > 0) /* do not turn polun[0] into gun */
       {
-        p1 = poleval((GEN)interp[i], Xm1);
+        p1 = TR_pol((GEN)interp[i], mun);
         interp[i] = (long)FpXQX_red(p1, NULL,p);
       }
       if (degpol(bezoutC[i]) > 0)
       {
-        p1 = poleval((GEN)bezoutC[i], Xm1);
+        p1 = TR_pol((GEN)bezoutC[i], mun);
         bezoutC[i] = (long)FpXQX_red(p1, NULL,p);
       }
     }
+    p1 = cgetg(lff, t_VEC);
+    for (i=1; i<lff; i++)
+      p1[i] = (long)FpX_red(TR_pol((GEN)ff[i], mTR), p);
+    ff = p1;
   }
   else
   {
-    GEN firstroot;    
-    long r = lg(ff);
+    GEN firstroot = cgetg(lff, t_VECSMALL);
     DATA = cgetg(15,t_VEC);
     DATA[2] = (long)p;
     DATA[4] = (long)T;
-    interp = cgetg(r, t_VEC);
-    firstroot = cgetg(r, t_VECSMALL);
+    interp = cgetg(lff, t_VEC);
     fk = cgetg(N+1,t_VEC);
-    for (l=1,j=1; j<r; j++)
+    for (l=1,j=1; j<lff; j++)
     { /* compute roots and fix ordering (Frobenius cycles) */
-      p1 = Fp_factor_irred((GEN)ff[j], p, (GEN)DATA[4]);
+      p1 = Fp_factor_irred((GEN)ff[j], p,T);
       interp[j] = (long)interpol(p1,T,p);
       firstroot[j] = l;
       for (i=1; i<lg(p1); i++,l++) fk[l] = p1[i];
@@ -799,7 +832,23 @@ compute_data(GEN DATA, struct poldata PD, long d, GEN ff, GEN T,GEN p)
   e = logint(shifti(vecmax(MM),20), p, &pe); /* overlift 2^20 [for d-1 test] */
   DATA[3] = (long)pe;
   DATA[6] = (long)roots_from_deg1(fk);
+
+#if 0
   fhk = hensel_lift_fact(pol,fk,(GEN)DATA[4],p,pe,e);
+#else
+  /* first lift in Zp to precision p^e */
+  ffL = hensel_lift_fact(pol,ff, NULL,p,pe,e);
+  fhk = NULL;
+  for (l=i=1; i<lff; i++)
+  { /* lift factorization of ff[i] in Qp[X] / T */
+    GEN L = (GEN)ffL[i];
+    long di = degpol(L);
+    p2 = cgetg(di+1, t_VEC);
+    for (j=1; j<=di; j++) p2[j] = fk[l++];
+    p1 = hensel_lift_fact(L,p2,T,p,pe,e);
+    fhk = fhk? concatsp(fhk, p1): p1;
+  }
+#endif
   DATA[5] = (long)roots_from_deg1(fhk);
 
   p1 = gmul(stoi(N), gsqrt(gpuigs(stoi(N-1),N-1),DEFAULTPREC));
@@ -832,7 +881,7 @@ static GEN
 subfields_of_given_degree(struct poldata PD,long d)
 {
   ulong av,av2;
-  long llist,i,nn;
+  long llist,i,nn,v;
   GEN listpotbl,ff,A,CSF,ESF,LSB,p,T,DATA,listdelta;
   GEN pol = PD.pol, dpol = PD.dis;
 
@@ -840,7 +889,8 @@ subfields_of_given_degree(struct poldata PD,long d)
   av = avma;
   p = choose_prime(pol,dpol,degpol(pol)/d,&ff,&listpotbl,&nn);
   if (!listpotbl) { avma=av; return cgetg(1,t_VEC); }
-  T = lift_intern(ffinit(p,nn, fetch_var()));
+  v = fetch_var(); name_var(v,"y");
+  T = lift_intern(ffinit(p,nn, v));
   DATA = NULL; LSB = cgetg(1,t_VEC); 
   i = 1; llist = lg(listpotbl);
 CHANGE: 
